/*
 * Copyright (c) 2016, Alliance for Open Media. All rights reserved
 *
 * This source code is subject to the terms of the BSD 2 Clause License and
 * the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
 * was not distributed with this source code in the LICENSE file, you can
 * obtain it at www.aomedia.org/license/software. If the Alliance for Open
 * Media Patent License 1.0 was not distributed with this source code in the
 * PATENTS file, you can obtain it at www.aomedia.org/license/patent.
 */

#include "./aom_config.h"
#include "aom/aom_integer.h"
#include "aom_mem/aom_mem.h"
#include "av1/common/blockd.h"
#include "av1/common/entropy.h"
#include "av1/common/entropymode.h"
#include "av1/common/onyxc_int.h"
#include "av1/common/scan.h"
#if CONFIG_LV_MAP
#include "av1/common/txb_common.h"
#endif

// Unconstrained Node Tree
/* clang-format off */
const aom_tree_index av1_coef_con_tree[TREE_SIZE(ENTROPY_TOKENS)] = {
  2, 6,                                // 0 = LOW_VAL
  -TWO_TOKEN, 4,                       // 1 = TWO
  -THREE_TOKEN, -FOUR_TOKEN,           // 2 = THREE
  8, 10,                               // 3 = HIGH_LOW
  -CATEGORY1_TOKEN, -CATEGORY2_TOKEN,  // 4 = CAT_ONE
  12, 14,                              // 5 = CAT_THREEFOUR
  -CATEGORY3_TOKEN, -CATEGORY4_TOKEN,  // 6 = CAT_THREE
  -CATEGORY5_TOKEN, -CATEGORY6_TOKEN   // 7 = CAT_FIVE
};
/* clang-format on */

#if CONFIG_NEW_MULTISYMBOL
/* Extra bits coded from LSB to MSB */
const aom_cdf_prob av1_cat1_cdf0[CDF_SIZE(2)] = { AOM_ICDF(20352),
                                                  AOM_ICDF(32768), 0 };
const aom_cdf_prob *av1_cat1_cdf[] = { av1_cat1_cdf0 };

const aom_cdf_prob av1_cat2_cdf0[CDF_SIZE(4)] = {
  AOM_ICDF(11963), AOM_ICDF(21121), AOM_ICDF(27719), AOM_ICDF(32768), 0
};
const aom_cdf_prob *av1_cat2_cdf[] = { av1_cat2_cdf0 };
const aom_cdf_prob av1_cat3_cdf0[CDF_SIZE(8)] = {
  AOM_ICDF(7001),  AOM_ICDF(12802), AOM_ICDF(17911),
  AOM_ICDF(22144), AOM_ICDF(25503), AOM_ICDF(28286),
  AOM_ICDF(30737), AOM_ICDF(32768), 0
};
const aom_cdf_prob *av1_cat3_cdf[] = { av1_cat3_cdf0 };

const aom_cdf_prob av1_cat4_cdf0[CDF_SIZE(16)] = { AOM_ICDF(3934),
                                                   AOM_ICDF(7460),
                                                   AOM_ICDF(10719),
                                                   AOM_ICDF(13640),
                                                   AOM_ICDF(16203),
                                                   AOM_ICDF(18500),
                                                   AOM_ICDF(20624),
                                                   AOM_ICDF(22528),
                                                   AOM_ICDF(24316),
                                                   AOM_ICDF(25919),
                                                   AOM_ICDF(27401),
                                                   AOM_ICDF(28729),
                                                   AOM_ICDF(29894),
                                                   AOM_ICDF(30938),
                                                   AOM_ICDF(31903),
                                                   AOM_ICDF(32768),
                                                   0 };
const aom_cdf_prob *av1_cat4_cdf[] = { av1_cat4_cdf0 };

const aom_cdf_prob av1_cat5_cdf0[CDF_SIZE(16)] = { AOM_ICDF(2942),
                                                   AOM_ICDF(5794),
                                                   AOM_ICDF(8473),
                                                   AOM_ICDF(11069),
                                                   AOM_ICDF(13469),
                                                   AOM_ICDF(15795),
                                                   AOM_ICDF(17980),
                                                   AOM_ICDF(20097),
                                                   AOM_ICDF(21952),
                                                   AOM_ICDF(23750),
                                                   AOM_ICDF(25439),
                                                   AOM_ICDF(27076),
                                                   AOM_ICDF(28589),
                                                   AOM_ICDF(30056),
                                                   AOM_ICDF(31434),
                                                   AOM_ICDF(32768),
                                                   0 };
const aom_cdf_prob av1_cat5_cdf1[CDF_SIZE(2)] = { AOM_ICDF(23040),
                                                  AOM_ICDF(32768), 0 };
const aom_cdf_prob *av1_cat5_cdf[] = { av1_cat5_cdf0, av1_cat5_cdf1 };

const aom_cdf_prob av1_cat6_cdf0[CDF_SIZE(16)] = {
  AOM_ICDF(2382),  AOM_ICDF(4727),  AOM_ICDF(7036),  AOM_ICDF(9309),
  AOM_ICDF(11512), AOM_ICDF(13681), AOM_ICDF(15816), AOM_ICDF(17918),
  AOM_ICDF(19892), AOM_ICDF(21835), AOM_ICDF(23748), AOM_ICDF(25632),
  AOM_ICDF(27458), AOM_ICDF(29255), AOM_ICDF(31024), AOM_ICDF(32768)
};
const aom_cdf_prob av1_cat6_cdf1[CDF_SIZE(16)] = {
  AOM_ICDF(9314),  AOM_ICDF(15584), AOM_ICDF(19741), AOM_ICDF(22540),
  AOM_ICDF(25391), AOM_ICDF(27310), AOM_ICDF(28583), AOM_ICDF(29440),
  AOM_ICDF(30493), AOM_ICDF(31202), AOM_ICDF(31672), AOM_ICDF(31988),
  AOM_ICDF(32310), AOM_ICDF(32527), AOM_ICDF(32671), AOM_ICDF(32768)
};
const aom_cdf_prob av1_cat6_cdf2[CDF_SIZE(16)] = {
  AOM_ICDF(29548), AOM_ICDF(31129), AOM_ICDF(31960), AOM_ICDF(32004),
  AOM_ICDF(32473), AOM_ICDF(32498), AOM_ICDF(32511), AOM_ICDF(32512),
  AOM_ICDF(32745), AOM_ICDF(32757), AOM_ICDF(32763), AOM_ICDF(32764),
  AOM_ICDF(32765), AOM_ICDF(32766), AOM_ICDF(32767), AOM_ICDF(32768)
};
const aom_cdf_prob av1_cat6_cdf3[CDF_SIZE(16)] = {
  AOM_ICDF(32006), AOM_ICDF(32258), AOM_ICDF(32510), AOM_ICDF(32512),
  AOM_ICDF(32638), AOM_ICDF(32639), AOM_ICDF(32640), AOM_ICDF(32641),
  AOM_ICDF(32761), AOM_ICDF(32762), AOM_ICDF(32763), AOM_ICDF(32764),
  AOM_ICDF(32765), AOM_ICDF(32766), AOM_ICDF(32767), AOM_ICDF(32768)
};
const aom_cdf_prob av1_cat6_cdf4[CDF_SIZE(4)] = {
  AOM_ICDF(32513), AOM_ICDF(32641), AOM_ICDF(32767), AOM_ICDF(32768)
};
const aom_cdf_prob *av1_cat6_cdf[] = {
  av1_cat6_cdf0, av1_cat6_cdf1, av1_cat6_cdf2, av1_cat6_cdf3, av1_cat6_cdf4
};
#endif
/* Extra bits coded from MSB to LSB */
const aom_prob av1_cat1_prob[] = { 159 };
const aom_prob av1_cat2_prob[] = { 165, 145 };
const aom_prob av1_cat3_prob[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob[] = {
  255, 255, 255, 255, 254, 254, 254, 252, 249,
  243, 230, 196, 177, 153, 140, 133, 130, 129
};

const uint16_t band_count_table[TX_SIZES_ALL][8] = {
#if CONFIG_CHROMA_2X2
  { 1, 2, 2, 3, 0, 0, 0 },
#endif
  { 1, 2, 3, 4, 3, 16 - 13, 0 },    { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 },  { 1, 2, 3, 4, 11, 1024 - 21, 0 },
#if CONFIG_TX64X64
  { 1, 2, 3, 4, 11, 4096 - 21, 0 },
#endif  // CONFIG_TX64X64
  { 1, 2, 3, 4, 8, 32 - 18, 0 },    { 1, 2, 3, 4, 8, 32 - 18, 0 },
  { 1, 2, 3, 4, 11, 128 - 21, 0 },  { 1, 2, 3, 4, 11, 128 - 21, 0 },
  { 1, 2, 3, 4, 11, 512 - 21, 0 },  { 1, 2, 3, 4, 11, 512 - 21, 0 },
  { 1, 2, 3, 4, 11, 64 - 21, 0 },   { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 },  { 1, 2, 3, 4, 11, 256 - 21, 0 },
};

const uint16_t band_cum_count_table[TX_SIZES_ALL][8] = {
#if CONFIG_CHROMA_2X2
  { 0, 1, 3, 6, 10, 13, 16, 0 },
#endif
  { 0, 1, 3, 6, 10, 13, 16, 0 },   { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 },  { 0, 1, 3, 6, 10, 21, 1024, 0 },
#if CONFIG_TX64X64
  { 0, 1, 3, 6, 10, 21, 4096, 0 },
#endif  // CONFIG_TX64X64
  { 0, 1, 3, 6, 10, 18, 32, 0 },   { 0, 1, 3, 6, 10, 18, 32, 0 },
  { 0, 1, 3, 6, 10, 21, 128, 0 },  { 0, 1, 3, 6, 10, 21, 128, 0 },
  { 0, 1, 3, 6, 10, 21, 512, 0 },  { 0, 1, 3, 6, 10, 21, 512, 0 },
  { 0, 1, 3, 6, 10, 21, 64, 0 },   { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 },  { 0, 1, 3, 6, 10, 21, 256, 0 },
};

const uint8_t av1_coefband_trans_8x8plus[MAX_TX_SQUARE] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5,
  // beyond MAXBAND_INDEX+1 all values are filled as 5
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
#if CONFIG_TX64X64
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5
#endif  // CONFIG_TX64X64
};

const uint8_t av1_coefband_trans_4x8_8x4[32] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
  4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
};

const uint8_t av1_coefband_trans_4x4[16] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5,
};

const uint8_t av1_pt_energy_class[ENTROPY_TOKENS] = { 0, 1, 2, 3, 3, 4,
                                                      4, 5, 5, 5, 5, 5 };

// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probablity of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.

// beta = 8

// Every odd line in this table can be generated from the even lines
// by averaging :
// av1_pareto8_full[l][node] = (av1_pareto8_full[l-1][node] +
//                              av1_pareto8_full[l+1][node] ) >> 1;
// Values for tokens ONE_TOKEN through CATEGORY6_TOKEN included here.
const aom_prob av1_pareto8_full[COEFF_PROB_MODELS][MODEL_NODES] = {
  { 3, 86, 128, 6, 86, 23, 88, 29 },
  { 6, 86, 128, 11, 87, 42, 91, 52 },
  { 9, 86, 129, 17, 88, 61, 94, 76 },
  { 12, 86, 129, 22, 88, 77, 97, 93 },
  { 15, 87, 129, 28, 89, 93, 100, 110 },
  { 17, 87, 129, 33, 90, 105, 103, 123 },
  { 20, 88, 130, 38, 91, 118, 106, 136 },
  { 23, 88, 130, 43, 91, 128, 108, 146 },
  { 26, 89, 131, 48, 92, 139, 111, 156 },
  { 28, 89, 131, 53, 93, 147, 114, 163 },
  { 31, 90, 131, 58, 94, 156, 117, 171 },
  { 34, 90, 131, 62, 94, 163, 119, 177 },
  { 37, 90, 132, 66, 95, 171, 122, 184 },
  { 39, 90, 132, 70, 96, 177, 124, 189 },
  { 42, 91, 132, 75, 97, 183, 127, 194 },
  { 44, 91, 132, 79, 97, 188, 129, 198 },
  { 47, 92, 133, 83, 98, 193, 132, 202 },
  { 49, 92, 133, 86, 99, 197, 134, 205 },
  { 52, 93, 133, 90, 100, 201, 137, 208 },
  { 54, 93, 133, 94, 100, 204, 139, 211 },
  { 57, 94, 134, 98, 101, 208, 142, 214 },
  { 59, 94, 134, 101, 102, 211, 144, 216 },
  { 62, 94, 135, 105, 103, 214, 146, 218 },
  { 64, 94, 135, 108, 103, 216, 148, 220 },
  { 66, 95, 135, 111, 104, 219, 151, 222 },
  { 68, 95, 135, 114, 105, 221, 153, 223 },
  { 71, 96, 136, 117, 106, 224, 155, 225 },
  { 73, 96, 136, 120, 106, 225, 157, 226 },
  { 76, 97, 136, 123, 107, 227, 159, 228 },
  { 78, 97, 136, 126, 108, 229, 160, 229 },
  { 80, 98, 137, 129, 109, 231, 162, 231 },
  { 82, 98, 137, 131, 109, 232, 164, 232 },
  { 84, 98, 138, 134, 110, 234, 166, 233 },
  { 86, 98, 138, 137, 111, 235, 168, 234 },
  { 89, 99, 138, 140, 112, 236, 170, 235 },
  { 91, 99, 138, 142, 112, 237, 171, 235 },
  { 93, 100, 139, 145, 113, 238, 173, 236 },
  { 95, 100, 139, 147, 114, 239, 174, 237 },
  { 97, 101, 140, 149, 115, 240, 176, 238 },
  { 99, 101, 140, 151, 115, 241, 177, 238 },
  { 101, 102, 140, 154, 116, 242, 179, 239 },
  { 103, 102, 140, 156, 117, 242, 180, 239 },
  { 105, 103, 141, 158, 118, 243, 182, 240 },
  { 107, 103, 141, 160, 118, 243, 183, 240 },
  { 109, 104, 141, 162, 119, 244, 185, 241 },
  { 111, 104, 141, 164, 119, 244, 186, 241 },
  { 113, 104, 142, 166, 120, 245, 187, 242 },
  { 114, 104, 142, 168, 121, 245, 188, 242 },
  { 116, 105, 143, 170, 122, 246, 190, 243 },
  { 118, 105, 143, 171, 122, 246, 191, 243 },
  { 120, 106, 143, 173, 123, 247, 192, 244 },
  { 121, 106, 143, 175, 124, 247, 193, 244 },
  { 123, 107, 144, 177, 125, 248, 195, 244 },
  { 125, 107, 144, 178, 125, 248, 196, 244 },
  { 127, 108, 145, 180, 126, 249, 197, 245 },
  { 128, 108, 145, 181, 127, 249, 198, 245 },
  { 130, 109, 145, 183, 128, 249, 199, 245 },
  { 132, 109, 145, 184, 128, 249, 200, 245 },
  { 134, 110, 146, 186, 129, 250, 201, 246 },
  { 135, 110, 146, 187, 130, 250, 202, 246 },
  { 137, 111, 147, 189, 131, 251, 203, 246 },
  { 138, 111, 147, 190, 131, 251, 204, 246 },
  { 140, 112, 147, 192, 132, 251, 205, 247 },
  { 141, 112, 147, 193, 132, 251, 206, 247 },
  { 143, 113, 148, 194, 133, 251, 207, 247 },
  { 144, 113, 148, 195, 134, 251, 207, 247 },
  { 146, 114, 149, 197, 135, 252, 208, 248 },
  { 147, 114, 149, 198, 135, 252, 209, 248 },
  { 149, 115, 149, 199, 136, 252, 210, 248 },
  { 150, 115, 149, 200, 137, 252, 210, 248 },
  { 152, 115, 150, 201, 138, 252, 211, 248 },
  { 153, 115, 150, 202, 138, 252, 212, 248 },
  { 155, 116, 151, 204, 139, 253, 213, 249 },
  { 156, 116, 151, 205, 139, 253, 213, 249 },
  { 158, 117, 151, 206, 140, 253, 214, 249 },
  { 159, 117, 151, 207, 141, 253, 215, 249 },
  { 161, 118, 152, 208, 142, 253, 216, 249 },
  { 162, 118, 152, 209, 142, 253, 216, 249 },
  { 163, 119, 153, 210, 143, 253, 217, 249 },
  { 164, 119, 153, 211, 143, 253, 217, 249 },
  { 166, 120, 153, 212, 144, 254, 218, 250 },
  { 167, 120, 153, 212, 145, 254, 219, 250 },
  { 168, 121, 154, 213, 146, 254, 220, 250 },
  { 169, 121, 154, 214, 146, 254, 220, 250 },
  { 171, 122, 155, 215, 147, 254, 221, 250 },
  { 172, 122, 155, 216, 147, 254, 221, 250 },
  { 173, 123, 155, 217, 148, 254, 222, 250 },
  { 174, 123, 155, 217, 149, 254, 222, 250 },
  { 176, 124, 156, 218, 150, 254, 223, 250 },
  { 177, 124, 156, 219, 150, 254, 223, 250 },
  { 178, 125, 157, 220, 151, 254, 224, 251 },
  { 179, 125, 157, 220, 151, 254, 224, 251 },
  { 180, 126, 157, 221, 152, 254, 225, 251 },
  { 181, 126, 157, 221, 152, 254, 225, 251 },
  { 183, 127, 158, 222, 153, 254, 226, 251 },
  { 184, 127, 158, 223, 154, 254, 226, 251 },
  { 185, 128, 159, 224, 155, 255, 227, 251 },
  { 186, 128, 159, 224, 155, 255, 227, 251 },
  { 187, 129, 160, 225, 156, 255, 228, 251 },
  { 188, 130, 160, 225, 156, 255, 228, 251 },
  { 189, 131, 160, 226, 157, 255, 228, 251 },
  { 190, 131, 160, 226, 158, 255, 228, 251 },
  { 191, 132, 161, 227, 159, 255, 229, 251 },
  { 192, 132, 161, 227, 159, 255, 229, 251 },
  { 193, 133, 162, 228, 160, 255, 230, 252 },
  { 194, 133, 162, 229, 160, 255, 230, 252 },
  { 195, 134, 163, 230, 161, 255, 231, 252 },
  { 196, 134, 163, 230, 161, 255, 231, 252 },
  { 197, 135, 163, 231, 162, 255, 231, 252 },
  { 198, 135, 163, 231, 162, 255, 231, 252 },
  { 199, 136, 164, 232, 163, 255, 232, 252 },
  { 200, 136, 164, 232, 164, 255, 232, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 202, 138, 166, 233, 166, 255, 233, 252 },
  { 203, 138, 166, 233, 166, 255, 233, 252 },
  { 204, 139, 166, 234, 167, 255, 234, 252 },
  { 205, 139, 166, 234, 167, 255, 234, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 207, 141, 168, 236, 169, 255, 235, 252 },
  { 208, 141, 168, 236, 170, 255, 235, 252 },
  { 209, 142, 169, 237, 171, 255, 236, 252 },
  { 209, 143, 169, 237, 171, 255, 236, 252 },
  { 210, 144, 169, 237, 172, 255, 236, 252 },
  { 211, 144, 169, 237, 172, 255, 236, 252 },
  { 212, 145, 170, 238, 173, 255, 237, 252 },
  { 213, 145, 170, 238, 173, 255, 237, 252 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 216, 148, 173, 240, 176, 255, 238, 253 },
  { 217, 148, 173, 240, 176, 255, 238, 253 },
  { 218, 149, 173, 241, 177, 255, 239, 253 },
  { 218, 149, 173, 241, 178, 255, 239, 253 },
  { 219, 150, 174, 241, 179, 255, 239, 253 },
  { 219, 151, 174, 241, 179, 255, 239, 253 },
  { 220, 152, 175, 242, 180, 255, 240, 253 },
  { 221, 152, 175, 242, 180, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 225, 156, 178, 244, 184, 255, 241, 253 },
  { 225, 157, 178, 244, 184, 255, 241, 253 },
  { 226, 158, 179, 244, 185, 255, 242, 253 },
  { 227, 158, 179, 244, 185, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 230, 161, 182, 246, 188, 255, 243, 253 },
  { 230, 162, 182, 246, 188, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 234, 166, 185, 247, 192, 255, 244, 253 },
  { 234, 167, 185, 247, 192, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 237, 171, 189, 249, 196, 255, 245, 254 },
  { 237, 172, 189, 249, 196, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 240, 175, 192, 249, 199, 255, 246, 254 },
  { 240, 176, 192, 249, 199, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 242, 179, 195, 250, 202, 255, 246, 254 },
  { 242, 180, 195, 250, 202, 255, 246, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 243, 182, 197, 251, 204, 255, 247, 254 },
  { 243, 183, 197, 251, 204, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 245, 186, 200, 251, 207, 255, 247, 254 },
  { 245, 187, 200, 251, 207, 255, 247, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 189, 202, 252, 208, 255, 248, 254 },
  { 246, 190, 202, 252, 208, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 192, 204, 252, 210, 255, 248, 254 },
  { 247, 193, 204, 252, 210, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 195, 206, 252, 212, 255, 249, 254 },
  { 248, 196, 206, 252, 212, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 198, 208, 253, 214, 255, 249, 254 },
  { 249, 199, 209, 253, 214, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 201, 211, 253, 215, 255, 249, 254 },
  { 250, 202, 211, 253, 215, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 251, 204, 213, 253, 217, 255, 250, 254 },
  { 251, 205, 213, 253, 217, 255, 250, 254 },
  { 251, 206, 214, 254, 218, 255, 250, 254 },
  { 251, 206, 215, 254, 218, 255, 250, 254 },
  { 252, 207, 216, 254, 219, 255, 250, 254 },
  { 252, 208, 216, 254, 219, 255, 250, 254 },
  { 252, 209, 217, 254, 220, 255, 250, 254 },
  { 252, 210, 217, 254, 220, 255, 250, 254 },
  { 252, 211, 218, 254, 221, 255, 250, 254 },
  { 252, 212, 218, 254, 221, 255, 250, 254 },
  { 253, 213, 219, 254, 222, 255, 250, 254 },
  { 253, 213, 220, 254, 222, 255, 250, 254 },
  { 253, 214, 221, 254, 223, 255, 250, 254 },
  { 253, 215, 221, 254, 223, 255, 250, 254 },
  { 253, 216, 222, 254, 224, 255, 251, 254 },
  { 253, 217, 223, 254, 224, 255, 251, 254 },
  { 253, 218, 224, 254, 225, 255, 251, 254 },
  { 253, 219, 224, 254, 225, 255, 251, 254 },
  { 254, 220, 225, 254, 225, 255, 251, 254 },
  { 254, 221, 226, 254, 225, 255, 251, 254 },
  { 254, 222, 227, 255, 226, 255, 251, 254 },
  { 254, 223, 227, 255, 226, 255, 251, 254 },
  { 254, 224, 228, 255, 227, 255, 251, 254 },
  { 254, 225, 229, 255, 227, 255, 251, 254 },
  { 254, 226, 230, 255, 228, 255, 251, 254 },
  { 254, 227, 230, 255, 229, 255, 251, 254 },
  { 255, 228, 231, 255, 230, 255, 251, 254 },
  { 255, 229, 232, 255, 230, 255, 251, 254 },
  { 255, 230, 233, 255, 231, 255, 252, 254 },
  { 255, 231, 234, 255, 231, 255, 252, 254 },
  { 255, 232, 235, 255, 232, 255, 252, 254 },
  { 255, 233, 236, 255, 232, 255, 252, 254 },
  { 255, 235, 237, 255, 233, 255, 252, 254 },
  { 255, 236, 238, 255, 234, 255, 252, 254 },
  { 255, 238, 240, 255, 235, 255, 252, 255 },
  { 255, 239, 241, 255, 235, 255, 252, 254 },
  { 255, 241, 243, 255, 236, 255, 252, 254 },
  { 255, 243, 245, 255, 237, 255, 252, 254 },
  { 255, 246, 247, 255, 239, 255, 253, 255 },
};

// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probability of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.
//
// The full source code of the generating program is available in:
// tools/gen_constrained_tokenset.py
//
// Values for tokens TWO_TOKEN through CATEGORY6_TOKEN included
// in the table here : the ONE_TOKEN probability is
// removed and the probabilities rescaled.
//
// ZERO_TOKEN and ONE_TOKEN are coded as one CDF,
// and EOB_TOKEN is coded as flags outside this coder.
const aom_cdf_prob av1_pareto8_tail_probs[COEFF_PROB_MODELS][TAIL_NODES] = {
  { 128, 127, 127, 252, 497, 969, 1839, 3318, 25511 },
  { 256, 254, 251, 496, 966, 1834, 3308, 5408, 19995 },
  { 383, 378, 373, 732, 1408, 2605, 4470, 6646, 15773 },
  { 511, 502, 493, 961, 1824, 3289, 5373, 7298, 12517 },
  { 638, 625, 611, 1182, 2215, 3894, 6064, 7548, 9991 },
  { 766, 746, 726, 1396, 2582, 4428, 6578, 7529, 8017 },
  { 893, 866, 839, 1603, 2927, 4896, 6945, 7332, 6467 },
  { 1020, 984, 950, 1803, 3250, 5305, 7191, 7022, 5243 },
  { 1147, 1102, 1059, 1996, 3552, 5659, 7338, 6646, 4269 },
  { 1274, 1218, 1166, 2183, 3835, 5963, 7403, 6234, 3492 },
  { 1400, 1334, 1270, 2363, 4099, 6223, 7401, 5809, 2869 },
  { 1527, 1447, 1372, 2537, 4345, 6442, 7346, 5386, 2366 },
  { 1654, 1560, 1473, 2704, 4574, 6624, 7247, 4973, 1959 },
  { 1780, 1672, 1571, 2866, 4787, 6771, 7114, 4579, 1628 },
  { 1906, 1782, 1667, 3022, 4984, 6889, 6954, 4206, 1358 },
  { 2032, 1891, 1762, 3172, 5167, 6979, 6773, 3856, 1136 },
  { 2158, 2000, 1854, 3316, 5335, 7044, 6577, 3530, 954 },
  { 2284, 2106, 1944, 3455, 5490, 7087, 6370, 3229, 803 },
  { 2410, 2212, 2032, 3588, 5632, 7109, 6155, 2951, 679 },
  { 2535, 2317, 2119, 3717, 5761, 7113, 5936, 2695, 575 },
  { 2661, 2420, 2203, 3840, 5880, 7101, 5714, 2461, 488 },
  { 2786, 2522, 2286, 3958, 5987, 7074, 5493, 2246, 416 },
  { 2911, 2624, 2367, 4072, 6083, 7033, 5273, 2050, 355 },
  { 3037, 2724, 2446, 4180, 6170, 6981, 5055, 1871, 304 },
  { 3162, 2822, 2523, 4284, 6247, 6919, 4842, 1708, 261 },
  { 3286, 2920, 2599, 4384, 6315, 6848, 4633, 1559, 224 },
  { 3411, 3017, 2672, 4478, 6374, 6768, 4430, 1424, 194 },
  { 3536, 3112, 2745, 4569, 6426, 6681, 4232, 1300, 167 },
  { 3660, 3207, 2815, 4656, 6469, 6588, 4040, 1188, 145 },
  { 3785, 3300, 2883, 4738, 6505, 6490, 3855, 1086, 126 },
  { 3909, 3392, 2950, 4817, 6534, 6387, 3677, 993, 109 },
  { 4033, 3483, 3015, 4891, 6557, 6281, 3505, 908, 95 },
  { 4157, 3573, 3079, 4962, 6573, 6170, 3340, 831, 83 },
  { 4281, 3662, 3141, 5029, 6584, 6058, 3181, 760, 72 },
  { 4405, 3750, 3201, 5093, 6588, 5943, 3029, 696, 63 },
  { 4529, 3837, 3260, 5152, 6587, 5826, 2883, 638, 56 },
  { 4652, 3922, 3317, 5209, 6582, 5709, 2744, 584, 49 },
  { 4775, 4007, 3373, 5262, 6572, 5590, 2610, 536, 43 },
  { 4899, 4090, 3427, 5312, 6557, 5470, 2483, 492, 38 },
  { 5022, 4173, 3480, 5359, 6538, 5351, 2361, 451, 33 },
  { 5145, 4254, 3531, 5403, 6515, 5231, 2246, 414, 29 },
  { 5268, 4334, 3581, 5443, 6489, 5112, 2135, 380, 26 },
  { 5391, 4414, 3629, 5481, 6458, 4993, 2029, 350, 23 },
  { 5514, 4492, 3676, 5515, 6425, 4875, 1929, 321, 21 },
  { 5637, 4569, 3721, 5548, 6388, 4758, 1833, 296, 18 },
  { 5759, 4645, 3766, 5577, 6349, 4642, 1742, 272, 16 },
  { 5881, 4720, 3808, 5604, 6307, 4528, 1656, 250, 14 },
  { 6004, 4794, 3849, 5628, 6262, 4414, 1573, 231, 13 },
  { 6126, 4867, 3890, 5649, 6215, 4302, 1495, 213, 11 },
  { 6248, 4939, 3928, 5669, 6166, 4192, 1420, 196, 10 },
  { 6370, 5010, 3966, 5686, 6114, 4083, 1349, 181, 9 },
  { 6492, 5080, 4002, 5700, 6061, 3976, 1282, 167, 8 },
  { 6614, 5149, 4037, 5712, 6006, 3871, 1218, 154, 7 },
  { 6735, 5217, 4070, 5723, 5950, 3767, 1157, 142, 7 },
  { 6857, 5284, 4103, 5731, 5891, 3666, 1099, 131, 6 },
  { 6978, 5351, 4134, 5737, 5832, 3566, 1044, 121, 5 },
  { 7099, 5415, 4164, 5741, 5771, 3469, 992, 112, 5 },
  { 7221, 5479, 4192, 5743, 5709, 3373, 943, 104, 4 },
  { 7342, 5542, 4220, 5743, 5646, 3279, 896, 96, 4 },
  { 7462, 5604, 4246, 5742, 5583, 3187, 851, 89, 4 },
  { 7584, 5665, 4272, 5739, 5518, 3097, 808, 82, 3 },
  { 7704, 5725, 4296, 5734, 5453, 3009, 768, 76, 3 },
  { 7825, 5784, 4318, 5727, 5386, 2924, 730, 71, 3 },
  { 7945, 5843, 4341, 5719, 5320, 2840, 693, 65, 2 },
  { 8066, 5900, 4361, 5709, 5252, 2758, 659, 61, 2 },
  { 8186, 5956, 4381, 5698, 5185, 2678, 626, 56, 2 },
  { 8306, 6011, 4400, 5685, 5117, 2600, 595, 52, 2 },
  { 8426, 6066, 4418, 5671, 5049, 2523, 565, 48, 2 },
  { 8547, 6119, 4434, 5655, 4981, 2449, 537, 45, 1 },
  { 8666, 6171, 4450, 5638, 4912, 2377, 511, 42, 1 },
  { 8786, 6223, 4465, 5620, 4843, 2306, 485, 39, 1 },
  { 8906, 6274, 4478, 5600, 4775, 2237, 461, 36, 1 },
  { 9025, 6323, 4491, 5580, 4706, 2170, 438, 34, 1 },
  { 9144, 6372, 4503, 5558, 4637, 2105, 417, 31, 1 },
  { 9264, 6420, 4514, 5535, 4568, 2041, 396, 29, 1 },
  { 9383, 6467, 4524, 5511, 4500, 1979, 376, 27, 1 },
  { 9502, 6513, 4532, 5486, 4432, 1919, 358, 25, 1 },
  { 9621, 6558, 4541, 5460, 4364, 1860, 340, 23, 1 },
  { 9740, 6602, 4548, 5433, 4296, 1803, 323, 22, 1 },
  { 9859, 6645, 4554, 5405, 4229, 1748, 307, 20, 1 },
  { 9978, 6688, 4559, 5376, 4161, 1694, 292, 19, 1 },
  { 10096, 6729, 4564, 5347, 4094, 1641, 278, 18, 1 },
  { 10215, 6770, 4568, 5316, 4028, 1590, 264, 16, 1 },
  { 10333, 6809, 4571, 5285, 3962, 1541, 251, 15, 1 },
  { 10452, 6848, 4573, 5253, 3896, 1492, 239, 14, 1 },
  { 10570, 6886, 4574, 5220, 3831, 1446, 227, 13, 1 },
  { 10688, 6923, 4575, 5186, 3767, 1400, 216, 12, 1 },
  { 10806, 6959, 4575, 5152, 3702, 1356, 205, 12, 1 },
  { 10924, 6994, 4574, 5117, 3639, 1313, 195, 11, 1 },
  { 11041, 7029, 4572, 5082, 3576, 1271, 186, 10, 1 },
  { 11159, 7062, 4570, 5046, 3513, 1231, 177, 9, 1 },
  { 11277, 7095, 4566, 5009, 3451, 1192, 168, 9, 1 },
  { 11394, 7127, 4563, 4972, 3390, 1153, 160, 8, 1 },
  { 11512, 7158, 4558, 4934, 3329, 1116, 152, 8, 1 },
  { 11629, 7188, 4553, 4896, 3269, 1080, 145, 7, 1 },
  { 11746, 7217, 4547, 4857, 3210, 1045, 138, 7, 1 },
  { 11864, 7245, 4540, 4818, 3151, 1012, 131, 6, 1 },
  { 11980, 7273, 4533, 4779, 3093, 979, 124, 6, 1 },
  { 12097, 7300, 4525, 4739, 3035, 947, 118, 6, 1 },
  { 12215, 7326, 4516, 4698, 2978, 916, 113, 5, 1 },
  { 12331, 7351, 4507, 4658, 2922, 886, 107, 5, 1 },
  { 12448, 7375, 4497, 4617, 2866, 857, 102, 5, 1 },
  { 12564, 7398, 4487, 4576, 2812, 829, 97, 4, 1 },
  { 12681, 7421, 4476, 4534, 2757, 802, 92, 4, 1 },
  { 12797, 7443, 4464, 4492, 2704, 775, 88, 4, 1 },
  { 12914, 7464, 4452, 4450, 2651, 749, 84, 3, 1 },
  { 13030, 7484, 4439, 4408, 2599, 725, 79, 3, 1 },
  { 13147, 7503, 4426, 4365, 2547, 700, 76, 3, 1 },
  { 13262, 7522, 4412, 4322, 2497, 677, 72, 3, 1 },
  { 13378, 7539, 4398, 4280, 2447, 654, 68, 3, 1 },
  { 13494, 7556, 4383, 4237, 2397, 632, 65, 3, 1 },
  { 13610, 7573, 4368, 4193, 2348, 611, 62, 2, 1 },
  { 13726, 7588, 4352, 4150, 2300, 590, 59, 2, 1 },
  { 13841, 7602, 4335, 4107, 2253, 571, 56, 2, 1 },
  { 13957, 7616, 4318, 4063, 2207, 551, 53, 2, 1 },
  { 14072, 7629, 4301, 4019, 2161, 532, 51, 2, 1 },
  { 14188, 7641, 4283, 3976, 2115, 514, 48, 2, 1 },
  { 14302, 7652, 4265, 3932, 2071, 497, 46, 2, 1 },
  { 14418, 7663, 4246, 3888, 2027, 480, 44, 1, 1 },
  { 14533, 7673, 4227, 3844, 1984, 463, 42, 1, 1 },
  { 14649, 7682, 4207, 3800, 1941, 447, 40, 1, 1 },
  { 14763, 7690, 4187, 3757, 1899, 432, 38, 1, 1 },
  { 14878, 7698, 4166, 3713, 1858, 417, 36, 1, 1 },
  { 14993, 7705, 4146, 3669, 1817, 402, 34, 1, 1 },
  { 15109, 7711, 4124, 3625, 1777, 388, 32, 1, 1 },
  { 15223, 7715, 4103, 3581, 1738, 375, 31, 1, 1 },
  { 15337, 7720, 4081, 3538, 1699, 362, 29, 1, 1 },
  { 15452, 7724, 4058, 3494, 1661, 349, 28, 1, 1 },
  { 15567, 7727, 4035, 3450, 1624, 337, 26, 1, 1 },
  { 15681, 7729, 4012, 3407, 1587, 325, 25, 1, 1 },
  { 15795, 7730, 3989, 3364, 1551, 313, 24, 1, 1 },
  { 15909, 7731, 3965, 3320, 1516, 302, 23, 1, 1 },
  { 16024, 7731, 3940, 3277, 1481, 291, 22, 1, 1 },
  { 16138, 7730, 3916, 3234, 1446, 281, 21, 1, 1 },
  { 16252, 7728, 3891, 3191, 1413, 271, 20, 1, 1 },
  { 16366, 7726, 3866, 3148, 1380, 261, 19, 1, 1 },
  { 16480, 7723, 3840, 3106, 1347, 252, 18, 1, 1 },
  { 16594, 7720, 3814, 3063, 1315, 243, 17, 1, 1 },
  { 16708, 7715, 3788, 3021, 1284, 234, 16, 1, 1 },
  { 16822, 7710, 3762, 2979, 1253, 225, 15, 1, 1 },
  { 16936, 7704, 3735, 2937, 1223, 217, 14, 1, 1 },
  { 17050, 7697, 3708, 2895, 1193, 209, 14, 1, 1 },
  { 17162, 7690, 3681, 2854, 1164, 202, 13, 1, 1 },
  { 17276, 7682, 3654, 2812, 1136, 194, 12, 1, 1 },
  { 17389, 7673, 3626, 2771, 1108, 187, 12, 1, 1 },
  { 17504, 7663, 3598, 2730, 1080, 180, 11, 1, 1 },
  { 17617, 7653, 3570, 2689, 1053, 173, 11, 1, 1 },
  { 17730, 7642, 3541, 2649, 1027, 167, 10, 1, 1 },
  { 17843, 7630, 3513, 2608, 1001, 161, 10, 1, 1 },
  { 17957, 7618, 3484, 2569, 975, 154, 9, 1, 1 },
  { 18069, 7605, 3455, 2529, 950, 149, 9, 1, 1 },
  { 18183, 7591, 3426, 2489, 926, 143, 8, 1, 1 },
  { 18296, 7576, 3396, 2450, 902, 138, 8, 1, 1 },
  { 18410, 7562, 3366, 2411, 878, 132, 7, 1, 1 },
  { 18523, 7545, 3337, 2372, 855, 127, 7, 1, 1 },
  { 18636, 7529, 3306, 2333, 833, 122, 7, 1, 1 },
  { 18749, 7511, 3276, 2295, 811, 118, 6, 1, 1 },
  { 18862, 7493, 3246, 2257, 789, 113, 6, 1, 1 },
  { 18975, 7474, 3215, 2219, 768, 109, 6, 1, 1 },
  { 19088, 7455, 3185, 2182, 747, 104, 5, 1, 1 },
  { 19201, 7435, 3154, 2144, 727, 100, 5, 1, 1 },
  { 19314, 7414, 3123, 2107, 707, 96, 5, 1, 1 },
  { 19427, 7392, 3092, 2071, 687, 92, 5, 1, 1 },
  { 19541, 7370, 3060, 2034, 668, 89, 4, 1, 1 },
  { 19654, 7347, 3029, 1998, 649, 85, 4, 1, 1 },
  { 19766, 7323, 2997, 1963, 631, 82, 4, 1, 1 },
  { 19878, 7299, 2966, 1927, 613, 79, 4, 1, 1 },
  { 19991, 7274, 2934, 1892, 596, 75, 4, 1, 1 },
  { 20105, 7248, 2902, 1857, 579, 72, 3, 1, 1 },
  { 20218, 7222, 2870, 1822, 562, 69, 3, 1, 1 },
  { 20331, 7195, 2838, 1788, 545, 66, 3, 1, 1 },
  { 20443, 7167, 2806, 1754, 529, 64, 3, 1, 1 },
  { 20556, 7138, 2774, 1720, 514, 61, 3, 1, 1 },
  { 20670, 7109, 2741, 1687, 498, 58, 3, 1, 1 },
  { 20783, 7079, 2709, 1654, 483, 56, 2, 1, 1 },
  { 20895, 7049, 2676, 1621, 469, 54, 2, 1, 1 },
  { 21008, 7017, 2644, 1589, 455, 51, 2, 1, 1 },
  { 21121, 6985, 2611, 1557, 441, 49, 2, 1, 1 },
  { 21234, 6953, 2578, 1525, 427, 47, 2, 1, 1 },
  { 21347, 6919, 2545, 1494, 414, 45, 2, 1, 1 },
  { 21460, 6885, 2513, 1462, 401, 43, 2, 1, 1 },
  { 21573, 6850, 2480, 1432, 388, 41, 2, 1, 1 },
  { 21687, 6815, 2447, 1401, 375, 39, 2, 1, 1 },
  { 21801, 6778, 2414, 1371, 363, 38, 1, 1, 1 },
  { 21914, 6741, 2381, 1341, 352, 36, 1, 1, 1 },
  { 22028, 6704, 2348, 1311, 340, 34, 1, 1, 1 },
  { 22141, 6665, 2315, 1282, 329, 33, 1, 1, 1 },
  { 22255, 6626, 2282, 1253, 318, 31, 1, 1, 1 },
  { 22368, 6586, 2249, 1225, 307, 30, 1, 1, 1 },
  { 22482, 6546, 2216, 1196, 297, 28, 1, 1, 1 },
  { 22595, 6505, 2183, 1169, 286, 27, 1, 1, 1 },
  { 22709, 6463, 2149, 1141, 277, 26, 1, 1, 1 },
  { 22823, 6420, 2116, 1114, 267, 25, 1, 1, 1 },
  { 22938, 6377, 2083, 1087, 257, 23, 1, 1, 1 },
  { 23053, 6332, 2050, 1060, 248, 22, 1, 1, 1 },
  { 23167, 6287, 2017, 1034, 239, 21, 1, 1, 1 },
  { 23280, 6242, 1984, 1008, 231, 20, 1, 1, 1 },
  { 23396, 6195, 1951, 982, 222, 19, 1, 1, 1 },
  { 23510, 6148, 1918, 957, 214, 18, 1, 1, 1 },
  { 23625, 6100, 1885, 932, 206, 17, 1, 1, 1 },
  { 23741, 6051, 1852, 907, 198, 16, 1, 1, 1 },
  { 23855, 6002, 1819, 883, 190, 16, 1, 1, 1 },
  { 23971, 5951, 1786, 859, 183, 15, 1, 1, 1 },
  { 24087, 5900, 1753, 835, 176, 14, 1, 1, 1 },
  { 24203, 5848, 1720, 812, 169, 13, 1, 1, 1 },
  { 24318, 5796, 1687, 789, 162, 13, 1, 1, 1 },
  { 24435, 5742, 1655, 766, 155, 12, 1, 1, 1 },
  { 24552, 5688, 1622, 743, 149, 11, 1, 1, 1 },
  { 24669, 5632, 1589, 721, 143, 11, 1, 1, 1 },
  { 24786, 5576, 1557, 699, 137, 10, 1, 1, 1 },
  { 24903, 5519, 1524, 678, 131, 10, 1, 1, 1 },
  { 25021, 5462, 1491, 657, 125, 9, 1, 1, 1 },
  { 25139, 5403, 1459, 636, 120, 8, 1, 1, 1 },
  { 25258, 5343, 1427, 615, 114, 8, 1, 1, 1 },
  { 25376, 5283, 1394, 595, 109, 8, 1, 1, 1 },
  { 25496, 5221, 1362, 575, 104, 7, 1, 1, 1 },
  { 25614, 5159, 1330, 556, 99, 7, 1, 1, 1 },
  { 25735, 5096, 1298, 536, 94, 6, 1, 1, 1 },
  { 25856, 5031, 1265, 517, 90, 6, 1, 1, 1 },
  { 25977, 4966, 1233, 499, 85, 5, 1, 1, 1 },
  { 26098, 4899, 1202, 480, 81, 5, 1, 1, 1 },
  { 26220, 4831, 1170, 462, 77, 5, 1, 1, 1 },
  { 26343, 4763, 1138, 444, 73, 4, 1, 1, 1 },
  { 26466, 4693, 1106, 427, 69, 4, 1, 1, 1 },
  { 26589, 4622, 1075, 410, 65, 4, 1, 1, 1 },
  { 26713, 4550, 1043, 393, 62, 4, 1, 1, 1 },
  { 26840, 4476, 1012, 376, 58, 3, 1, 1, 1 },
  { 26966, 4401, 980, 360, 55, 3, 1, 1, 1 },
  { 27092, 4325, 949, 344, 52, 3, 1, 1, 1 },
  { 27220, 4248, 918, 328, 48, 3, 1, 1, 1 },
  { 27350, 4169, 886, 313, 45, 2, 1, 1, 1 },
  { 27480, 4088, 855, 298, 42, 2, 1, 1, 1 },
  { 27610, 4006, 824, 283, 40, 2, 1, 1, 1 },
  { 27743, 3922, 793, 268, 37, 2, 1, 1, 1 },
  { 27876, 3837, 762, 254, 34, 2, 1, 1, 1 },
  { 28011, 3749, 731, 240, 32, 2, 1, 1, 1 },
  { 28147, 3659, 701, 227, 30, 1, 1, 1, 1 },
  { 28286, 3568, 670, 213, 27, 1, 1, 1, 1 },
  { 28426, 3474, 639, 200, 25, 1, 1, 1, 1 },
  { 28569, 3377, 608, 187, 23, 1, 1, 1, 1 },
  { 28714, 3278, 577, 174, 21, 1, 1, 1, 1 },
  { 28860, 3176, 547, 162, 19, 1, 1, 1, 1 },
  { 29010, 3071, 516, 150, 17, 1, 1, 1, 1 },
  { 29163, 2962, 485, 138, 16, 1, 1, 1, 1 },
  { 29320, 2849, 454, 127, 14, 1, 1, 1, 1 },
  { 29483, 2731, 423, 115, 12, 1, 1, 1, 1 },
  { 29650, 2608, 391, 104, 11, 1, 1, 1, 1 },
  { 29823, 2479, 360, 93, 9, 1, 1, 1, 1 },
  { 30002, 2343, 328, 83, 8, 1, 1, 1, 1 },
  { 30192, 2198, 295, 72, 7, 1, 1, 1, 1 },
  { 30393, 2041, 262, 62, 6, 1, 1, 1, 1 },
  { 30612, 1869, 227, 52, 4, 1, 1, 1, 1 },
  { 30853, 1676, 191, 41, 3, 1, 1, 1, 1 },
  { 31131, 1448, 152, 31, 2, 1, 1, 1, 1 },
  { 31486, 1150, 107, 20, 1, 1, 1, 1, 1 },
};

/* clang-format off */
#if CONFIG_Q_ADAPT_PROBS
const av1_coeff_probs_model
default_qctx_coef_probs[QCTX_BINS][TX_SIZES][PLANE_TYPES] = {
    {  // Q_Index 0
#if CONFIG_CHROMA_2X2
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {182,  34, 137}, { 79,  39, 103}, { 10,  28,  51},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 45,  88, 147}, { 46,  80, 140}, { 25,  69, 119},
                        { 12,  57,  96}, {  4,  41,  65}, {  1,  20,  31},
                    },
                    {  // band 2
                        { 58, 124, 190}, { 39, 106, 178}, { 16,  86, 147},
                        {  7,  69, 114}, {  3,  50,  80}, {  1,  25,  42},
                    },
                    {  // band 3
                        { 90, 138, 215}, { 54, 116, 198}, { 18,  86, 155},
                        {  5,  62, 112}, {  1,  38,  68}, {  1,  17,  30},
                    },
                    {  // band 4
                        {126, 149, 231}, { 82, 114, 211}, { 21,  80, 157},
                        {  6,  56, 105}, {  1,  36,  64}, {  1,  17,  31},
                    },
                    {  // band 5
                        {171,  56, 236}, {140,  54, 219}, { 57,  45, 167},
                        { 26,  36, 113}, { 11,  29,  72}, {  3,  18,  39},
                    },
                },
                {  // Intra
                    {  // band 0
                        {153, 122, 186}, {106, 109, 171}, { 36,  84, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 27, 151, 201}, { 34, 131, 199}, { 23, 102, 161},
                        { 10,  80, 120}, {  4,  52,  78}, {  1,  24,  37},
                    },
                    {  // band 2
                        { 43, 158, 213}, { 35, 133, 203}, {  8,  92, 151},
                        {  2,  64, 106}, {  1,  36,  60}, {  1,  13,  24},
                    },
                    {  // band 3
                        { 68, 167, 223}, { 36, 135, 211}, {  9,  94, 157},
                        {  2,  67, 112}, {  1,  40,  68}, {  1,  17,  31},
                    },
                    {  // band 4
                        {131, 146, 237}, { 72, 119, 223}, { 17,  82, 164},
                        {  4,  55, 107}, {  1,  34,  63}, {  1,  16,  29},
                    },
                    {  // band 5
                        {184,  68, 244}, {153,  59, 232}, { 68,  51, 179},
                        { 31,  40, 123}, { 13,  29,  77}, {  4,  17,  37},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {203,  41, 203}, {127,  56, 174}, { 49,  56, 127},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {110, 121, 217}, {119, 113, 213}, { 64,  95, 185},
                        { 30,  72, 144}, {  8,  42,  76}, {  2,  17,  25},
                    },
                    {  // band 2
                        {127, 159, 229}, {115, 134, 223}, { 36, 100, 189},
                        { 11,  75, 142}, {  3,  48,  83}, {  1,  19,  33},
                    },
                    {  // band 3
                        {150, 172, 241}, { 90, 133, 231}, { 28, 102, 192},
                        {  7,  81, 147}, {  1,  53,  91}, {  1,  25,  42},
                    },
                    {  // band 4
                        {184, 144, 248}, {114, 117, 237}, { 37,  89, 192},
                        { 10,  63, 130}, {  4,  42,  76}, {  1,  19,  38},
                    },
                    {  // band 5
                        {207,  79, 250}, {179,  74, 241}, { 83,  67, 199},
                        { 38,  51, 142}, { 17,  37,  97}, { 10,  14,  55},
                    },
                },
                {  // Inter
                    {  // band 0
                        {220,  82, 232}, {150,  93, 214}, { 66,  95, 177},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 160, 227}, {136, 141, 227}, { 67, 114, 190},
                        { 40,  94, 148}, { 21,  70, 107}, { 10,  43,  63},
                    },
                    {  // band 2
                        {124, 173, 235}, {105, 147, 226}, { 27, 107, 184},
                        { 10,  80, 142}, {  3,  50,  86}, {  1,  16,  32},
                    },
                    {  // band 3
                        {149, 179, 243}, { 89, 147, 234}, { 29, 112, 193},
                        {  9,  94, 157}, {  1,  64, 111}, {  1,  25,  43},
                    },
                    {  // band 4
                        {187, 153, 248}, {127, 130, 241}, { 52,  99, 202},
                        { 20,  79, 152}, {  4,  50,  93}, {  1,  19,  32},
                    },
                    {  // band 5
                        {215,  82, 251}, {195,  80, 246}, { 93,  70, 204},
                        { 39,  54, 147}, { 14,  33,  88}, {  6,  14,  39},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {182,  34, 137}, { 79,  39, 103}, { 10,  28,  51},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 45,  88, 147}, { 46,  80, 140}, { 25,  69, 119},
                        { 12,  57,  96}, {  4,  41,  65}, {  1,  20,  31},
                    },
                    {  // band 2
                        { 58, 124, 190}, { 39, 106, 178}, { 16,  86, 147},
                        {  7,  69, 114}, {  3,  50,  80}, {  1,  25,  42},
                    },
                    {  // band 3
                        { 90, 138, 215}, { 54, 116, 198}, { 18,  86, 155},
                        {  5,  62, 112}, {  1,  38,  68}, {  1,  17,  30},
                    },
                    {  // band 4
                        {126, 149, 231}, { 82, 114, 211}, { 21,  80, 157},
                        {  6,  56, 105}, {  1,  36,  64}, {  1,  17,  31},
                    },
                    {  // band 5
                        {171,  56, 236}, {140,  54, 219}, { 57,  45, 167},
                        { 26,  36, 113}, { 11,  29,  72}, {  3,  18,  39},
                    },
                },
                {  // Intra
                    {  // band 0
                        {153, 122, 186}, {106, 109, 171}, { 36,  84, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 27, 151, 201}, { 34, 131, 199}, { 23, 102, 161},
                        { 10,  80, 120}, {  4,  52,  78}, {  1,  24,  37},
                    },
                    {  // band 2
                        { 43, 158, 213}, { 35, 133, 203}, {  8,  92, 151},
                        {  2,  64, 106}, {  1,  36,  60}, {  1,  13,  24},
                    },
                    {  // band 3
                        { 68, 167, 223}, { 36, 135, 211}, {  9,  94, 157},
                        {  2,  67, 112}, {  1,  40,  68}, {  1,  17,  31},
                    },
                    {  // band 4
                        {131, 146, 237}, { 72, 119, 223}, { 17,  82, 164},
                        {  4,  55, 107}, {  1,  34,  63}, {  1,  16,  29},
                    },
                    {  // band 5
                        {184,  68, 244}, {153,  59, 232}, { 68,  51, 179},
                        { 31,  40, 123}, { 13,  29,  77}, {  4,  17,  37},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {203,  41, 203}, {127,  56, 174}, { 49,  56, 127},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {110, 121, 217}, {119, 113, 213}, { 64,  95, 185},
                        { 30,  72, 144}, {  8,  42,  76}, {  2,  17,  25},
                    },
                    {  // band 2
                        {127, 159, 229}, {115, 134, 223}, { 36, 100, 189},
                        { 11,  75, 142}, {  3,  48,  83}, {  1,  19,  33},
                    },
                    {  // band 3
                        {150, 172, 241}, { 90, 133, 231}, { 28, 102, 192},
                        {  7,  81, 147}, {  1,  53,  91}, {  1,  25,  42},
                    },
                    {  // band 4
                        {184, 144, 248}, {114, 117, 237}, { 37,  89, 192},
                        { 10,  63, 130}, {  4,  42,  76}, {  1,  19,  38},
                    },
                    {  // band 5
                        {207,  79, 250}, {179,  74, 241}, { 83,  67, 199},
                        { 38,  51, 142}, { 17,  37,  97}, { 10,  14,  55},
                    },
                },
                {  // Inter
                    {  // band 0
                        {220,  82, 232}, {150,  93, 214}, { 66,  95, 177},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 160, 227}, {136, 141, 227}, { 67, 114, 190},
                        { 40,  94, 148}, { 21,  70, 107}, { 10,  43,  63},
                    },
                    {  // band 2
                        {124, 173, 235}, {105, 147, 226}, { 27, 107, 184},
                        { 10,  80, 142}, {  3,  50,  86}, {  1,  16,  32},
                    },
                    {  // band 3
                        {149, 179, 243}, { 89, 147, 234}, { 29, 112, 193},
                        {  9,  94, 157}, {  1,  64, 111}, {  1,  25,  43},
                    },
                    {  // band 4
                        {187, 153, 248}, {127, 130, 241}, { 52,  99, 202},
                        { 20,  79, 152}, {  4,  50,  93}, {  1,  19,  32},
                    },
                    {  // band 5
                        {215,  82, 251}, {195,  80, 246}, { 93,  70, 204},
                        { 39,  54, 147}, { 14,  33,  88}, {  6,  14,  39},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {116,  43, 131}, { 39,  41,  94}, {  4,  28,  47},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 28, 101, 141}, { 27,  95, 140}, { 18,  80, 121},
                        { 10,  61,  95}, {  4,  39,  60}, {  1,  19,  26},
                    },
                    {  // band 2
                        { 29, 150, 183}, { 19, 127, 175}, {  8,  98, 147},
                        {  3,  76, 115}, {  1,  55,  84}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 26, 168, 202}, { 12, 138, 188}, {  2,  98, 149},
                        {  1,  69, 110}, {  1,  40,  65}, {  1,  17,  25},
                    },
                    {  // band 4
                        { 33, 188, 225}, { 12, 155, 207}, {  2, 101, 155},
                        {  1,  65, 106}, {  1,  36,  60}, {  1,  18,  26},
                    },
                    {  // band 5
                        { 79, 205, 242}, { 30, 168, 224}, {  5, 106, 164},
                        {  1,  68, 110}, {  1,  39,  65}, {  1,  18,  28},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 96,  80, 201}, { 51,  88, 168}, { 14,  78, 116},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  6, 167, 216}, { 32, 152, 211}, { 24, 121, 182},
                        { 13,  98, 149}, { 12,  76, 108}, {  8,  48,  62},
                    },
                    {  // band 2
                        { 17, 176, 225}, { 13, 147, 209}, {  3,  96, 155},
                        {  1,  65, 108}, {  2,  43,  63}, {  2,  23,  25},
                    },
                    {  // band 3
                        { 18, 183, 232}, { 10, 153, 214}, {  1,  96, 154},
                        {  1,  63, 105}, {  1,  39,  59}, {  1,  21,  24},
                    },
                    {  // band 4
                        { 23, 191, 239}, {  8, 159, 221}, {  1,  97, 158},
                        {  1,  61, 105}, {  1,  37,  60}, {  1,  20,  26},
                    },
                    {  // band 5
                        { 70, 201, 243}, { 29, 163, 228}, {  4, 102, 169},
                        {  1,  67, 114}, {  1,  39,  66}, {  1,  17,  29},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {181,  38, 192}, { 95,  47, 151}, { 29,  49, 102},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 72, 131, 202}, { 93, 120, 205}, { 50, 103, 179},
                        { 24,  79, 143}, { 11,  47,  78}, {  7,  19,  25},
                    },
                    {  // band 2
                        { 84, 176, 221}, { 56, 144, 214}, { 21, 108, 182},
                        {  8,  83, 139}, {  3,  55,  90}, {  2,  27,  41},
                    },
                    {  // band 3
                        { 84, 195, 234}, { 42, 156, 222}, { 10, 109, 180},
                        {  4,  77, 133}, {  1,  48,  80}, {  1,  23,  35},
                    },
                    {  // band 4
                        { 89, 210, 238}, { 35, 165, 221}, {  6, 106, 172},
                        {  2,  70, 123}, {  1,  44,  74}, {  1,  21,  30},
                    },
                    {  // band 5
                        {114, 221, 247}, { 49, 170, 234}, {  7, 113, 184},
                        {  2,  77, 132}, {  1,  48,  79}, {  1,  25,  33},
                    },
                },
                {  // Inter
                    {  // band 0
                        {192,  66, 237}, {113,  84, 211}, { 35,  84, 154},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 81, 180, 234}, {127, 165, 229}, { 58, 137, 204},
                        { 41, 114, 174}, { 44,  94, 136}, { 29,  66,  86},
                    },
                    {  // band 2
                        { 82, 193, 240}, { 39, 162, 223}, {  8, 113, 179},
                        {  3,  83, 136}, {  6,  62,  84}, {  5,  45,  45},
                    },
                    {  // band 3
                        { 78, 203, 242}, { 31, 170, 227}, {  4, 115, 181},
                        {  1,  82, 135}, {  2,  59,  82}, {  1,  45,  47},
                    },
                    {  // band 4
                        { 76, 210, 239}, { 25, 170, 213}, {  2,  99, 152},
                        {  1,  69, 115}, {  1,  49,  80}, {  1,  47,  57},
                    },
                    {  // band 5
                        {103, 217, 250}, { 42, 180, 237}, {  3, 124, 191},
                        {  1,  90, 150}, {  1,  69, 116}, {  1,  52,  46},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 58,  38,  99}, {  9,  26,  51}, {  1,  14,  22},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 14,  78, 109}, { 16,  73, 105}, { 11,  62,  92},
                        {  6,  47,  72}, {  2,  29,  45}, {  1,  12,  18},
                    },
                    {  // band 2
                        { 17, 131, 148}, { 11, 112, 140}, {  5,  87, 118},
                        {  2,  63,  90}, {  1,  42,  63}, {  1,  19,  31},
                    },
                    {  // band 3
                        { 12, 151, 168}, {  6, 116, 152}, {  1,  76, 115},
                        {  1,  50,  81}, {  1,  32,  52}, {  1,  14,  23},
                    },
                    {  // band 4
                        { 10, 174, 191}, {  3, 130, 172}, {  1,  80, 126},
                        {  1,  53,  88}, {  1,  32,  55}, {  1,  14,  24},
                    },
                    {  // band 5
                        { 19, 219, 237}, {  3, 168, 211}, {  1,  90, 142},
                        {  1,  53,  91}, {  1,  29,  51}, {  1,  12,  21},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 21,  46, 184}, { 10,  53, 130}, {  2,  49,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  3, 169, 198}, { 37, 165, 196}, { 26, 134, 176},
                        { 11, 108, 149}, {  5,  81, 112}, {  3,  47,  64},
                    },
                    {  // band 2
                        { 11, 183, 215}, {  8, 142, 192}, {  2,  91, 141},
                        {  1,  62, 100}, {  1,  38,  62}, {  1,  17,  28},
                    },
                    {  // band 3
                        { 12, 190, 223}, {  6, 149, 199}, {  1,  88, 139},
                        {  1,  56,  93}, {  1,  31,  54}, {  1,  13,  21},
                    },
                    {  // band 4
                        { 11, 197, 230}, {  3, 154, 204}, {  1,  83, 134},
                        {  1,  50,  86}, {  1,  28,  49}, {  1,  12,  21},
                    },
                    {  // band 5
                        { 17, 211, 240}, {  2, 167, 217}, {  1,  88, 143},
                        {  1,  53,  91}, {  1,  30,  53}, {  1,  14,  24},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {151,  30, 151}, { 50,  36, 105}, {  8,  34,  66},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 39, 111, 160}, { 62, 111, 165}, { 37,  99, 147},
                        { 15,  77, 118}, {  3,  47,  73}, {  1,  17,  27},
                    },
                    {  // band 2
                        { 48, 170, 190}, { 32, 135, 180}, { 11, 100, 149},
                        {  4,  76, 116}, {  1,  51,  80}, {  1,  22,  36},
                    },
                    {  // band 3
                        { 39, 191, 208}, { 18, 141, 191}, {  3,  96, 150},
                        {  1,  66, 110}, {  1,  41,  69}, {  1,  17,  28},
                    },
                    {  // band 4
                        { 32, 209, 219}, {  8, 152, 201}, {  1,  96, 153},
                        {  1,  63, 106}, {  1,  38,  66}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 33, 230, 237}, {  5, 173, 214}, {  1, 100, 155},
                        {  1,  62, 105}, {  1,  38,  66}, {  1,  18,  32},
                    },
                },
                {  // Inter
                    {  // band 0
                        {149,  38, 231}, { 59,  51, 186}, { 12,  54, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53, 179, 226}, {126, 176, 223}, { 58, 147, 202},
                        { 28, 118, 174}, { 15,  94, 138}, { 14,  63,  87},
                    },
                    {  // band 2
                        { 58, 196, 232}, { 26, 158, 213}, {  5, 106, 166},
                        {  1,  75, 124}, {  1,  46,  79}, {  1,  23,  39},
                    },
                    {  // band 3
                        { 46, 203, 235}, { 17, 162, 213}, {  2, 104, 165},
                        {  1,  72, 120}, {  1,  44,  74}, {  1,  22,  33},
                    },
                    {  // band 4
                        { 37, 213, 238}, {  8, 167, 216}, {  1, 104, 168},
                        {  1,  68, 119}, {  1,  40,  67}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 30, 228, 239}, {  4, 181, 213}, {  1, 103, 153},
                        {  1,  65, 110}, {  1,  43,  79}, {  1,  27,  56},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 1
#if CONFIG_CHROMA_2X2
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {174,  30, 159}, { 76,  38, 115}, { 15,  33,  65},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 60,  80, 153}, { 72,  75, 147}, { 36,  68, 129},
                        { 15,  59, 104}, {  4,  45,  74}, {  1,  28,  45},
                    },
                    {  // band 2
                        { 70, 122, 186}, { 55, 104, 175}, { 21,  83, 144},
                        {  8,  67, 112}, {  2,  51,  82}, {  1,  34,  57},
                    },
                    {  // band 3
                        { 97, 144, 207}, { 52, 109, 195}, { 16,  77, 153},
                        {  4,  58, 113}, {  1,  43,  77}, {  1,  27,  48},
                    },
                    {  // band 4
                        {128, 148, 229}, { 76, 104, 210}, { 18,  77, 159},
                        {  4,  65, 110}, {  1,  52,  82}, {  1,  31,  55},
                    },
                    {  // band 5
                        {165,  51, 238}, {128,  50, 230}, { 57,  49, 185},
                        { 28,  47, 130}, { 12,  44,  96}, {  3,  36,  60},
                    },
                },
                {  // Intra
                    {  // band 0
                        {169, 103, 203}, {117,  96, 176}, { 56,  81, 137},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 31, 150, 224}, { 49, 128, 212}, { 19,  92, 165},
                        {  6,  67, 116}, {  2,  43,  71}, {  1,  21,  36},
                    },
                    {  // band 2
                        { 58, 156, 230}, { 47, 130, 215}, {  7,  87, 158},
                        {  2,  63, 114}, {  1,  39,  71}, {  1,  18,  36},
                    },
                    {  // band 3
                        { 85, 176, 240}, { 43, 138, 226}, {  8,  93, 172},
                        {  2,  70, 127}, {  1,  46,  81}, {  1,  26,  47},
                    },
                    {  // band 4
                        {155, 144, 248}, { 93, 116, 235}, { 21,  83, 180},
                        {  4,  59, 119}, {  1,  43,  80}, {  1,  25,  50},
                    },
                    {  // band 5
                        {203,  61, 250}, {171,  57, 243}, { 71,  57, 199},
                        { 31,  49, 144}, { 13,  42,  96}, {  7,  30,  52},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {204,  44, 204}, {137,  57, 184}, { 72,  62, 152},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {145, 117, 236}, {151, 112, 231}, { 87,  95, 208},
                        { 31,  77, 165}, {  5,  49,  98}, {  1,  24,  39},
                    },
                    {  // band 2
                        {146, 152, 241}, {140, 132, 236}, { 41, 103, 209},
                        { 10,  86, 165}, {  2,  55, 106}, {  1,  25,  58},
                    },
                    {  // band 3
                        {154, 181, 249}, { 84, 143, 240}, { 23, 114, 210},
                        {  6, 102, 182}, {  2,  71, 137}, {  1,  35,  90},
                    },
                    {  // band 4
                        {184, 150, 251}, {115, 130, 244}, { 34, 105, 215},
                        { 15,  89, 173}, {  1,  51, 141}, {128, 128, 128},
                    },
                    {  // band 5
                        {211,  71, 253}, {193,  78, 249}, {106,  91, 232},
                        { 61,  87, 198}, { 85, 153, 254}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {232, 104, 242}, {165, 114, 227}, { 96, 120, 206},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {137, 178, 250}, {146, 153, 245}, { 74, 108, 205},
                        { 41,  81, 149}, { 24,  55, 104}, { 13,  36,  68},
                    },
                    {  // band 2
                        {147, 185, 252}, {127, 161, 246}, { 30, 104, 208},
                        { 11,  74, 154}, {  6,  54, 100}, {  2,  29,  63},
                    },
                    {  // band 3
                        {163, 191, 254}, {101, 161, 249}, { 22, 114, 215},
                        {  6,  89, 173}, {  1,  65, 120}, {  1,   1, 170},
                    },
                    {  // band 4
                        {197, 160, 254}, {142, 141, 251}, { 39, 102, 218},
                        { 10,  76, 158}, {  1,  56, 122}, {128, 128, 128},
                    },
                    {  // band 5
                        {224,  76, 254}, {215,  84, 253}, {107,  85, 232},
                        { 43,  71, 177}, {  1,   1, 254}, {128, 128, 128},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {174,  30, 159}, { 76,  38, 115}, { 15,  33,  65},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 60,  80, 153}, { 72,  75, 147}, { 36,  68, 129},
                        { 15,  59, 104}, {  4,  45,  74}, {  1,  28,  45},
                    },
                    {  // band 2
                        { 70, 122, 186}, { 55, 104, 175}, { 21,  83, 144},
                        {  8,  67, 112}, {  2,  51,  82}, {  1,  34,  57},
                    },
                    {  // band 3
                        { 97, 144, 207}, { 52, 109, 195}, { 16,  77, 153},
                        {  4,  58, 113}, {  1,  43,  77}, {  1,  27,  48},
                    },
                    {  // band 4
                        {128, 148, 229}, { 76, 104, 210}, { 18,  77, 159},
                        {  4,  65, 110}, {  1,  52,  82}, {  1,  31,  55},
                    },
                    {  // band 5
                        {165,  51, 238}, {128,  50, 230}, { 57,  49, 185},
                        { 28,  47, 130}, { 12,  44,  96}, {  3,  36,  60},
                    },
                },
                {  // Intra
                    {  // band 0
                        {169, 103, 203}, {117,  96, 176}, { 56,  81, 137},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 31, 150, 224}, { 49, 128, 212}, { 19,  92, 165},
                        {  6,  67, 116}, {  2,  43,  71}, {  1,  21,  36},
                    },
                    {  // band 2
                        { 58, 156, 230}, { 47, 130, 215}, {  7,  87, 158},
                        {  2,  63, 114}, {  1,  39,  71}, {  1,  18,  36},
                    },
                    {  // band 3
                        { 85, 176, 240}, { 43, 138, 226}, {  8,  93, 172},
                        {  2,  70, 127}, {  1,  46,  81}, {  1,  26,  47},
                    },
                    {  // band 4
                        {155, 144, 248}, { 93, 116, 235}, { 21,  83, 180},
                        {  4,  59, 119}, {  1,  43,  80}, {  1,  25,  50},
                    },
                    {  // band 5
                        {203,  61, 250}, {171,  57, 243}, { 71,  57, 199},
                        { 31,  49, 144}, { 13,  42,  96}, {  7,  30,  52},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {204,  44, 204}, {137,  57, 184}, { 72,  62, 152},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {145, 117, 236}, {151, 112, 231}, { 87,  95, 208},
                        { 31,  77, 165}, {  5,  49,  98}, {  1,  24,  39},
                    },
                    {  // band 2
                        {146, 152, 241}, {140, 132, 236}, { 41, 103, 209},
                        { 10,  86, 165}, {  2,  55, 106}, {  1,  25,  58},
                    },
                    {  // band 3
                        {154, 181, 249}, { 84, 143, 240}, { 23, 114, 210},
                        {  6, 102, 182}, {  2,  71, 137}, {  1,  35,  90},
                    },
                    {  // band 4
                        {184, 150, 251}, {115, 130, 244}, { 34, 105, 215},
                        { 15,  89, 173}, {  1,  51, 141}, {128, 128, 128},
                    },
                    {  // band 5
                        {211,  71, 253}, {193,  78, 249}, {106,  91, 232},
                        { 61,  87, 198}, { 85, 153, 254}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {232, 104, 242}, {165, 114, 227}, { 96, 120, 206},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {137, 178, 250}, {146, 153, 245}, { 74, 108, 205},
                        { 41,  81, 149}, { 24,  55, 104}, { 13,  36,  68},
                    },
                    {  // band 2
                        {147, 185, 252}, {127, 161, 246}, { 30, 104, 208},
                        { 11,  74, 154}, {  6,  54, 100}, {  2,  29,  63},
                    },
                    {  // band 3
                        {163, 191, 254}, {101, 161, 249}, { 22, 114, 215},
                        {  6,  89, 173}, {  1,  65, 120}, {  1,   1, 170},
                    },
                    {  // band 4
                        {197, 160, 254}, {142, 141, 251}, { 39, 102, 218},
                        { 10,  76, 158}, {  1,  56, 122}, {128, 128, 128},
                    },
                    {  // band 5
                        {224,  76, 254}, {215,  84, 253}, {107,  85, 232},
                        { 43,  71, 177}, {  1,   1, 254}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 68,  37, 120}, { 21,  34,  82}, {  5,  26,  49},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41,  89, 138}, { 56,  83, 132}, { 31,  73, 115},
                        { 16,  62,  92}, {  5,  45,  62}, {  1,  24,  32},
                    },
                    {  // band 2
                        { 48, 139, 165}, { 30, 114, 160}, { 13,  92, 132},
                        {  6,  72, 103}, {  3,  49,  72}, {  1,  26,  41},
                    },
                    {  // band 3
                        { 44, 162, 191}, { 20, 127, 175}, {  5,  90, 137},
                        {  1,  62, 100}, {  1,  38,  63}, {  1,  20,  32},
                    },
                    {  // band 4
                        { 51, 184, 213}, { 16, 137, 193}, {  2,  89, 143},
                        {  1,  60, 102}, {  1,  39,  66}, {  1,  23,  37},
                    },
                    {  // band 5
                        { 76, 200, 235}, { 27, 150, 216}, {  3,  99, 164},
                        {  1,  70, 119}, {  1,  45,  77}, {  1,  22,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 81, 112, 199}, { 49, 101, 164}, { 19,  80, 119},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 12, 181, 217}, { 48, 151, 212}, { 38, 118, 180},
                        { 22,  95, 140}, { 11,  67,  92}, { 13,  46,  44},
                    },
                    {  // band 2
                        { 29, 188, 226}, { 19, 147, 210}, {  5,  95, 154},
                        {  4,  68, 106}, {  3,  44,  60}, {  1,  24,  27},
                    },
                    {  // band 3
                        { 30, 195, 234}, { 15, 153, 216}, {  3,  95, 156},
                        {  2,  66, 108}, {  2,  44,  62}, {  1,  24,  29},
                    },
                    {  // band 4
                        { 36, 203, 243}, { 12, 162, 225}, {  2,  98, 163},
                        {  2,  67, 113}, {  2,  45,  68}, {  1,  24,  34},
                    },
                    {  // band 5
                        { 86, 207, 248}, { 35, 165, 236}, {  3, 107, 180},
                        {  1,  73, 128}, {  1,  45,  78}, {  1,  20,  34},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  37, 205}, {118,  51, 172}, { 56,  57, 135},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 135, 225}, {144, 123, 221}, { 72, 103, 197},
                        { 35,  77, 153}, { 15,  47,  82}, {  6,  25,  34},
                    },
                    {  // band 2
                        {128, 171, 233}, { 82, 142, 226}, { 31, 106, 191},
                        { 16,  82, 146}, {  9,  59,  98}, {  4,  33,  54},
                    },
                    {  // band 3
                        {126, 197, 241}, { 66, 155, 230}, { 18, 108, 190},
                        {  7,  82, 148}, {  3,  58,  98}, {  1,  25,  50},
                    },
                    {  // band 4
                        {117, 207, 244}, { 44, 163, 233}, {  9, 112, 191},
                        {  5,  84, 148}, {  3,  61,  87}, {  1,  28,  38},
                    },
                    {  // band 5
                        {112, 214, 249}, { 39, 174, 240}, {  6, 125, 205},
                        {  4,  96, 163}, {  5,  66, 100}, {  1, 128, 254},
                    },
                },
                {  // Inter
                    {  // band 0
                        {227,  70, 234}, {145,  91, 213}, { 61, 100, 173},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {108, 198, 243}, {171, 172, 240}, {118, 130, 210},
                        {104, 107, 165}, { 64,  85, 114}, { 55,  64,  60},
                    },
                    {  // band 2
                        {110, 208, 247}, { 64, 175, 237}, { 24, 112, 187},
                        { 24,  81, 133}, { 24,  63,  83}, { 21,  47,  53},
                    },
                    {  // band 3
                        { 91, 218, 249}, { 46, 188, 238}, {  8, 113, 184},
                        {  5,  83, 137}, {  6,  62,  95}, { 17,  44,  94},
                    },
                    {  // band 4
                        { 84, 216, 248}, { 30, 187, 237}, {  2, 117, 188},
                        {  1,  88, 141}, {  3,  63,  98}, {  1,   1,   1},
                    },
                    {  // band 5
                        {116, 218, 252}, { 47, 186, 242}, {  2, 132, 204},
                        {  1, 106, 175}, {  1,  88, 104}, {  1, 254, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 35,  41, 129}, { 12,  30,  70}, {  2,  19,  32},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 30,  77, 116}, { 39,  70, 110}, { 20,  58,  96},
                        {  8,  47,  77}, {  2,  33,  52}, {  1,  17,  26},
                    },
                    {  // band 2
                        { 31, 123, 146}, { 18, 103, 140}, {  7,  81, 119},
                        {  2,  62,  95}, {  1,  44,  70}, {  1,  26,  42},
                    },
                    {  // band 3
                        { 21, 149, 170}, {  9, 114, 158}, {  2,  80, 126},
                        {  1,  57,  94}, {  1,  36,  61}, {  1,  18,  31},
                    },
                    {  // band 4
                        { 20, 178, 199}, {  6, 134, 183}, {  1,  87, 139},
                        {  1,  60, 100}, {  1,  37,  64}, {  1,  18,  31},
                    },
                    {  // band 5
                        { 36, 218, 233}, {  6, 160, 207}, {  1,  92, 147},
                        {  1,  59, 101}, {  1,  35,  62}, {  1,  18,  31},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 17,  62, 211}, { 14,  62, 153}, {  5,  50,  84},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 11, 180, 205}, { 87, 160, 205}, { 53, 128, 184},
                        { 27, 106, 156}, { 13,  79, 115}, {  6,  46,  67},
                    },
                    {  // band 2
                        { 32, 194, 220}, { 20, 145, 202}, {  4,  96, 152},
                        {  1,  67, 111}, {  1,  42,  70}, {  1,  21,  37},
                    },
                    {  // band 3
                        { 30, 204, 228}, { 14, 152, 207}, {  1,  92, 149},
                        {  1,  61, 103}, {  1,  34,  59}, {  1,  16,  28},
                    },
                    {  // band 4
                        { 27, 213, 235}, {  7, 159, 210}, {  1,  88, 143},
                        {  1,  55,  94}, {  1,  31,  53}, {  1,  16,  27},
                    },
                    {  // band 5
                        { 28, 223, 243}, {  4, 173, 217}, {  1,  91, 146},
                        {  1,  58,  98}, {  1,  35,  60}, {  1,  19,  33},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {172,  37, 202}, { 83,  51, 156}, { 24,  53, 110},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 134, 206}, {110, 124, 200}, { 47, 106, 180},
                        { 15,  82, 145}, {  3,  48,  83}, {  1,  19,  32},
                    },
                    {  // band 2
                        { 80, 176, 220}, { 49, 145, 212}, { 17, 112, 180},
                        {  7,  84, 140}, {  1,  53,  89}, {  1,  27,  43},
                    },
                    {  // band 3
                        { 74, 201, 232}, { 38, 158, 221}, {  8, 112, 179},
                        {  2,  79, 132}, {  1,  47,  82}, {  1,  26,  42},
                    },
                    {  // band 4
                        { 73, 215, 239}, { 28, 169, 227}, {  3, 112, 176},
                        {  1,  74, 126}, {  1,  48,  79}, {  1,  27,  44},
                    },
                    {  // band 5
                        { 71, 233, 244}, { 18, 180, 230}, {  1, 114, 180},
                        {  1,  80, 134}, {  1,  51,  85}, {  1,  26,  36},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  34, 244}, {126,  57, 212}, { 46,  67, 151},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {120, 202, 245}, {198, 173, 241}, {119, 146, 224},
                        { 76, 126, 195}, { 44, 102, 159}, { 40,  76, 115},
                    },
                    {  // band 2
                        {120, 215, 248}, { 69, 171, 237}, { 23, 119, 194},
                        { 10,  86, 147}, {  2,  56,  94}, {  1,  25,  44},
                    },
                    {  // band 3
                        {102, 226, 250}, { 53, 183, 239}, {  9, 118, 188},
                        {  2,  78, 131}, {  1,  48,  89}, {  1,  17,  36},
                    },
                    {  // band 4
                        { 86, 235, 252}, { 34, 194, 240}, {  2, 109, 173},
                        {  1,  68, 118}, {  1,  44,  79}, {  1,   1,  38},
                    },
                    {  // band 5
                        { 59, 236, 243}, { 11, 189, 228}, {  1, 112, 187},
                        {  1,  88, 145}, {  1,  55,  92}, {  1,   1, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 2
#if CONFIG_CHROMA_2X2
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {181,  22, 175}, { 96,  37, 147}, { 35,  41, 105},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80,  95, 197}, {111,  92, 193}, { 59,  87, 175},
                        { 29,  79, 150}, { 10,  65, 118}, {  2,  47,  82},
                    },
                    {  // band 2
                        { 90, 141, 216}, { 77, 120, 210}, { 23,  95, 184},
                        { 11,  81, 151}, {  6,  75, 130}, {  2,  58, 113},
                    },
                    {  // band 3
                        {122, 167, 231}, { 66, 119, 225}, { 26,  87, 189},
                        {  7,  76, 151}, {  2,  63, 125}, {  1,  59,  77},
                    },
                    {  // band 4
                        {162, 147, 244}, {110,  97, 236}, { 32,  88, 204},
                        { 11,  89, 174}, {  5,  78, 151}, {128, 128, 128},
                    },
                    {  // band 5
                        {205,  59, 251}, {176,  68, 248}, { 90,  71, 223},
                        { 49,  72, 188}, { 17,  74, 203}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {188,  70, 207}, {140,  73, 189}, { 85,  73, 163},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 59, 144, 239}, { 79, 126, 237}, { 31, 102, 202},
                        { 10,  81, 153}, {  3,  56, 102}, {  2,  33,  59},
                    },
                    {  // band 2
                        {100, 152, 243}, { 80, 129, 236}, { 14,  94, 194},
                        {  4,  72, 150}, {  1,  50, 103}, {  1,  35,  60},
                    },
                    {  // band 3
                        {130, 183, 247}, { 70, 139, 242}, { 19, 100, 203},
                        {  4,  83, 159}, {  1,  59, 119}, {  1,  44,  72},
                    },
                    {  // band 4
                        {197, 138, 252}, {135, 107, 247}, { 31,  86, 210},
                        {  7,  74, 160}, {  1,  53, 107}, {128, 128, 128},
                    },
                    {  // band 5
                        {229,  54, 254}, {200,  51, 251}, { 83,  61, 226},
                        { 33,  55, 177}, { 12,  74, 145}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {229,  20, 235}, {183,  37, 221}, {127,  47, 198},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {188, 115, 251}, {208, 110, 250}, {101,  99, 235},
                        { 38,  81, 197}, {  9,  56, 132}, {  9,  52,  63},
                    },
                    {  // band 2
                        {189, 150, 252}, {186, 137, 251}, { 54, 107, 236},
                        { 14,  90, 195}, {  1,  89, 104}, {128, 128, 128},
                    },
                    {  // band 3
                        {209, 180, 254}, {142, 145, 253}, { 51, 130, 236},
                        {  6, 128, 214}, {  1, 128, 254}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 140, 254}, {194, 128, 254}, { 75, 119, 233},
                        {128,  23, 230}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {244,  59, 254}, {239,  81, 254}, {128,  85, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {246,  55, 247}, {197,  64, 235}, {141,  74, 218},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {178, 163, 254}, {192, 138, 252}, { 85, 103, 231},
                        { 49,  81, 179}, { 32,  54, 133}, { 12,  26,  98},
                    },
                    {  // band 2
                        {189, 173, 254}, {179, 150, 253}, { 60,  94, 237},
                        { 34,  81, 198}, { 20,  53, 187}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 191, 254}, {157, 160, 254}, { 57, 117, 240},
                        { 28, 105, 211}, {  1, 128,   1}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 146, 254}, {208, 133, 254}, { 66,  78, 233},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  49, 254}, {246,  63, 254}, { 85, 142, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {181,  22, 175}, { 96,  37, 147}, { 35,  41, 105},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80,  95, 197}, {111,  92, 193}, { 59,  87, 175},
                        { 29,  79, 150}, { 10,  65, 118}, {  2,  47,  82},
                    },
                    {  // band 2
                        { 90, 141, 216}, { 77, 120, 210}, { 23,  95, 184},
                        { 11,  81, 151}, {  6,  75, 130}, {  2,  58, 113},
                    },
                    {  // band 3
                        {122, 167, 231}, { 66, 119, 225}, { 26,  87, 189},
                        {  7,  76, 151}, {  2,  63, 125}, {  1,  59,  77},
                    },
                    {  // band 4
                        {162, 147, 244}, {110,  97, 236}, { 32,  88, 204},
                        { 11,  89, 174}, {  5,  78, 151}, {128, 128, 128},
                    },
                    {  // band 5
                        {205,  59, 251}, {176,  68, 248}, { 90,  71, 223},
                        { 49,  72, 188}, { 17,  74, 203}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {188,  70, 207}, {140,  73, 189}, { 85,  73, 163},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 59, 144, 239}, { 79, 126, 237}, { 31, 102, 202},
                        { 10,  81, 153}, {  3,  56, 102}, {  2,  33,  59},
                    },
                    {  // band 2
                        {100, 152, 243}, { 80, 129, 236}, { 14,  94, 194},
                        {  4,  72, 150}, {  1,  50, 103}, {  1,  35,  60},
                    },
                    {  // band 3
                        {130, 183, 247}, { 70, 139, 242}, { 19, 100, 203},
                        {  4,  83, 159}, {  1,  59, 119}, {  1,  44,  72},
                    },
                    {  // band 4
                        {197, 138, 252}, {135, 107, 247}, { 31,  86, 210},
                        {  7,  74, 160}, {  1,  53, 107}, {128, 128, 128},
                    },
                    {  // band 5
                        {229,  54, 254}, {200,  51, 251}, { 83,  61, 226},
                        { 33,  55, 177}, { 12,  74, 145}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {229,  20, 235}, {183,  37, 221}, {127,  47, 198},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {188, 115, 251}, {208, 110, 250}, {101,  99, 235},
                        { 38,  81, 197}, {  9,  56, 132}, {  9,  52,  63},
                    },
                    {  // band 2
                        {189, 150, 252}, {186, 137, 251}, { 54, 107, 236},
                        { 14,  90, 195}, {  1,  89, 104}, {128, 128, 128},
                    },
                    {  // band 3
                        {209, 180, 254}, {142, 145, 253}, { 51, 130, 236},
                        {  6, 128, 214}, {  1, 128, 254}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 140, 254}, {194, 128, 254}, { 75, 119, 233},
                        {128,  23, 230}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {244,  59, 254}, {239,  81, 254}, {128,  85, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {246,  55, 247}, {197,  64, 235}, {141,  74, 218},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {178, 163, 254}, {192, 138, 252}, { 85, 103, 231},
                        { 49,  81, 179}, { 32,  54, 133}, { 12,  26,  98},
                    },
                    {  // band 2
                        {189, 173, 254}, {179, 150, 253}, { 60,  94, 237},
                        { 34,  81, 198}, { 20,  53, 187}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 191, 254}, {157, 160, 254}, { 57, 117, 240},
                        { 28, 105, 211}, {  1, 128,   1}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 146, 254}, {208, 133, 254}, { 66,  78, 233},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  49, 254}, {246,  63, 254}, { 85, 142, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 45,  28, 124}, { 23,  35, 107}, { 10,  34,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53,  99, 177}, { 82,  96, 174}, { 46,  89, 158},
                        { 21,  76, 133}, {  6,  56,  94}, {  1,  33,  54},
                    },
                    {  // band 2
                        { 68, 147, 201}, { 42, 124, 195}, { 17,  98, 166},
                        {  7,  75, 131}, {  2,  53,  93}, {  1,  33,  59},
                    },
                    {  // band 3
                        { 65, 176, 217}, { 30, 137, 206}, {  6,  97, 167},
                        {  2,  70, 128}, {  1,  47,  88}, {  1,  29,  46},
                    },
                    {  // band 4
                        { 69, 195, 232}, { 24, 146, 218}, {  4, 100, 175},
                        {  2,  72, 134}, {  1,  51,  93}, {  1,  29,  52},
                    },
                    {  // band 5
                        { 96, 212, 246}, { 39, 158, 234}, {  6, 109, 192},
                        {  2,  77, 144}, {  1,  50,  95}, {  1,  20,  45},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 71,  80, 213}, { 53,  73, 181}, { 25,  66, 141},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 168, 231}, { 91, 150, 229}, { 49, 122, 202},
                        { 22,  97, 162}, { 10,  68, 108}, {  9,  48,  57},
                    },
                    {  // band 2
                        { 56, 178, 236}, { 32, 148, 225}, {  9,  99, 176},
                        {  4,  69, 127}, {  2,  44,  78}, {  1,  25,  41},
                    },
                    {  // band 3
                        { 57, 191, 242}, { 27, 155, 230}, {  5, 102, 180},
                        {  2,  71, 133}, {  1,  44,  78}, {  1,  27,  41},
                    },
                    {  // band 4
                        { 67, 201, 247}, { 24, 162, 237}, {  3, 106, 188},
                        {  3,  74, 137}, {  1,  46,  85}, {  1,  34,  48},
                    },
                    {  // band 5
                        {111, 210, 251}, { 47, 166, 244}, {  3, 113, 199},
                        {  2,  77, 146}, {  1,  48,  93}, {  1,  38,  22},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {206,  21, 221}, {150,  36, 195}, { 94,  44, 164},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {147, 128, 239}, {194, 122, 238}, { 95, 104, 220},
                        { 39,  81, 183}, { 13,  53, 111}, {  3,  24,  49},
                    },
                    {  // band 2
                        {164, 163, 244}, {106, 142, 239}, { 50, 112, 215},
                        { 26,  90, 177}, { 12,  67, 130}, {  1,   1,  64},
                    },
                    {  // band 3
                        {155, 193, 249}, { 88, 158, 244}, { 26, 124, 220},
                        { 10,  98, 173}, {  1,  77, 126}, {128, 128, 128},
                    },
                    {  // band 4
                        {141, 205, 252}, { 64, 174, 248}, { 17, 124, 221},
                        { 12,  92, 176}, {  1,  29, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {150, 217, 254}, { 74, 191, 252}, { 30, 144, 215},
                        {  1, 106, 137}, {128,   1, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {241,  37, 242}, {175,  48, 223}, { 99,  53, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {153, 183, 248}, {212, 156, 247}, {134, 124, 221},
                        { 88, 103, 184}, { 59,  86, 132}, { 29,  61,  67},
                    },
                    {  // band 2
                        {162, 199, 250}, {106, 167, 247}, { 56, 110, 207},
                        { 32,  85, 165}, { 16,  71, 130}, {  1,  93, 254},
                    },
                    {  // band 3
                        {143, 213, 252}, { 86, 187, 250}, { 23, 124, 220},
                        {  7,  95, 176}, {  1, 109, 102}, {128, 128, 128},
                    },
                    {  // band 4
                        {130, 219, 254}, { 70, 201, 253}, { 15, 128, 215},
                        {  1, 101, 201}, {  1,  64, 170}, {128, 128, 128},
                    },
                    {  // band 5
                        {155, 219, 254}, {105, 207, 254}, { 28, 155, 229},
                        {  1, 153, 191}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 18,  26, 117}, { 10,  29,  82}, {  3,  25,  52},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35,  88, 152}, { 62,  85, 150}, { 36,  77, 137},
                        { 16,  66, 116}, {  4,  47,  81}, {  1,  26,  44},
                    },
                    {  // band 2
                        { 55, 141, 182}, { 32, 119, 177}, { 12,  93, 154},
                        {  4,  71, 123}, {  1,  51,  89}, {  1,  32,  56},
                    },
                    {  // band 3
                        { 46, 171, 202}, { 21, 130, 191}, {  5,  91, 154},
                        {  1,  64, 115}, {  1,  42,  77}, {  1,  25,  41},
                    },
                    {  // band 4
                        { 43, 195, 219}, { 12, 142, 203}, {  1,  91, 156},
                        {  1,  63, 115}, {  1,  41,  77}, {  1,  22,  43},
                    },
                    {  // band 5
                        { 42, 221, 238}, {  8, 162, 219}, {  1,  98, 167},
                        {  1,  67, 123}, {  1,  43,  83}, {  1,  25,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 16,  51, 216}, { 20,  48, 168}, {  9,  44, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 34, 164, 226}, {124, 148, 226}, { 72, 127, 207},
                        { 36, 107, 175}, { 15,  81, 129}, {  6,  51,  79},
                    },
                    {  // band 2
                        { 61, 182, 234}, { 35, 148, 220}, {  9, 101, 178},
                        {  4,  71, 134}, {  1,  46,  90}, {  1,  24,  51},
                    },
                    {  // band 3
                        { 54, 198, 239}, { 25, 156, 224}, {  3,  98, 173},
                        {  1,  66, 124}, {  1,  41,  78}, {  1,  15,  37},
                    },
                    {  // band 4
                        { 48, 209, 242}, { 12, 162, 226}, {  1,  96, 169},
                        {  1,  63, 119}, {  1,  40,  78}, {  1,  18,  45},
                    },
                    {  // band 5
                        { 44, 223, 247}, {  6, 173, 232}, {  1, 105, 178},
                        {  1,  71, 131}, {  1,  44,  84}, {  1,  13,  46},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  26, 214}, {121,  42, 181}, { 66,  49, 149},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {136, 128, 233}, {172, 124, 230}, { 80, 106, 211},
                        { 27,  81, 174}, {  6,  49,  98}, {  8,  28,  49},
                    },
                    {  // band 2
                        {145, 166, 239}, { 92, 141, 229}, { 28, 108, 196},
                        {  8,  87, 154}, {  1,  58, 105}, {  1,  27,  59},
                    },
                    {  // band 3
                        {131, 193, 242}, { 66, 151, 231}, { 13, 112, 192},
                        {  2,  81, 152}, {  1,  66, 121}, {  1,  23,  64},
                    },
                    {  // band 4
                        {112, 211, 246}, { 41, 164, 235}, {  5, 117, 202},
                        {  1,  83, 162}, {  1,  64, 111}, {128, 128, 128},
                    },
                    {  // band 5
                        { 96, 230, 250}, { 28, 185, 243}, {  2, 132, 204},
                        {  1,  91, 166}, {  1,  85,  46}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {238,  23, 242}, {157,  29, 215}, { 73,  27, 162},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {165, 173, 250}, {222, 151, 247}, {152, 134, 235},
                        {114, 120, 210}, { 86, 109, 176}, { 53,  88, 145},
                    },
                    {  // band 2
                        {164, 194, 249}, {100, 158, 241}, { 35, 111, 212},
                        { 17,  85, 167}, {  1,  52, 112}, {  1,  73,   1},
                    },
                    {  // band 3
                        {151, 215, 252}, { 83, 172, 245}, { 16, 122, 208},
                        {  6, 101, 165}, {  1,  74, 113}, {  1,   1,   1},
                    },
                    {  // band 4
                        {138, 230, 253}, { 65, 184, 248}, {  8, 128, 212},
                        {  1, 111, 182}, {128,   1,   1}, {128, 128, 128},
                    },
                    {  // band 5
                        {123, 240, 253}, { 36, 201, 250}, {  3, 127, 211},
                        {  1,  68, 204}, {128,   1,   1}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 3
#if CONFIG_CHROMA_2X2
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {186,  16, 200}, {122,  31, 187}, { 78,  40, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {146, 119, 245}, {182, 115, 244}, {130, 113, 238},
                        { 88, 110, 225}, { 47, 103, 208}, {  5, 102, 188},
                    },
                    {  // band 2
                        {164, 157, 248}, {155, 141, 250}, { 71, 116, 243},
                        { 88, 129, 233}, { 50,  99, 228}, { 26, 148, 191},
                    },
                    {  // band 3
                        {200, 158, 253}, {177, 118, 252}, { 99, 113, 245},
                        { 77, 120, 210}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 104, 254}, {209,  82, 254}, {143, 112, 252},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {250,  36, 254}, {243,  55, 254}, {223, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {207,  37, 226}, {164,  46, 218}, {122,  58, 201},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {149, 154, 253}, {170, 137, 253}, { 94, 123, 247},
                        { 42, 113, 222}, { 16,  97, 174}, { 49,  98, 159},
                    },
                    {  // band 2
                        {177, 162, 253}, {165, 142, 252}, { 51, 108, 243},
                        { 18, 108, 213}, {  1,  98, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {211, 152, 254}, {184, 116, 254}, { 70, 110, 244},
                        {  8, 108, 237}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {236,  89, 254}, {210,  67, 254}, {112, 111, 248},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  26, 254}, {233,  35, 254}, {128,   1, 254},
                        {254, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {247,   2, 247}, {226,   8, 242}, {191,  14, 235},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {231,  94, 254}, {248,  91, 254}, {186,  89, 252},
                        {128,  92, 244}, { 79, 112, 254}, {128, 128, 128},
                    },
                    {  // band 2
                        {228, 145, 253}, {240, 130, 254}, {223, 105, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {245, 153, 253}, {240, 120, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 128, 254}, {204, 128, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {253,   7, 249}, {224,   9, 244}, {182,  13, 231},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {234, 109, 254}, {242, 104, 254}, {160,  98, 254},
                        {123,  85, 243}, { 82,  43, 217}, {128, 128, 128},
                    },
                    {  // band 2
                        {243, 137, 254}, {240, 118, 254}, {136,  53, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {251, 173, 254}, {229, 129, 250}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 119, 254}, {254, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {186,  16, 200}, {122,  31, 187}, { 78,  40, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {146, 119, 245}, {182, 115, 244}, {130, 113, 238},
                        { 88, 110, 225}, { 47, 103, 208}, {  5, 102, 188},
                    },
                    {  // band 2
                        {164, 157, 248}, {155, 141, 250}, { 71, 116, 243},
                        { 88, 129, 233}, { 50,  99, 228}, { 26, 148, 191},
                    },
                    {  // band 3
                        {200, 158, 253}, {177, 118, 252}, { 99, 113, 245},
                        { 77, 120, 210}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 104, 254}, {209,  82, 254}, {143, 112, 252},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {250,  36, 254}, {243,  55, 254}, {223, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {207,  37, 226}, {164,  46, 218}, {122,  58, 201},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {149, 154, 253}, {170, 137, 253}, { 94, 123, 247},
                        { 42, 113, 222}, { 16,  97, 174}, { 49,  98, 159},
                    },
                    {  // band 2
                        {177, 162, 253}, {165, 142, 252}, { 51, 108, 243},
                        { 18, 108, 213}, {  1,  98, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {211, 152, 254}, {184, 116, 254}, { 70, 110, 244},
                        {  8, 108, 237}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {236,  89, 254}, {210,  67, 254}, {112, 111, 248},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  26, 254}, {233,  35, 254}, {128,   1, 254},
                        {254, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {247,   2, 247}, {226,   8, 242}, {191,  14, 235},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {231,  94, 254}, {248,  91, 254}, {186,  89, 252},
                        {128,  92, 244}, { 79, 112, 254}, {128, 128, 128},
                    },
                    {  // band 2
                        {228, 145, 253}, {240, 130, 254}, {223, 105, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {245, 153, 253}, {240, 120, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 128, 254}, {204, 128, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {253,   7, 249}, {224,   9, 244}, {182,  13, 231},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {234, 109, 254}, {242, 104, 254}, {160,  98, 254},
                        {123,  85, 243}, { 82,  43, 217}, {128, 128, 128},
                    },
                    {  // band 2
                        {243, 137, 254}, {240, 118, 254}, {136,  53, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {251, 173, 254}, {229, 129, 250}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 119, 254}, {254, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 49,  26, 159}, { 36,  34, 150}, { 26,  38, 124},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 122, 226}, {143, 119, 225}, { 90, 113, 213},
                        { 46, 102, 193}, { 14,  84, 157}, {  3,  59, 107},
                    },
                    {  // band 2
                        {109, 164, 237}, { 74, 142, 233}, { 29, 112, 216},
                        { 14,  92, 184}, { 10,  80, 156}, {  1,  52, 137},
                    },
                    {  // band 3
                        {110, 191, 245}, { 59, 156, 240}, { 18, 121, 220},
                        {  8,  97, 184}, {  3,  84, 150}, {128, 128, 128},
                    },
                    {  // band 4
                        {115, 203, 250}, { 59, 167, 246}, { 16, 130, 226},
                        {  7,  97, 192}, {  1,  71,  99}, {128, 128, 128},
                    },
                    {  // band 5
                        {149, 218, 253}, { 93, 171, 251}, { 28, 125, 233},
                        { 28,  99, 192}, {128,  85,  85}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 97,  45, 229}, { 79,  52, 205}, { 46,  58, 171},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 180, 249}, {156, 165, 249}, { 73, 141, 237},
                        { 31, 116, 208}, { 13,  81, 153}, {  5,  42,  86},
                    },
                    {  // band 2
                        {113, 188, 251}, { 68, 161, 244}, { 16, 108, 216},
                        {  6,  81, 168}, {  2,  65, 118}, {128,   1,   1},
                    },
                    {  // band 3
                        {117, 201, 252}, { 62, 171, 248}, { 12, 119, 221},
                        {  5,  90, 182}, {  4,  66, 116}, {128, 128, 128},
                    },
                    {  // band 4
                        {128, 207, 253}, { 70, 176, 251}, { 11, 126, 228},
                        {  6,  89, 189}, {  1,  44, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {162, 218, 254}, {107, 170, 253}, { 22, 131, 238},
                        {  1,  77, 182}, {  1, 254, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {235,   5, 238}, {194,  14, 223}, {152,  22, 205},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {200, 121, 251}, {241, 115, 252}, {167, 108, 248},
                        { 93,  93, 233}, { 36,  66, 189}, {128, 128, 128},
                    },
                    {  // band 2
                        {220, 151, 253}, {176, 135, 252}, { 95, 124, 254},
                        { 64, 105, 217}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {225, 189, 254}, {175, 155, 254}, {102, 119, 254},
                        {  1,   1,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {218, 195, 254}, {125, 157, 253}, {128, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {221, 197, 254}, { 85, 210, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {250,   9, 246}, {204,  13, 234}, {144,  18, 211},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {213, 157, 253}, {243, 138, 253}, {170, 117, 250},
                        {109,  91, 233}, { 66,  77, 163}, { 64,  85, 254},
                    },
                    {  // band 2
                        {221, 169, 254}, {182, 141, 253}, {112, 120, 239},
                        { 85, 165, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {226, 192, 254}, {189, 174, 251}, {153, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {232, 192, 254}, {195, 187, 247}, {  1, 191, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {247, 185, 254}, {254,  93, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 14,  30, 136}, { 15,  33, 120}, { 10,  33,  90},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 92, 109, 209}, {113, 108, 207}, { 77, 102, 193},
                        { 39,  91, 171}, { 11,  70, 129}, {  2,  44,  77},
                    },
                    {  // band 2
                        { 99, 158, 223}, { 66, 135, 217}, { 23, 109, 194},
                        {  9,  85, 160}, {  3,  66, 124}, {  1,  51, 100},
                    },
                    {  // band 3
                        { 89, 189, 234}, { 46, 149, 225}, { 10, 110, 194},
                        {  2,  83, 156}, {  1,  57, 113}, {  1,  47,  73},
                    },
                    {  // band 4
                        { 78, 206, 242}, { 28, 161, 232}, {  3, 114, 200},
                        {  1,  86, 161}, {  1,  62, 118}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 72, 227, 250}, { 20, 182, 242}, {  3, 126, 210},
                        {  2,  91, 166}, {  1,  64, 126}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 23,  42, 227}, { 41,  43, 195}, { 25,  45, 146},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {100, 172, 245}, {165, 158, 246}, { 88, 137, 234},
                        { 44, 116, 203}, { 18,  85, 149}, {  7,  56,  92},
                    },
                    {  // band 2
                        {117, 188, 247}, { 70, 155, 239}, { 18, 105, 204},
                        {  7,  78, 158}, {  2,  50, 111}, {  1,  38,  77},
                    },
                    {  // band 3
                        {104, 207, 250}, { 54, 166, 241}, {  6, 110, 199},
                        {  1,  78, 155}, {  1,  45, 100}, {  1,   1,   1},
                    },
                    {  // band 4
                        { 87, 216, 251}, { 30, 177, 243}, {  1, 114, 203},
                        {  1,  85, 157}, {  1,  53, 108}, {128, 128, 128},
                    },
                    {  // band 5
                        { 80, 230, 253}, { 23, 193, 248}, {  1, 127, 215},
                        {  1,  94, 170}, {  1,  71,  59}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {222,   9, 234}, {161,  20, 210}, {113,  30, 185},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {195, 120, 248}, {231, 124, 247}, {148, 116, 238},
                        { 64,  98, 207}, { 20,  70, 147}, { 87,  68, 100},
                    },
                    {  // band 2
                        {186, 161, 250}, {124, 148, 245}, { 44, 123, 230},
                        { 23, 107, 205}, {  1,  80, 131}, {128, 128, 128},
                    },
                    {  // band 3
                        {172, 196, 252}, {110, 160, 248}, { 37, 134, 235},
                        { 23, 125, 200}, {128, 254, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {173, 209, 253}, {103, 175, 250}, {  1, 120, 240},
                        {  1, 146, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {184, 235, 254}, { 81, 186, 251}, {128, 109, 254},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {248,   8, 243}, {185,  11, 225}, {108,  11, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {208, 158, 254}, {244, 147, 252}, {195, 132, 248},
                        {161, 122, 224}, {129, 114, 188}, { 59, 119, 159},
                    },
                    {  // band 2
                        {202, 182, 253}, {143, 161, 251}, { 73, 115, 247},
                        {146, 175, 204}, {128,   1, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 204, 254}, {131, 174, 251}, { 18, 153, 207},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {192, 221, 254}, {114, 190, 254}, {128, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {166, 236, 254}, {119, 200, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
};
#else
static const av1_coeff_probs_model default_coef_probs_4x4[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {97, 27, 144}, {81, 38, 128}, {51, 42, 99}
      },
      { // Band 1
        {74, 113, 204}, {68, 101, 199}, {50, 87, 173},
        {31, 76, 133}, {13, 55, 86}, {3, 30, 39}
      },
      { // Band 2
        {83, 156, 222}, {74, 127, 215}, {46, 101, 179},
        {30, 80, 129}, {14, 57, 81}, {3, 27, 37}
      },
      { // Band 3
        {105, 164, 233}, {84, 128, 224}, {49, 92, 175},
        {28, 60, 114}, {12, 34, 53}, {20, 59, 98}
      },
      { // Band 4
        {131, 159, 243}, {98, 123, 228}, {40, 78, 151},
        {19, 46, 97}, {13, 47, 19}, {19, 16, 19}
      },
      { // Band 5
        {192, 71, 241}, {174, 70, 226}, {125, 46, 153},
        {108, 49, 116}, {82, 24, 46}, {60, 14, 30}
      }
    },
    { // Inter
      { // Band 0
        {111, 66, 218}, {107, 87, 211}, {93, 99, 207}
      },
      { // Band 1
        {107, 166, 250}, {107, 143, 247}, {73, 119, 221},
        {43, 91, 166}, {17, 74, 102}, {3, 70, 53}
      },
      { // Band 2
        {126, 177, 251}, {109, 148, 246}, {64, 99, 204},
        {42, 68, 140}, {28, 52, 84}, {20, 34, 1}
      },
      { // Band 3
        {143, 178, 252}, {114, 144, 245}, {46, 92, 188},
        {45, 65, 104}, {40, 44, 76}, {1, 1, 1}
      },
      { // Band 4
        {163, 159, 251}, {120, 131, 243}, {47, 81, 182},
        {32, 39, 128}, {33, 44, 56}, {1, 17, 34}
      },
      { // Band 5
        {209, 94, 251}, {190, 81, 241}, {139, 45, 147},
        {123, 35, 73}, {118, 1, 118}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {189, 37, 229}, {145, 68, 205}, {99, 74, 171}
      },
      { // Band 1
        {153, 139, 242}, {135, 125, 235}, {84, 100, 200},
        {49, 75, 162}, {9, 21, 84}, {3, 31, 69}
      },
      { // Band 2
        {165, 165, 244}, {128, 144, 240}, {68, 94, 204},
        {39, 72, 132}, {22, 44, 93}, {26, 73, 26}
      },
      { // Band 3
        {181, 174, 246}, {142, 132, 241}, {81, 96, 212},
        {41, 70, 166}, {9, 48, 92}, {1, 19, 38}
      },
      { // Band 4
        {197, 159, 251}, {168, 121, 245}, {107, 75, 218},
        {70, 43, 158}, {1, 128, 1}, {1, 18, 37}
      },
      { // Band 5
        {231, 79, 255}, {211, 74, 249}, {157, 104, 210},
        {128, 102, 213}, {12, 34, 96}, {2, 20, 47}
      }
    },
    { // Inter
      { // Band 0
        {220, 53, 252}, {191, 80, 248}, {154, 100, 245}
      },
      { // Band 1
        {205, 153, 255}, {182, 147, 254}, {110, 131, 231},
        {68, 114, 161}, {50, 114, 140}, {1, 33, 57}
      },
      { // Band 2
        {213, 171, 255}, {184, 163, 254}, {116, 104, 235},
        {79, 71, 207}, {1, 41, 79}, {1, 20, 39}
      },
      { // Band 3
        {223, 158, 255}, {203, 137, 255}, {111, 142, 244},
        {2, 255, 133}, {1, 44, 85}, {1, 22, 47}
      },
      { // Band 4
        {232, 148, 255}, {222, 123, 255}, {255, 128, 255},
        {3, 61, 124}, {1, 41, 84}, {1, 21, 52}
      },
      { // Band 5
        {248, 92, 255}, {248, 96, 255}, {69, 58, 184},
        {31, 44, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
static const av1_coeff_probs_model default_coef_probs_8x8[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {112, 31, 159}, {72, 37, 119}, {22, 35, 68}
      },
      { // Band 1
        {42, 109, 174}, {45, 99, 172}, {32, 84, 149},
        {18, 69, 119}, {6, 46, 76}, {1, 19, 31}
      },
      { // Band 2
        {40, 154, 202}, {35, 126, 191}, {19, 98, 160},
        {10, 75, 122}, {5, 53, 82}, {1, 23, 39}
      },
      { // Band 3
        {39, 176, 215}, {28, 135, 200}, {11, 93, 156},
        {5, 63, 109}, {1, 36, 64}, {1, 14, 26}
      },
      { // Band 4
        {41, 191, 230}, {25, 147, 212}, {9, 97, 160},
        {3, 65, 109}, {1, 33, 58}, {1, 14, 20}
      },
      { // Band 5
        {68, 203, 242}, {40, 159, 220}, {12, 97, 153},
        {5, 58, 97}, {1, 29, 55}, {1, 11, 18}
      }
    },
    { // Inter
      { // Band 0
        {99, 67, 221}, {86, 80, 204}, {60, 87, 184}
      },
      { // Band 1
        {73, 169, 246}, {79, 158, 242}, {50, 135, 220},
        {30, 113, 181}, {18, 76, 126}, {5, 54, 85}
      },
      { // Band 2
        {90, 184, 250}, {78, 162, 243}, {47, 118, 214},
        {35, 85, 171}, {32, 53, 115}, {20, 28, 76}
      },
      { // Band 3
        {109, 197, 252}, {89, 172, 247}, {52, 119, 217},
        {37, 80, 161}, {23, 44, 100}, {1, 18, 34}
      },
      { // Band 4
        {132, 202, 254}, {110, 175, 251}, {63, 128, 228},
        {37, 86, 168}, {64, 91, 102}, {1, 17, 34}
      },
      { // Band 5
        {126, 204, 253}, {100, 174, 250}, {50, 148, 237},
        {25, 90, 133}, {1, 64, 85}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {195, 35, 235}, {137, 63, 201}, {62, 70, 145}
      },
      { // Band 1
        {110, 158, 233}, {102, 143, 227}, {60, 120, 199},
        {30, 85, 156}, {9, 50, 90}, {1, 16, 33}
      },
      { // Band 2
        {102, 185, 233}, {71, 152, 224}, {29, 111, 187},
        {18, 74, 138}, {4, 56, 87}, {1, 18, 46}
      },
      { // Band 3
        {101, 205, 239}, {66, 161, 229}, {23, 109, 183},
        {9, 85, 135}, {5, 71, 142}, {1, 1, 102}
      },
      { // Band 4
        {109, 216, 243}, {69, 168, 233}, {23, 119, 191},
        {8, 137, 115}, {1, 54, 98}, {1, 1, 255}
      },
      { // Band 5
        {139, 224, 249}, {98, 176, 238}, {55, 129, 187},
        {25, 101, 131}, {26, 59, 154}, {2, 20, 47}
      }
    },
    { // Inter
      { // Band 0
        {220, 72, 254}, {176, 108, 251}, {114, 132, 247}
      },
      { // Band 1
        {161, 185, 255}, {141, 185, 254}, {131, 180, 249},
        {111, 164, 186}, {50, 98, 142}, {1, 128, 1}
      },
      { // Band 2
        {171, 195, 255}, {133, 184, 254}, {68, 140, 231},
        {102, 96, 205}, {1, 1, 128}, {1, 20, 39}
      },
      { // Band 3
        {180, 206, 255}, {148, 191, 254}, {83, 157, 241},
        {128, 171, 128}, {1, 44, 85}, {1, 22, 47}
      },
      { // Band 4
        {194, 214, 255}, {159, 188, 255}, {122, 148, 250},
        {3, 255, 124}, {1, 41, 84}, {1, 21, 52}
      },
      { // Band 5
        {231, 217, 255}, {209, 149, 255}, {205, 145, 205},
        {31, 44, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
static const av1_coeff_probs_model default_coef_probs_16x16[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {91, 31, 117}, {49, 31, 89}, {14, 25, 48}
      },
      { // Band 1
        {31, 97, 151}, {33, 89, 148}, {28, 76, 133},
        {17, 60, 106}, {7, 42, 72}, {1, 19, 32}
      },
      { // Band 2
        {28, 152, 182}, {28, 120, 174}, {15, 93, 146},
        {9, 72, 116}, {5, 47, 82}, {1, 21, 37}
      },
      { // Band 3
        {29, 174, 203}, {23, 127, 187}, {9, 89, 145},
        {2, 56, 100}, {1, 31, 56}, {1, 12, 25}
      },
      { // Band 4
        {28, 193, 220}, {17, 141, 197}, {4, 87, 142},
        {1, 54, 95}, {1, 31, 56}, {1, 12, 26}
      },
      { // Band 5
        {29, 221, 240}, {11, 167, 215}, {2, 93, 149},
        {1, 58, 100}, {1, 35, 61}, {1, 16, 28}
      }
    },
    { // Inter
      { // Band 0
        {108, 52, 214}, {84, 60, 186}, {45, 69, 161}
      },
      { // Band 1
        {43, 164, 236}, {57, 161, 233}, {38, 146, 214},
        {24, 120, 182}, {15, 80, 126}, {5, 28, 66}
      },
      { // Band 2
        {58, 187, 242}, {47, 163, 234}, {28, 118, 204},
        {26, 82, 165}, {21, 54, 112}, {4, 28, 55}
      },
      { // Band 3
        {65, 201, 248}, {51, 170, 239}, {22, 117, 204},
        {11, 81, 159}, {10, 43, 102}, {1, 1, 1}
      },
      { // Band 4
        {80, 206, 252}, {57, 179, 245}, {25, 129, 214},
        {16, 97, 170}, {6, 60, 130}, {1, 128, 1}
      },
      { // Band 5
        {97, 217, 253}, {68, 186, 250}, {26, 138, 216},
        {20, 105, 166}, {11, 78, 111}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {181, 37, 233}, {121, 55, 192}, {46, 52, 124}
      },
      { // Band 1
        {108, 157, 221}, {98, 140, 215}, {59, 124, 187},
        {34, 92, 158}, {9, 68, 112}, {1, 41, 70}
      },
      { // Band 2
        {80, 188, 223}, {46, 153, 204}, {25, 91, 173},
        {11, 73, 131}, {5, 43, 82}, {1, 17, 91}
      },
      { // Band 3
        {63, 209, 228}, {31, 157, 206}, {8, 104, 167},
        {3, 63, 122}, {1, 44, 87}, {1, 43, 51}
      },
      { // Band 4
        {52, 220, 234}, {22, 165, 216}, {4, 104, 163},
        {2, 62, 129}, {1, 33, 50}, {1, 26, 28}
      },
      { // Band 5
        {58, 238, 242}, {24, 183, 224}, {4, 109, 172},
        {2, 87, 141}, {1, 52, 79}, {1, 51, 64}
      }
    },
    { // Inter
      { // Band 0
        {224, 52, 250}, {188, 81, 239}, {138, 114, 228}
      },
      { // Band 1
        {131, 206, 255}, {128, 193, 254}, {119, 173, 247},
        {106, 127, 187}, {50, 100, 124}, {1, 96, 1}
      },
      { // Band 2
        {123, 214, 254}, {86, 194, 254}, {64, 119, 221},
        {43, 51, 128}, {1, 32, 110}, {1, 20, 39}
      },
      { // Band 3
        {115, 223, 255}, {78, 200, 254}, {75, 164, 203},
        {128, 85, 255}, {1, 44, 85}, {1, 22, 47}
      },
      { // Band 4
        {132, 226, 255}, {88, 207, 254}, {20, 140, 225},
        {3, 61, 124}, {1, 41, 84}, {1, 21, 52}
      },
      { // Band 5
        {180, 236, 255}, {138, 223, 254}, {73, 166, 238},
        {31, 255, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
static const av1_coeff_probs_model default_coef_probs_32x32[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {163, 26, 188}, {78, 29, 105}, {22, 22, 48}
      },
      { // Band 1
        {72, 93, 168}, {74, 91, 170}, {62, 72, 151},
        {37, 55, 112}, {10, 33, 63}, {1, 14, 23}
      },
      { // Band 2
        {41, 163, 182}, {36, 136, 177}, {20, 102, 153},
        {10, 76, 114}, {5, 45, 71}, {1, 17, 27}
      },
      { // Band 3
        {43, 202, 213}, {28, 142, 193}, {10, 90, 141},
        {2, 51, 93}, {1, 24, 48}, {1, 10, 19}
      },
      { // Band 4
        {46, 216, 220}, {26, 150, 199}, {7, 87, 136},
        {2, 49, 86}, {1, 28, 47}, {1, 12, 24}
      },
      { // Band 5
        {19, 241, 237}, {5, 172, 200}, {1, 82, 126},
        {1, 47, 79}, {1, 29, 47}, {1, 14, 25}
      }
    },
    { // Inter
      { // Band 0
        {185, 20, 226}, {151, 26, 187}, {109, 34, 144}
      },
      { // Band 1
        {56, 151, 227}, {76, 165, 232}, {62, 161, 222},
        {47, 139, 201}, {29, 106, 150}, {14, 61, 98}
      },
      { // Band 2
        {57, 200, 237}, {43, 164, 227}, {22, 106, 190},
        {14, 68, 140}, {10, 48, 90}, {1, 15, 40}
      },
      { // Band 3
        {46, 209, 238}, {28, 165, 225}, {7, 107, 180},
        {2, 69, 125}, {2, 36, 94}, {1, 1, 1}
      },
      { // Band 4
        {55, 225, 248}, {28, 181, 237}, {7, 117, 198},
        {6, 77, 144}, {3, 60, 90}, {1, 1, 1}
      },
      { // Band 5
        {63, 243, 251}, {27, 193, 242}, {4, 124, 200},
        {1, 58, 153}, {1, 59, 124}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {208, 28, 218}, {183, 32, 188}, {169, 21, 189}
      },
      { // Band 1
        {205, 124, 247}, {190, 96, 240}, {233, 89, 233},
        {177, 44, 212}, {59, 58, 59}, {32, 33, 38}
      },
      { // Band 2
        {194, 195, 250}, {179, 190, 226}, {32, 174, 128},
        {32, 85, 128}, {12, 64, 122}, {1, 85, 90}
      },
      { // Band 3
        {149, 232, 249}, {95, 159, 227}, {28, 91, 171},
        {28, 102, 114}, {1, 1, 73}, {1, 19, 38}
      },
      { // Band 4
        {154, 239, 246}, {138, 151, 235}, {1, 123, 138},
        {128, 183, 255}, {1, 128, 1}, {1, 18, 37}
      },
      { // Band 5
        {157, 255, 253}, {75, 171, 241}, {43, 102, 171},
        {30, 44, 136}, {12, 34, 96}, {2, 20, 47}
      }
    },
    { // Inter
      { // Band 0
        {249, 13, 248}, {238, 14, 220}, {225, 16, 174}
      },
      { // Band 1
        {190, 189, 254}, {169, 134, 253}, {124, 179, 248},
        {138, 131, 223}, {64, 133, 192}, {1, 85, 128}
      },
      { // Band 2
        {139, 212, 254}, {126, 177, 255}, {93, 39, 186},
        {1, 1, 171}, {1, 41, 79}, {1, 20, 39}
      },
      { // Band 3
        {153, 216, 255}, {165, 204, 255}, {1, 1, 255},
        {2, 73, 133}, {1, 1, 1}, {1, 22, 47}
      },
      { // Band 4
        {147, 226, 254}, {119, 196, 255}, {1, 128, 255},
        {1, 1, 171}, {1, 1, 1}, {1, 21, 52}
      },
      { // Band 5
        {168, 240, 255}, {95, 179, 255}, {1, 171, 1},
        {31, 44, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};

#if CONFIG_TX64X64
// FIXME. Optimize for EC_MULTISYMBOL
static const av1_coeff_probs_model default_coef_probs_64x64[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {  17,  38, 140 }, {   7,  34,  80 }, {   1,  17,  29 }
      }, {  // Band 1
        {  37,  75, 128 }, {  41,  76, 128 }, {  26,  66, 116 },
        {  12,  52,  94 }, {   2,  32,  55 }, {   1,  10,  16 }
      }, {  // Band 2
        {  50, 127, 154 }, {  37, 109, 152 }, {  16,  82, 121 },
        {   5,  59,  85 }, {   1,  35,  54 }, {   1,  13,  20 }
      }, {  // Band 3
        {  40, 142, 167 }, {  17, 110, 157 }, {   2,  71, 112 },
        {   1,  44,  72 }, {   1,  27,  45 }, {   1,  11,  17 }
      }, {  // Band 4
        {  30, 175, 188 }, {   9, 124, 169 }, {   1,  74, 116 },
        {   1,  48,  78 }, {   1,  30,  49 }, {   1,  11,  18 }
      }, {  // Band 5
        {  10, 222, 223 }, {   2, 150, 194 }, {   1,  83, 128 },
        {   1,  48,  79 }, {   1,  27,  45 }, {   1,  11,  17 }
      }
    }, {  // Inter
      {  // Band 0
        {  36,  41, 235 }, {  29,  36, 193 }, {  10,  27, 111 }
      }, {  // Band 1
        {  85, 165, 222 }, { 177, 162, 215 }, { 110, 135, 195 },
        {  57, 113, 168 }, {  23,  83, 120 }, {  10,  49,  61 }
      }, {  // Band 2
        {  85, 190, 223 }, {  36, 139, 200 }, {   5,  90, 146 },
        {   1,  60, 103 }, {   1,  38,  65 }, {   1,  18,  30 }
      }, {  // Band 3
        {  72, 202, 223 }, {  23, 141, 199 }, {   2,  86, 140 },
        {   1,  56,  97 }, {   1,  36,  61 }, {   1,  16,  27 }
      }, {  // Band 4
        {  55, 218, 225 }, {  13, 145, 200 }, {   1,  86, 141 },
        {   1,  57,  99 }, {   1,  35,  61 }, {   1,  13,  22 }
      }, {  // Band 5
        {  15, 235, 212 }, {   1, 132, 184 }, {   1,  84, 139 },
        {   1,  57,  97 }, {   1,  34,  56 }, {   1,  14,  23 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 181,  21, 201 }, {  61,  37, 123 }, {  10,  38,  71 }
      }, {  // Band 1
        {  47, 106, 172 }, {  95, 104, 173 }, {  42,  93, 159 },
        {  18,  77, 131 }, {   4,  50,  81 }, {   1,  17,  23 }
      }, {  // Band 2
        {  62, 147, 199 }, {  44, 130, 189 }, {  28, 102, 154 },
        {  18,  75, 115 }, {   2,  44,  65 }, {   1,  12,  19 }
      }, {  // Band 3
        {  55, 153, 210 }, {  24, 130, 194 }, {   3,  93, 146 },
        {   1,  61,  97 }, {   1,  31,  50 }, {   1,  10,  16 }
      }, {  // Band 4
        {  49, 186, 223 }, {  17, 148, 204 }, {   1,  96, 142 },
        {   1,  53,  83 }, {   1,  26,  44 }, {   1,  11,  17 }
      }, {  // Band 5
        {  13, 217, 212 }, {   2, 136, 180 }, {   1,  78, 124 },
        {   1,  50,  83 }, {   1,  29,  49 }, {   1,  14,  23 }
      }
    }, {  // Inter
      {  // Band 0
        { 197,  13, 247 }, {  82,  17, 222 }, {  25,  17, 162 }
      }, {  // Band 1
        { 126, 186, 247 }, { 234, 191, 243 }, { 176, 177, 234 },
        { 104, 158, 220 }, {  66, 128, 186 }, {  55,  90, 137 }
      }, {  // Band 2
        { 111, 197, 242 }, {  46, 158, 219 }, {   9, 104, 171 },
        {   2,  65, 125 }, {   1,  44,  80 }, {   1,  17,  91 }
      }, {  // Band 3
        { 104, 208, 245 }, {  39, 168, 224 }, {   3, 109, 162 },
        {   1,  79, 124 }, {   1,  50, 102 }, {   1,  43, 102 }
      }, {  // Band 4
        {  84, 220, 246 }, {  31, 177, 231 }, {   2, 115, 180 },
        {   1,  79, 134 }, {   1,  55,  77 }, {   1,  60,  79 }
      }, {  // Band 5
        {  43, 243, 240 }, {   8, 180, 217 }, {   1, 115, 166 },
        {   1,  84, 121 }, {   1,  51,  67 }, {   1,  16,   6 }
      }
    }
  }
};
#endif  // CONFIG_TX64X64
#endif  // CONFIG_Q_ADAPT_PROBS
static const aom_prob av1_default_blockzero_probs[TX_SIZES][PLANE_TYPES]
                                           [REF_TYPES][BLOCKZ_CONTEXTS] = {
#if CONFIG_CHROMA_2X2
  { // TX_2x2
    { // Y plane
      { 195, 84, 8, },  // Intra
      { 191, 124, 25, },  // Inter
    },
    { // UV plane
      { 214, 132, 42, },  // Intra
      { 229, 143, 46, },  // Inter
    },
  },
#endif
  { // TX_4x4
    { // Y plane
      { 195, 84, 8, },  // Intra
      { 191, 124, 25, },  // Inter
    },
    { // UV plane
      { 214, 132, 42, },  // Intra
      { 229, 143, 46, },  // Inter
    },
  },
  { // TX_8x8
    { // Y plane
      { 125, 52, 6, },  // Intra
      { 202, 108, 18, },  // Inter
    },
    { // UV plane
      { 212, 113, 29, },  // Intra
      { 225, 144, 42, },  // Inter
    },
  },
  { // TX_16x16
    { // Y plane
      { 7, 5, 1, },  // Intra
      { 19, 19, 3, },  // Inter
    },
    { // UV plane
      { 211, 96, 22, },  // Intra
      { 233, 146, 43, },  // Inter
    },
  },
  { // TX_32x32
    { //  Y plane
      { 17, 7, 1, },  // Intra
      { 36, 29, 10, },  // Inter
    },
    { // UV plane
      { 181, 61, 10, },  // Intra
      { 197, 82, 25, },  // Inter
    },
  },
#if CONFIG_TX64X64
  { // TX_64x64 FIXME: currently the same as 32x32
    { //  Y plane
      { 17, 7, 1, },  // Intra
      { 36, 29, 10, },  // Inter
    },
    { // UV plane
      { 181, 61, 10, },  // Intra
      { 197, 82, 25, },  // Inter
    },
  },
#endif
};

static const coeff_cdf_model default_coef_head_cdf_4x4[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { AOM_ICDF(25024), AOM_ICDF(25863), AOM_ICDF(27361), AOM_ICDF(29796),
          AOM_ICDF(30374), AOM_ICDF(32768) },
        { AOM_ICDF(10816), AOM_ICDF(14127), AOM_ICDF(17116), AOM_ICDF(23516),
          AOM_ICDF(24999), AOM_ICDF(32768) },
        { AOM_ICDF(1088), AOM_ICDF(6358), AOM_ICDF(8428), AOM_ICDF(16648),
          AOM_ICDF(18276), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(14529), AOM_ICDF(18769), AOM_ICDF(29100), AOM_ICDF(29634),
         AOM_ICDF(32768) },
        {AOM_ICDF(12993), AOM_ICDF(17117), AOM_ICDF(28404), AOM_ICDF(28988),
         AOM_ICDF(32768) },
        {AOM_ICDF(11201), AOM_ICDF(14084), AOM_ICDF(25818), AOM_ICDF(26504),
         AOM_ICDF(32768) },
        {AOM_ICDF(9793), AOM_ICDF(11267), AOM_ICDF(21775), AOM_ICDF(22451),
         AOM_ICDF(32768) },
        {AOM_ICDF(7105), AOM_ICDF(7562), AOM_ICDF(15777), AOM_ICDF(16225),
         AOM_ICDF(32768) },
        {AOM_ICDF(3905), AOM_ICDF(3966), AOM_ICDF(8359), AOM_ICDF(8526),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(20033), AOM_ICDF(23643), AOM_ICDF(31102), AOM_ICDF(31374),
         AOM_ICDF(32768) },
        {AOM_ICDF(16321), AOM_ICDF(20350), AOM_ICDF(30167), AOM_ICDF(30546),
         AOM_ICDF(32768) },
        {AOM_ICDF(12993), AOM_ICDF(15512), AOM_ICDF(26859), AOM_ICDF(27396),
         AOM_ICDF(32768) },
        {AOM_ICDF(10305), AOM_ICDF(11659), AOM_ICDF(21669), AOM_ICDF(22330),
         AOM_ICDF(32768) },
        {AOM_ICDF(7361), AOM_ICDF(7819), AOM_ICDF(15450), AOM_ICDF(15940),
         AOM_ICDF(32768) },
        {AOM_ICDF(3521), AOM_ICDF(3580), AOM_ICDF(7805), AOM_ICDF(7976),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(21057), AOM_ICDF(25460), AOM_ICDF(31740), AOM_ICDF(31952),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(21173), AOM_ICDF(30761), AOM_ICDF(31092),
         AOM_ICDF(32768) },
        {AOM_ICDF(11841), AOM_ICDF(14615), AOM_ICDF(26188), AOM_ICDF(26824),
         AOM_ICDF(32768) },
        {AOM_ICDF(7745), AOM_ICDF(8991), AOM_ICDF(18937), AOM_ICDF(19707),
         AOM_ICDF(32768) },
        {AOM_ICDF(4417), AOM_ICDF(4706), AOM_ICDF(10342), AOM_ICDF(10890),
         AOM_ICDF(32768) },
        {AOM_ICDF(7617), AOM_ICDF(8392), AOM_ICDF(17295), AOM_ICDF(17915),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(20417), AOM_ICDF(26452), AOM_ICDF(32166), AOM_ICDF(32321),
         AOM_ICDF(32768) },
        {AOM_ICDF(15809), AOM_ICDF(21634), AOM_ICDF(30947), AOM_ICDF(31298),
         AOM_ICDF(32768) },
        {AOM_ICDF(10049), AOM_ICDF(12176), AOM_ICDF(23495), AOM_ICDF(24229),
         AOM_ICDF(32768) },
        {AOM_ICDF(5953), AOM_ICDF(6731), AOM_ICDF(16166), AOM_ICDF(16798),
         AOM_ICDF(32768) },
        {AOM_ICDF(6081), AOM_ICDF(6188), AOM_ICDF(8114), AOM_ICDF(8764),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2291), AOM_ICDF(4448), AOM_ICDF(5527),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(9153), AOM_ICDF(25905), AOM_ICDF(31431), AOM_ICDF(31934),
         AOM_ICDF(32768) },
        {AOM_ICDF(9025), AOM_ICDF(23345), AOM_ICDF(30033), AOM_ICDF(30965),
         AOM_ICDF(32768) },
        {AOM_ICDF(5953), AOM_ICDF(13835), AOM_ICDF(22032), AOM_ICDF(24664),
         AOM_ICDF(32768) },
        {AOM_ICDF(6337), AOM_ICDF(11435), AOM_ICDF(18366), AOM_ICDF(21418),
         AOM_ICDF(32768) },
        {AOM_ICDF(3137), AOM_ICDF(4871), AOM_ICDF(8519), AOM_ICDF(12426),
         AOM_ICDF(32768) },
        {AOM_ICDF(1857), AOM_ICDF(2727), AOM_ICDF(5540), AOM_ICDF(8757),
         AOM_ICDF(32768) } } },
    {   // Intra
      { // Band 0
        { AOM_ICDF(24512), AOM_ICDF(26673), AOM_ICDF(28962), AOM_ICDF(31929),
          AOM_ICDF(32126), AOM_ICDF(32768) },
        { AOM_ICDF(15936), AOM_ICDF(21711), AOM_ICDF(25569), AOM_ICDF(30899),
          AOM_ICDF(31305), AOM_ICDF(32768) },
        { AOM_ICDF(3264), AOM_ICDF(14756), AOM_ICDF(20107), AOM_ICDF(29407),
          AOM_ICDF(30032), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(21313), AOM_ICDF(26020), AOM_ICDF(32523), AOM_ICDF(32575),
         AOM_ICDF(32768) },
        {AOM_ICDF(18369), AOM_ICDF(24215), AOM_ICDF(32291), AOM_ICDF(32391),
         AOM_ICDF(32768) },
        {AOM_ICDF(15297), AOM_ICDF(19637), AOM_ICDF(30414), AOM_ICDF(30752),
         AOM_ICDF(32768) },
        {AOM_ICDF(11713), AOM_ICDF(14040), AOM_ICDF(25408), AOM_ICDF(26033),
         AOM_ICDF(32768) },
        {AOM_ICDF(9537), AOM_ICDF(10173), AOM_ICDF(18839), AOM_ICDF(19315),
         AOM_ICDF(32768) },
        {AOM_ICDF(9025), AOM_ICDF(9093), AOM_ICDF(13987), AOM_ICDF(14115),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(22721), AOM_ICDF(27599), AOM_ICDF(32592), AOM_ICDF(32636),
         AOM_ICDF(32768) },
        {AOM_ICDF(19009), AOM_ICDF(24676), AOM_ICDF(32258), AOM_ICDF(32367),
         AOM_ICDF(32768) },
        {AOM_ICDF(12737), AOM_ICDF(16769), AOM_ICDF(28739), AOM_ICDF(29247),
         AOM_ICDF(32768) },
        {AOM_ICDF(8769), AOM_ICDF(10956), AOM_ICDF(21941), AOM_ICDF(22840),
         AOM_ICDF(32768) },
        {AOM_ICDF(6721), AOM_ICDF(7678), AOM_ICDF(15319), AOM_ICDF(16290),
         AOM_ICDF(32768) },
        {AOM_ICDF(4417), AOM_ICDF(4430), AOM_ICDF(4583), AOM_ICDF(5712),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(22849), AOM_ICDF(28333), AOM_ICDF(32633), AOM_ICDF(32671),
         AOM_ICDF(32768) },
        {AOM_ICDF(18497), AOM_ICDF(24619), AOM_ICDF(32184), AOM_ICDF(32315),
         AOM_ICDF(32768) },
        {AOM_ICDF(11841), AOM_ICDF(14640), AOM_ICDF(27251), AOM_ICDF(27752),
         AOM_ICDF(32768) },
        {AOM_ICDF(8385), AOM_ICDF(10154), AOM_ICDF(18339), AOM_ICDF(19621),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(6977), AOM_ICDF(13787), AOM_ICDF(15289),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(194), AOM_ICDF(384), AOM_ICDF(479),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(20417), AOM_ICDF(28167), AOM_ICDF(32552), AOM_ICDF(32621),
         AOM_ICDF(32768) },
        {AOM_ICDF(16833), AOM_ICDF(23968), AOM_ICDF(31991), AOM_ICDF(32174),
         AOM_ICDF(32768) },
        {AOM_ICDF(10433), AOM_ICDF(13387), AOM_ICDF(26356), AOM_ICDF(26951),
         AOM_ICDF(32768) },
        {AOM_ICDF(5057), AOM_ICDF(6823), AOM_ICDF(18967), AOM_ICDF(19843),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(6479), AOM_ICDF(11672), AOM_ICDF(13052),
         AOM_ICDF(32768) },
        {AOM_ICDF(2241), AOM_ICDF(2265), AOM_ICDF(6355), AOM_ICDF(6432),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(12097), AOM_ICDF(28717), AOM_ICDF(32406), AOM_ICDF(32555),
         AOM_ICDF(32768) },
        {AOM_ICDF(10433), AOM_ICDF(26113), AOM_ICDF(31504), AOM_ICDF(31975),
         AOM_ICDF(32768) },
        {AOM_ICDF(5825), AOM_ICDF(14284), AOM_ICDF(21349), AOM_ICDF(24461),
         AOM_ICDF(32768) },
        {AOM_ICDF(4545), AOM_ICDF(8454), AOM_ICDF(12648), AOM_ICDF(17501),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(7173), AOM_ICDF(15272), AOM_ICDF(19322),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2183), AOM_ICDF(7202), AOM_ICDF(7377),
         AOM_ICDF(32768) } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { AOM_ICDF(27456), AOM_ICDF(28244), AOM_ICDF(31289), AOM_ICDF(32358),
          AOM_ICDF(32534), AOM_ICDF(32768) },
        { AOM_ICDF(16960), AOM_ICDF(21207), AOM_ICDF(26511), AOM_ICDF(30539),
          AOM_ICDF(31190), AOM_ICDF(32768) },
        { AOM_ICDF(5440), AOM_ICDF(13412), AOM_ICDF(18469), AOM_ICDF(26423),
          AOM_ICDF(27669), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(17857), AOM_ICDF(26327), AOM_ICDF(31983), AOM_ICDF(32219),
         AOM_ICDF(32768) },
        {AOM_ICDF(16065), AOM_ICDF(24198), AOM_ICDF(31431), AOM_ICDF(31785),
         AOM_ICDF(32768) },
        {AOM_ICDF(12865), AOM_ICDF(18011), AOM_ICDF(28454), AOM_ICDF(29166),
         AOM_ICDF(32768) },
        {AOM_ICDF(9665), AOM_ICDF(12501), AOM_ICDF(24331), AOM_ICDF(25147),
         AOM_ICDF(32768) },
        {AOM_ICDF(2753), AOM_ICDF(3121), AOM_ICDF(12661), AOM_ICDF(13034),
         AOM_ICDF(32768) },
        {AOM_ICDF(4033), AOM_ICDF(4140), AOM_ICDF(11834), AOM_ICDF(11977),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(21185), AOM_ICDF(28338), AOM_ICDF(32249), AOM_ICDF(32417),
         AOM_ICDF(32768) },
        {AOM_ICDF(18497), AOM_ICDF(25227), AOM_ICDF(31905), AOM_ICDF(32122),
         AOM_ICDF(32768) },
        {AOM_ICDF(12097), AOM_ICDF(16516), AOM_ICDF(28610), AOM_ICDF(29166),
         AOM_ICDF(32768) },
        {AOM_ICDF(9281), AOM_ICDF(11157), AOM_ICDF(21438), AOM_ICDF(22312),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(6566), AOM_ICDF(15585), AOM_ICDF(16340),
         AOM_ICDF(32768) },
        {AOM_ICDF(9409), AOM_ICDF(9659), AOM_ICDF(11827), AOM_ICDF(12911),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(22337), AOM_ICDF(29459), AOM_ICDF(32382), AOM_ICDF(32519),
         AOM_ICDF(32768) },
        {AOM_ICDF(16961), AOM_ICDF(25262), AOM_ICDF(31874), AOM_ICDF(32123),
         AOM_ICDF(32768) },
        {AOM_ICDF(12353), AOM_ICDF(17748), AOM_ICDF(29300), AOM_ICDF(29852),
         AOM_ICDF(32768) },
        {AOM_ICDF(9025), AOM_ICDF(11528), AOM_ICDF(24468), AOM_ICDF(25141),
         AOM_ICDF(32768) },
        {AOM_ICDF(6209), AOM_ICDF(6565), AOM_ICDF(15806), AOM_ICDF(16121),
         AOM_ICDF(32768) },
        {AOM_ICDF(2497), AOM_ICDF(2524), AOM_ICDF(7050), AOM_ICDF(7125),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(20417), AOM_ICDF(29779), AOM_ICDF(32552), AOM_ICDF(32636),
         AOM_ICDF(32768) },
        {AOM_ICDF(15553), AOM_ICDF(26420), AOM_ICDF(32063), AOM_ICDF(32295),
         AOM_ICDF(32768) },
        {AOM_ICDF(9665), AOM_ICDF(17946), AOM_ICDF(29385), AOM_ICDF(30096),
         AOM_ICDF(32768) },
        {AOM_ICDF(5569), AOM_ICDF(10207), AOM_ICDF(22410), AOM_ICDF(23836),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(16450), AOM_ICDF(16545), AOM_ICDF(16593),
         AOM_ICDF(32768) },
        {AOM_ICDF(2369), AOM_ICDF(2395), AOM_ICDF(6822), AOM_ICDF(6898),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(10177), AOM_ICDF(30567), AOM_ICDF(32725), AOM_ICDF(32745),
         AOM_ICDF(32768) },
        {AOM_ICDF(9537), AOM_ICDF(28243), AOM_ICDF(32179), AOM_ICDF(32423),
         AOM_ICDF(32768) },
        {AOM_ICDF(13377), AOM_ICDF(23187), AOM_ICDF(29322), AOM_ICDF(30382),
         AOM_ICDF(32768) },
        {AOM_ICDF(13121), AOM_ICDF(21346), AOM_ICDF(29507), AOM_ICDF(30326),
         AOM_ICDF(32768) },
        {AOM_ICDF(4417), AOM_ICDF(4939), AOM_ICDF(15104), AOM_ICDF(15535),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2680), AOM_ICDF(8218), AOM_ICDF(8338),
         AOM_ICDF(32768) } } },
    {   // Inter
      { // Band 0
        { AOM_ICDF(29376), AOM_ICDF(30098), AOM_ICDF(32421), AOM_ICDF(32766),
          AOM_ICDF(32767), AOM_ICDF(32768) },
        { AOM_ICDF(18368), AOM_ICDF(22916), AOM_ICDF(30116), AOM_ICDF(32541),
          AOM_ICDF(32650), AOM_ICDF(32768) },
        { AOM_ICDF(5952), AOM_ICDF(16505), AOM_ICDF(25955), AOM_ICDF(32163),
          AOM_ICDF(32365), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(19649), AOM_ICDF(30160), AOM_ICDF(32743), AOM_ICDF(32753),
         AOM_ICDF(32768) },
        {AOM_ICDF(18881), AOM_ICDF(28724), AOM_ICDF(32688), AOM_ICDF(32717),
         AOM_ICDF(32768) },
        {AOM_ICDF(16833), AOM_ICDF(23053), AOM_ICDF(31244), AOM_ICDF(31573),
         AOM_ICDF(32768) },
        {AOM_ICDF(14657), AOM_ICDF(17714), AOM_ICDF(26083), AOM_ICDF(26978),
         AOM_ICDF(32768) },
        {AOM_ICDF(14657), AOM_ICDF(16618), AOM_ICDF(24597), AOM_ICDF(25403),
         AOM_ICDF(32768) },
        {AOM_ICDF(4289), AOM_ICDF(4326), AOM_ICDF(10686), AOM_ICDF(10751),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(21953), AOM_ICDF(30956), AOM_ICDF(32748), AOM_ICDF(32757),
         AOM_ICDF(32768) },
        {AOM_ICDF(20929), AOM_ICDF(29412), AOM_ICDF(32700), AOM_ICDF(32725),
         AOM_ICDF(32768) },
        {AOM_ICDF(13377), AOM_ICDF(21495), AOM_ICDF(31216), AOM_ICDF(31569),
         AOM_ICDF(32768) },
        {AOM_ICDF(9153), AOM_ICDF(15097), AOM_ICDF(28295), AOM_ICDF(28990),
         AOM_ICDF(32768) },
        {AOM_ICDF(5313), AOM_ICDF(5363), AOM_ICDF(13839), AOM_ICDF(13894),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2652), AOM_ICDF(7276), AOM_ICDF(7351),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(20289), AOM_ICDF(31164), AOM_ICDF(32745), AOM_ICDF(32755),
         AOM_ICDF(32768) },
        {AOM_ICDF(17601), AOM_ICDF(29635), AOM_ICDF(32739), AOM_ICDF(32751),
         AOM_ICDF(32768) },
        {AOM_ICDF(18241), AOM_ICDF(24284), AOM_ICDF(32116), AOM_ICDF(32258),
         AOM_ICDF(32768) },
        {AOM_ICDF(32705), AOM_ICDF(32706), AOM_ICDF(32739), AOM_ICDF(32740),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(5750), AOM_ICDF(14739), AOM_ICDF(14792),
         AOM_ICDF(32768) },
        {AOM_ICDF(2881), AOM_ICDF(2913), AOM_ICDF(8427), AOM_ICDF(8498),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(19009), AOM_ICDF(31481), AOM_ICDF(32742), AOM_ICDF(32754),
         AOM_ICDF(32768) },
        {AOM_ICDF(15809), AOM_ICDF(30521), AOM_ICDF(32736), AOM_ICDF(32750),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(32705), AOM_ICDF(32737), AOM_ICDF(32753),
         AOM_ICDF(32768) },
        {AOM_ICDF(7873), AOM_ICDF(8039), AOM_ICDF(19981), AOM_ICDF(20068),
         AOM_ICDF(32768) },
        {AOM_ICDF(5313), AOM_ICDF(5366), AOM_ICDF(14376), AOM_ICDF(14430),
         AOM_ICDF(32768) },
        {AOM_ICDF(2753), AOM_ICDF(2789), AOM_ICDF(8909), AOM_ICDF(8979),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(11841), AOM_ICDF(32116), AOM_ICDF(32728), AOM_ICDF(32748),
         AOM_ICDF(32768) },
        {AOM_ICDF(12353), AOM_ICDF(32132), AOM_ICDF(32729), AOM_ICDF(32748),
         AOM_ICDF(32768) },
        {AOM_ICDF(7489), AOM_ICDF(12435), AOM_ICDF(25708), AOM_ICDF(26666),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(7486), AOM_ICDF(20238), AOM_ICDF(21009),
         AOM_ICDF(32768) },
        {AOM_ICDF(4929), AOM_ICDF(5579), AOM_ICDF(16402), AOM_ICDF(16866),
         AOM_ICDF(32768) },
        {AOM_ICDF(3009), AOM_ICDF(3246), AOM_ICDF(10158), AOM_ICDF(10533),
         AOM_ICDF(32768) } } } }
};
static const coeff_cdf_model default_coef_head_cdf_8x8[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { AOM_ICDF(16064), AOM_ICDF(18127), AOM_ICDF(22153), AOM_ICDF(27289),
          AOM_ICDF(28507), AOM_ICDF(32768) },
        { AOM_ICDF(6720), AOM_ICDF(10545), AOM_ICDF(13491), AOM_ICDF(20948),
          AOM_ICDF(22631), AOM_ICDF(32768) },
        { AOM_ICDF(832), AOM_ICDF(5270), AOM_ICDF(5918), AOM_ICDF(12645),
          AOM_ICDF(13532), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(14017), AOM_ICDF(16139), AOM_ICDF(26799), AOM_ICDF(27295),
         AOM_ICDF(32768) },
        {AOM_ICDF(12737), AOM_ICDF(15136), AOM_ICDF(26235), AOM_ICDF(26816),
         AOM_ICDF(32768) },
        {AOM_ICDF(10817), AOM_ICDF(12445), AOM_ICDF(23637), AOM_ICDF(24217),
         AOM_ICDF(32768) },
        {AOM_ICDF(8897), AOM_ICDF(9702), AOM_ICDF(20040), AOM_ICDF(20500),
         AOM_ICDF(32768) },
        {AOM_ICDF(5953), AOM_ICDF(6156), AOM_ICDF(13966), AOM_ICDF(14205),
         AOM_ICDF(32768) },
        {AOM_ICDF(2497), AOM_ICDF(2519), AOM_ICDF(6222), AOM_ICDF(6300),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(19777), AOM_ICDF(21403), AOM_ICDF(30054), AOM_ICDF(30269),
         AOM_ICDF(32768) },
        {AOM_ICDF(16193), AOM_ICDF(17913), AOM_ICDF(28593), AOM_ICDF(28883),
         AOM_ICDF(32768) },
        {AOM_ICDF(12609), AOM_ICDF(13572), AOM_ICDF(25248), AOM_ICDF(25534),
         AOM_ICDF(32768) },
        {AOM_ICDF(9665), AOM_ICDF(10118), AOM_ICDF(20721), AOM_ICDF(20968),
         AOM_ICDF(32768) },
        {AOM_ICDF(6849), AOM_ICDF(7028), AOM_ICDF(15202), AOM_ICDF(15391),
         AOM_ICDF(32768) },
        {AOM_ICDF(3009), AOM_ICDF(3036), AOM_ICDF(7601), AOM_ICDF(7675),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(22593), AOM_ICDF(23915), AOM_ICDF(31159), AOM_ICDF(31283),
         AOM_ICDF(32768) },
        {AOM_ICDF(17345), AOM_ICDF(18690), AOM_ICDF(29425), AOM_ICDF(29611),
         AOM_ICDF(32768) },
        {AOM_ICDF(11969), AOM_ICDF(12540), AOM_ICDF(24685), AOM_ICDF(24867),
         AOM_ICDF(32768) },
        {AOM_ICDF(8129), AOM_ICDF(8355), AOM_ICDF(18668), AOM_ICDF(18819),
         AOM_ICDF(32768) },
        {AOM_ICDF(4673), AOM_ICDF(4714), AOM_ICDF(11752), AOM_ICDF(11814),
         AOM_ICDF(32768) },
        {AOM_ICDF(1857), AOM_ICDF(1876), AOM_ICDF(5057), AOM_ICDF(5138),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(24513), AOM_ICDF(25718), AOM_ICDF(31947), AOM_ICDF(32014),
         AOM_ICDF(32768) },
        {AOM_ICDF(18881), AOM_ICDF(20029), AOM_ICDF(30409), AOM_ICDF(30527),
         AOM_ICDF(32768) },
        {AOM_ICDF(12481), AOM_ICDF(12953), AOM_ICDF(25201), AOM_ICDF(25341),
         AOM_ICDF(32768) },
        {AOM_ICDF(8385), AOM_ICDF(8528), AOM_ICDF(18815), AOM_ICDF(18910),
         AOM_ICDF(32768) },
        {AOM_ICDF(4289), AOM_ICDF(4327), AOM_ICDF(10797), AOM_ICDF(10861),
         AOM_ICDF(32768) },
        {AOM_ICDF(1857), AOM_ICDF(1872), AOM_ICDF(4332), AOM_ICDF(4415),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(26049), AOM_ICDF(27752), AOM_ICDF(32415), AOM_ICDF(32462),
         AOM_ICDF(32768) },
        {AOM_ICDF(20417), AOM_ICDF(22100), AOM_ICDF(31056), AOM_ICDF(31192),
         AOM_ICDF(32768) },
        {AOM_ICDF(12481), AOM_ICDF(13075), AOM_ICDF(24646), AOM_ICDF(24844),
         AOM_ICDF(32768) },
        {AOM_ICDF(7489), AOM_ICDF(7696), AOM_ICDF(17117), AOM_ICDF(17285),
         AOM_ICDF(32768) },
        {AOM_ICDF(3777), AOM_ICDF(3814), AOM_ICDF(10062), AOM_ICDF(10129),
         AOM_ICDF(32768) },
        {AOM_ICDF(1473), AOM_ICDF(1486), AOM_ICDF(3735), AOM_ICDF(3820),
         AOM_ICDF(32768) } } },
    {   // Intra
      { // Band 0
        { AOM_ICDF(25920), AOM_ICDF(27743), AOM_ICDF(29455), AOM_ICDF(32147),
          AOM_ICDF(32280), AOM_ICDF(32768) },
        { AOM_ICDF(13888), AOM_ICDF(19845), AOM_ICDF(23350), AOM_ICDF(30219),
          AOM_ICDF(30660), AOM_ICDF(32768) },
        { AOM_ICDF(2368), AOM_ICDF(12781), AOM_ICDF(16196), AOM_ICDF(27232),
          AOM_ICDF(27894), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(21697), AOM_ICDF(24758), AOM_ICDF(32358), AOM_ICDF(32417),
         AOM_ICDF(32768) },
        {AOM_ICDF(20289), AOM_ICDF(23960), AOM_ICDF(32111), AOM_ICDF(32213),
         AOM_ICDF(32768) },
        {AOM_ICDF(17345), AOM_ICDF(19966), AOM_ICDF(30630), AOM_ICDF(30841),
         AOM_ICDF(32768) },
        {AOM_ICDF(14529), AOM_ICDF(16070), AOM_ICDF(27461), AOM_ICDF(27777),
         AOM_ICDF(32768) },
        {AOM_ICDF(9793), AOM_ICDF(10613), AOM_ICDF(21146), AOM_ICDF(21566),
         AOM_ICDF(32768) },
        {AOM_ICDF(6977), AOM_ICDF(7162), AOM_ICDF(15591), AOM_ICDF(15776),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(23617), AOM_ICDF(26783), AOM_ICDF(32572), AOM_ICDF(32607),
         AOM_ICDF(32768) },
        {AOM_ICDF(20801), AOM_ICDF(24292), AOM_ICDF(32185), AOM_ICDF(32275),
         AOM_ICDF(32768) },
        {AOM_ICDF(15169), AOM_ICDF(17905), AOM_ICDF(29916), AOM_ICDF(30181),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(12972), AOM_ICDF(25565), AOM_ICDF(26064),
         AOM_ICDF(32768) },
        {AOM_ICDF(6849), AOM_ICDF(8334), AOM_ICDF(18543), AOM_ICDF(19446),
         AOM_ICDF(32768) },
        {AOM_ICDF(3649), AOM_ICDF(4346), AOM_ICDF(12351), AOM_ICDF(13169),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(25281), AOM_ICDF(28440), AOM_ICDF(32667), AOM_ICDF(32689),
         AOM_ICDF(32768) },
        {AOM_ICDF(22081), AOM_ICDF(25694), AOM_ICDF(32414), AOM_ICDF(32476),
         AOM_ICDF(32768) },
        {AOM_ICDF(15297), AOM_ICDF(18341), AOM_ICDF(30141), AOM_ICDF(30410),
         AOM_ICDF(32768) },
        {AOM_ICDF(10305), AOM_ICDF(12381), AOM_ICDF(24477), AOM_ICDF(25084),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(6673), AOM_ICDF(16325), AOM_ICDF(17080),
         AOM_ICDF(32768) },
        {AOM_ICDF(2369), AOM_ICDF(2393), AOM_ICDF(6466), AOM_ICDF(6543),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(25921), AOM_ICDF(29445), AOM_ICDF(32729), AOM_ICDF(32739),
         AOM_ICDF(32768) },
        {AOM_ICDF(22465), AOM_ICDF(26834), AOM_ICDF(32588), AOM_ICDF(32627),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(20062), AOM_ICDF(31016), AOM_ICDF(31233),
         AOM_ICDF(32768) },
        {AOM_ICDF(11073), AOM_ICDF(13165), AOM_ICDF(25353), AOM_ICDF(25896),
         AOM_ICDF(32768) },
        {AOM_ICDF(11713), AOM_ICDF(13837), AOM_ICDF(20144), AOM_ICDF(21734),
         AOM_ICDF(32768) },
        {AOM_ICDF(2241), AOM_ICDF(2265), AOM_ICDF(6355), AOM_ICDF(6432),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(26177), AOM_ICDF(29403), AOM_ICDF(32705), AOM_ICDF(32721),
         AOM_ICDF(32768) },
        {AOM_ICDF(22337), AOM_ICDF(26344), AOM_ICDF(32545), AOM_ICDF(32589),
         AOM_ICDF(32768) },
        {AOM_ICDF(19009), AOM_ICDF(21527), AOM_ICDF(31775), AOM_ICDF(31873),
         AOM_ICDF(32768) },
        {AOM_ICDF(11585), AOM_ICDF(12685), AOM_ICDF(22632), AOM_ICDF(23137),
         AOM_ICDF(32768) },
        {AOM_ICDF(8257), AOM_ICDF(8305), AOM_ICDF(16444), AOM_ICDF(16492),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2183), AOM_ICDF(7202), AOM_ICDF(7377),
         AOM_ICDF(32768) } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { AOM_ICDF(27200), AOM_ICDF(27981), AOM_ICDF(31389), AOM_ICDF(32444),
          AOM_ICDF(32592), AOM_ICDF(32768) },
        { AOM_ICDF(14528), AOM_ICDF(19068), AOM_ICDF(24887), AOM_ICDF(29901),
          AOM_ICDF(30688), AOM_ICDF(32768) },
        { AOM_ICDF(3776), AOM_ICDF(11778), AOM_ICDF(14700), AOM_ICDF(23745),
          AOM_ICDF(24854), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(20289), AOM_ICDF(25202), AOM_ICDF(31672), AOM_ICDF(31909),
         AOM_ICDF(32768) },
        {AOM_ICDF(18369), AOM_ICDF(23493), AOM_ICDF(31166), AOM_ICDF(31487),
         AOM_ICDF(32768) },
        {AOM_ICDF(15425), AOM_ICDF(18619), AOM_ICDF(28941), AOM_ICDF(29393),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(12535), AOM_ICDF(24287), AOM_ICDF(24792),
         AOM_ICDF(32768) },
        {AOM_ICDF(6465), AOM_ICDF(6810), AOM_ICDF(15764), AOM_ICDF(16080),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2137), AOM_ICDF(6125), AOM_ICDF(6203),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(23745), AOM_ICDF(27041), AOM_ICDF(31976), AOM_ICDF(32135),
         AOM_ICDF(32768) },
        {AOM_ICDF(19521), AOM_ICDF(22766), AOM_ICDF(31139), AOM_ICDF(31367),
         AOM_ICDF(32768) },
        {AOM_ICDF(14273), AOM_ICDF(15834), AOM_ICDF(27820), AOM_ICDF(28105),
         AOM_ICDF(32768) },
        {AOM_ICDF(9537), AOM_ICDF(10445), AOM_ICDF(22106), AOM_ICDF(22491),
         AOM_ICDF(32768) },
        {AOM_ICDF(7233), AOM_ICDF(7386), AOM_ICDF(15961), AOM_ICDF(16109),
         AOM_ICDF(32768) },
        {AOM_ICDF(2369), AOM_ICDF(2401), AOM_ICDF(7891), AOM_ICDF(7964),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(26305), AOM_ICDF(28703), AOM_ICDF(32352), AOM_ICDF(32435),
         AOM_ICDF(32768) },
        {AOM_ICDF(20673), AOM_ICDF(23490), AOM_ICDF(31517), AOM_ICDF(31680),
         AOM_ICDF(32768) },
        {AOM_ICDF(14017), AOM_ICDF(15251), AOM_ICDF(27458), AOM_ICDF(27702),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(11374), AOM_ICDF(22496), AOM_ICDF(22687),
         AOM_ICDF(32768) },
        {AOM_ICDF(9153), AOM_ICDF(9435), AOM_ICDF(22299), AOM_ICDF(22411),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(269), AOM_ICDF(13236), AOM_ICDF(13293),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(27713), AOM_ICDF(29770), AOM_ICDF(32522), AOM_ICDF(32575),
         AOM_ICDF(32768) },
        {AOM_ICDF(21569), AOM_ICDF(24342), AOM_ICDF(31785), AOM_ICDF(31919),
         AOM_ICDF(32768) },
        {AOM_ICDF(15297), AOM_ICDF(16497), AOM_ICDF(28367), AOM_ICDF(28569),
         AOM_ICDF(32768) },
        {AOM_ICDF(17601), AOM_ICDF(17828), AOM_ICDF(24444), AOM_ICDF(24582),
         AOM_ICDF(32768) },
        {AOM_ICDF(6977), AOM_ICDF(7035), AOM_ICDF(16901), AOM_ICDF(16947),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(384), AOM_ICDF(32706), AOM_ICDF(32707),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(28737), AOM_ICDF(30879), AOM_ICDF(32667), AOM_ICDF(32695),
         AOM_ICDF(32768) },
        {AOM_ICDF(22593), AOM_ICDF(26241), AOM_ICDF(32073), AOM_ICDF(32207),
         AOM_ICDF(32768) },
        {AOM_ICDF(16577), AOM_ICDF(19148), AOM_ICDF(28436), AOM_ICDF(28906),
         AOM_ICDF(32768) },
        {AOM_ICDF(12993), AOM_ICDF(14005), AOM_ICDF(23151), AOM_ICDF(23630),
         AOM_ICDF(32768) },
        {AOM_ICDF(7617), AOM_ICDF(9188), AOM_ICDF(22797), AOM_ICDF(23313),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2680), AOM_ICDF(8218), AOM_ICDF(8338),
         AOM_ICDF(32768) } } },
    {   // Inter
      { // Band 0
        { AOM_ICDF(28864), AOM_ICDF(29988), AOM_ICDF(32423), AOM_ICDF(32766),
          AOM_ICDF(32767), AOM_ICDF(32768) },
        { AOM_ICDF(18496), AOM_ICDF(24572), AOM_ICDF(30167), AOM_ICDF(32687),
          AOM_ICDF(32737), AOM_ICDF(32768) },
        { AOM_ICDF(5440), AOM_ICDF(19618), AOM_ICDF(25332), AOM_ICDF(32393),
          AOM_ICDF(32491), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(23745), AOM_ICDF(29427), AOM_ICDF(32751), AOM_ICDF(32757),
         AOM_ICDF(32768) },
        {AOM_ICDF(23745), AOM_ICDF(28704), AOM_ICDF(32716), AOM_ICDF(32731),
         AOM_ICDF(32768) },
        {AOM_ICDF(23105), AOM_ICDF(27943), AOM_ICDF(32524), AOM_ICDF(32587),
         AOM_ICDF(32768) },
        {AOM_ICDF(21057), AOM_ICDF(24773), AOM_ICDF(29589), AOM_ICDF(30282),
         AOM_ICDF(32768) },
        {AOM_ICDF(12609), AOM_ICDF(14823), AOM_ICDF(23831), AOM_ICDF(24713),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(16450), AOM_ICDF(16545), AOM_ICDF(16593),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(25025), AOM_ICDF(30203), AOM_ICDF(32754), AOM_ICDF(32759),
         AOM_ICDF(32768) },
        {AOM_ICDF(23617), AOM_ICDF(28361), AOM_ICDF(32715), AOM_ICDF(32729),
         AOM_ICDF(32768) },
        {AOM_ICDF(17985), AOM_ICDF(21562), AOM_ICDF(31354), AOM_ICDF(31543),
         AOM_ICDF(32768) },
        {AOM_ICDF(12353), AOM_ICDF(18915), AOM_ICDF(28742), AOM_ICDF(29548),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(289), AOM_ICDF(16545), AOM_ICDF(16593),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2652), AOM_ICDF(7276), AOM_ICDF(7351),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(26433), AOM_ICDF(30892), AOM_ICDF(32757), AOM_ICDF(32761),
         AOM_ICDF(32768) },
        {AOM_ICDF(24513), AOM_ICDF(29274), AOM_ICDF(32721), AOM_ICDF(32735),
         AOM_ICDF(32768) },
        {AOM_ICDF(20161), AOM_ICDF(24040), AOM_ICDF(32055), AOM_ICDF(32171),
         AOM_ICDF(32768) },
        {AOM_ICDF(21953), AOM_ICDF(24678), AOM_ICDF(27382), AOM_ICDF(28734),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(5750), AOM_ICDF(14739), AOM_ICDF(14792),
         AOM_ICDF(32768) },
        {AOM_ICDF(2881), AOM_ICDF(2913), AOM_ICDF(8427), AOM_ICDF(8498),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(27457), AOM_ICDF(31485), AOM_ICDF(32759), AOM_ICDF(32763),
         AOM_ICDF(32768) },
        {AOM_ICDF(24129), AOM_ICDF(29502), AOM_ICDF(32752), AOM_ICDF(32757),
         AOM_ICDF(32768) },
        {AOM_ICDF(19009), AOM_ICDF(25452), AOM_ICDF(32473), AOM_ICDF(32544),
         AOM_ICDF(32768) },
        {AOM_ICDF(32705), AOM_ICDF(32706), AOM_ICDF(32737), AOM_ICDF(32738),
         AOM_ICDF(32768) },
        {AOM_ICDF(5313), AOM_ICDF(5366), AOM_ICDF(14376), AOM_ICDF(14430),
         AOM_ICDF(32768) },
        {AOM_ICDF(2753), AOM_ICDF(2789), AOM_ICDF(8909), AOM_ICDF(8979),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(27841), AOM_ICDF(32288), AOM_ICDF(32759), AOM_ICDF(32764),
         AOM_ICDF(32768) },
        {AOM_ICDF(19137), AOM_ICDF(30271), AOM_ICDF(32742), AOM_ICDF(32753),
         AOM_ICDF(32768) },
        {AOM_ICDF(18625), AOM_ICDF(27739), AOM_ICDF(29979), AOM_ICDF(31099),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(7486), AOM_ICDF(20238), AOM_ICDF(21009),
         AOM_ICDF(32768) },
        {AOM_ICDF(4929), AOM_ICDF(5579), AOM_ICDF(16402), AOM_ICDF(16866),
         AOM_ICDF(32768) },
        {AOM_ICDF(3009), AOM_ICDF(3246), AOM_ICDF(10158), AOM_ICDF(10533),
         AOM_ICDF(32768) } } } }
};
static const coeff_cdf_model default_coef_head_cdf_16x16[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { AOM_ICDF(960), AOM_ICDF(4882), AOM_ICDF(9467), AOM_ICDF(17710),
          AOM_ICDF(20412), AOM_ICDF(32768) },
        { AOM_ICDF(704), AOM_ICDF(4657), AOM_ICDF(6561), AOM_ICDF(14507),
          AOM_ICDF(16279), AOM_ICDF(32768) },
        { AOM_ICDF(192), AOM_ICDF(3443), AOM_ICDF(3759), AOM_ICDF(9011),
          AOM_ICDF(9685), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(12481), AOM_ICDF(13958), AOM_ICDF(24487), AOM_ICDF(24997),
         AOM_ICDF(32768) },
        {AOM_ICDF(11457), AOM_ICDF(13075), AOM_ICDF(23820), AOM_ICDF(24406),
         AOM_ICDF(32768) },
        {AOM_ICDF(9793), AOM_ICDF(11127), AOM_ICDF(21775), AOM_ICDF(22387),
         AOM_ICDF(32768) },
        {AOM_ICDF(7745), AOM_ICDF(8457), AOM_ICDF(18155), AOM_ICDF(18655),
         AOM_ICDF(32768) },
        {AOM_ICDF(5441), AOM_ICDF(5668), AOM_ICDF(13180), AOM_ICDF(13467),
         AOM_ICDF(32768) },
        {AOM_ICDF(2497), AOM_ICDF(2520), AOM_ICDF(6340), AOM_ICDF(6417),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(19521), AOM_ICDF(20572), AOM_ICDF(28965), AOM_ICDF(29177),
         AOM_ICDF(32768) },
        {AOM_ICDF(15425), AOM_ICDF(16741), AOM_ICDF(27247), AOM_ICDF(27554),
         AOM_ICDF(32768) },
        {AOM_ICDF(11969), AOM_ICDF(12690), AOM_ICDF(23872), AOM_ICDF(24141),
         AOM_ICDF(32768) },
        {AOM_ICDF(9281), AOM_ICDF(9678), AOM_ICDF(19970), AOM_ICDF(20207),
         AOM_ICDF(32768) },
        {AOM_ICDF(6081), AOM_ICDF(6266), AOM_ICDF(14682), AOM_ICDF(14876),
         AOM_ICDF(32768) },
        {AOM_ICDF(2753), AOM_ICDF(2779), AOM_ICDF(7150), AOM_ICDF(7225),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(22337), AOM_ICDF(23293), AOM_ICDF(30630), AOM_ICDF(30753),
         AOM_ICDF(32768) },
        {AOM_ICDF(16321), AOM_ICDF(17427), AOM_ICDF(28368), AOM_ICDF(28570),
         AOM_ICDF(32768) },
        {AOM_ICDF(11457), AOM_ICDF(11907), AOM_ICDF(23570), AOM_ICDF(23741),
         AOM_ICDF(32768) },
        {AOM_ICDF(7233), AOM_ICDF(7331), AOM_ICDF(17258), AOM_ICDF(17334),
         AOM_ICDF(32768) },
        {AOM_ICDF(4033), AOM_ICDF(4070), AOM_ICDF(10375), AOM_ICDF(10441),
         AOM_ICDF(32768) },
        {AOM_ICDF(1601), AOM_ICDF(1619), AOM_ICDF(4706), AOM_ICDF(4788),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(24769), AOM_ICDF(25536), AOM_ICDF(31660), AOM_ICDF(31722),
         AOM_ICDF(32768) },
        {AOM_ICDF(18113), AOM_ICDF(18886), AOM_ICDF(29420), AOM_ICDF(29534),
         AOM_ICDF(32768) },
        {AOM_ICDF(11201), AOM_ICDF(11412), AOM_ICDF(23207), AOM_ICDF(23291),
         AOM_ICDF(32768) },
        {AOM_ICDF(6977), AOM_ICDF(7033), AOM_ICDF(16599), AOM_ICDF(16646),
         AOM_ICDF(32768) },
        {AOM_ICDF(4033), AOM_ICDF(4070), AOM_ICDF(10375), AOM_ICDF(10441),
         AOM_ICDF(32768) },
        {AOM_ICDF(1601), AOM_ICDF(1620), AOM_ICDF(4827), AOM_ICDF(4909),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(28353), AOM_ICDF(28831), AOM_ICDF(32502), AOM_ICDF(32517),
         AOM_ICDF(32768) },
        {AOM_ICDF(21441), AOM_ICDF(21869), AOM_ICDF(30977), AOM_ICDF(31017),
         AOM_ICDF(32768) },
        {AOM_ICDF(11969), AOM_ICDF(12088), AOM_ICDF(24116), AOM_ICDF(24158),
         AOM_ICDF(32768) },
        {AOM_ICDF(7489), AOM_ICDF(7547), AOM_ICDF(17413), AOM_ICDF(17458),
         AOM_ICDF(32768) },
        {AOM_ICDF(4545), AOM_ICDF(4585), AOM_ICDF(11325), AOM_ICDF(11388),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2133), AOM_ICDF(5526), AOM_ICDF(5606),
         AOM_ICDF(32768) } } },
    {   // Intra
      { // Band 0
        { AOM_ICDF(2496), AOM_ICDF(8717), AOM_ICDF(17280), AOM_ICDF(28922),
          AOM_ICDF(29751), AOM_ICDF(32768) },
        { AOM_ICDF(2496), AOM_ICDF(9665), AOM_ICDF(15235), AOM_ICDF(26542),
          AOM_ICDF(27580), AOM_ICDF(32768) },
        { AOM_ICDF(448), AOM_ICDF(9240), AOM_ICDF(11886), AOM_ICDF(24124),
          AOM_ICDF(24898), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(21057), AOM_ICDF(22896), AOM_ICDF(31877), AOM_ICDF(31953),
         AOM_ICDF(32768) },
        {AOM_ICDF(20673), AOM_ICDF(23151), AOM_ICDF(31706), AOM_ICDF(31825),
         AOM_ICDF(32768) },
        {AOM_ICDF(18753), AOM_ICDF(20519), AOM_ICDF(30497), AOM_ICDF(30668),
         AOM_ICDF(32768) },
        {AOM_ICDF(15425), AOM_ICDF(16608), AOM_ICDF(27789), AOM_ICDF(28027),
         AOM_ICDF(32768) },
        {AOM_ICDF(10305), AOM_ICDF(10977), AOM_ICDF(21405), AOM_ICDF(21749),
         AOM_ICDF(32768) },
        {AOM_ICDF(3649), AOM_ICDF(3812), AOM_ICDF(11213), AOM_ICDF(11445),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(24001), AOM_ICDF(25899), AOM_ICDF(32307), AOM_ICDF(32360),
         AOM_ICDF(32768) },
        {AOM_ICDF(20929), AOM_ICDF(22941), AOM_ICDF(31775), AOM_ICDF(31867),
         AOM_ICDF(32768) },
        {AOM_ICDF(15169), AOM_ICDF(16734), AOM_ICDF(29228), AOM_ICDF(29425),
         AOM_ICDF(32768) },
        {AOM_ICDF(10561), AOM_ICDF(12047), AOM_ICDF(24918), AOM_ICDF(25324),
         AOM_ICDF(32768) },
        {AOM_ICDF(6977), AOM_ICDF(7929), AOM_ICDF(18311), AOM_ICDF(18918),
         AOM_ICDF(32768) },
        {AOM_ICDF(3649), AOM_ICDF(3760), AOM_ICDF(9962), AOM_ICDF(10162),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(25793), AOM_ICDF(27526), AOM_ICDF(32565), AOM_ICDF(32591),
         AOM_ICDF(32768) },
        {AOM_ICDF(21825), AOM_ICDF(23885), AOM_ICDF(32064), AOM_ICDF(32135),
         AOM_ICDF(32768) },
        {AOM_ICDF(15041), AOM_ICDF(16286), AOM_ICDF(29203), AOM_ICDF(29360),
         AOM_ICDF(32768) },
        {AOM_ICDF(10433), AOM_ICDF(11058), AOM_ICDF(24349), AOM_ICDF(24538),
         AOM_ICDF(32768) },
        {AOM_ICDF(5569), AOM_ICDF(6016), AOM_ICDF(16460), AOM_ICDF(16794),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(194), AOM_ICDF(384), AOM_ICDF(479),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(26433), AOM_ICDF(28398), AOM_ICDF(32682), AOM_ICDF(32696),
         AOM_ICDF(32768) },
        {AOM_ICDF(22977), AOM_ICDF(25086), AOM_ICDF(32367), AOM_ICDF(32412),
         AOM_ICDF(32768) },
        {AOM_ICDF(16577), AOM_ICDF(17928), AOM_ICDF(30144), AOM_ICDF(30275),
         AOM_ICDF(32768) },
        {AOM_ICDF(12481), AOM_ICDF(13352), AOM_ICDF(25993), AOM_ICDF(26211),
         AOM_ICDF(32768) },
        {AOM_ICDF(7745), AOM_ICDF(8069), AOM_ICDF(20501), AOM_ICDF(20657),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(16450), AOM_ICDF(16545), AOM_ICDF(16593),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(27841), AOM_ICDF(29700), AOM_ICDF(32721), AOM_ICDF(32730),
         AOM_ICDF(32768) },
        {AOM_ICDF(23873), AOM_ICDF(26202), AOM_ICDF(32578), AOM_ICDF(32604),
         AOM_ICDF(32768) },
        {AOM_ICDF(17729), AOM_ICDF(19046), AOM_ICDF(30448), AOM_ICDF(30568),
         AOM_ICDF(32768) },
        {AOM_ICDF(13505), AOM_ICDF(14508), AOM_ICDF(26034), AOM_ICDF(26304),
         AOM_ICDF(32768) },
        {AOM_ICDF(10049), AOM_ICDF(10494), AOM_ICDF(19945), AOM_ICDF(20233),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2183), AOM_ICDF(7202), AOM_ICDF(7377),
         AOM_ICDF(32768) } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { AOM_ICDF(27072), AOM_ICDF(27916), AOM_ICDF(31095), AOM_ICDF(32400),
          AOM_ICDF(32553), AOM_ICDF(32768) },
        { AOM_ICDF(12352), AOM_ICDF(16792), AOM_ICDF(22516), AOM_ICDF(28853),
          AOM_ICDF(29797), AOM_ICDF(32768) },
        { AOM_ICDF(2880), AOM_ICDF(9023), AOM_ICDF(11126), AOM_ICDF(20602),
          AOM_ICDF(21713), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(20161), AOM_ICDF(24785), AOM_ICDF(31070), AOM_ICDF(31430),
         AOM_ICDF(32768) },
        {AOM_ICDF(17985), AOM_ICDF(22773), AOM_ICDF(30430), AOM_ICDF(30880),
         AOM_ICDF(32768) },
        {AOM_ICDF(15937), AOM_ICDF(18802), AOM_ICDF(28265), AOM_ICDF(28788),
         AOM_ICDF(32768) },
        {AOM_ICDF(11841), AOM_ICDF(13587), AOM_ICDF(24798), AOM_ICDF(25335),
         AOM_ICDF(32768) },
        {AOM_ICDF(8769), AOM_ICDF(9160), AOM_ICDF(19316), AOM_ICDF(19566),
         AOM_ICDF(32768) },
        {AOM_ICDF(5313), AOM_ICDF(5357), AOM_ICDF(12874), AOM_ICDF(12932),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(24129), AOM_ICDF(26501), AOM_ICDF(31672), AOM_ICDF(31844),
         AOM_ICDF(32768) },
        {AOM_ICDF(19649), AOM_ICDF(21553), AOM_ICDF(30130), AOM_ICDF(30370),
         AOM_ICDF(32768) },
        {AOM_ICDF(11713), AOM_ICDF(13134), AOM_ICDF(25983), AOM_ICDF(26321),
         AOM_ICDF(32768) },
        {AOM_ICDF(9409), AOM_ICDF(9948), AOM_ICDF(21408), AOM_ICDF(21663),
         AOM_ICDF(32768) },
        {AOM_ICDF(5569), AOM_ICDF(5757), AOM_ICDF(14335), AOM_ICDF(14533),
         AOM_ICDF(32768) },
        {AOM_ICDF(2241), AOM_ICDF(2305), AOM_ICDF(13152), AOM_ICDF(13209),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(26817), AOM_ICDF(28135), AOM_ICDF(32130), AOM_ICDF(32209),
         AOM_ICDF(32768) },
        {AOM_ICDF(20161), AOM_ICDF(21412), AOM_ICDF(30331), AOM_ICDF(30481),
         AOM_ICDF(32768) },
        {AOM_ICDF(13377), AOM_ICDF(13798), AOM_ICDF(26065), AOM_ICDF(26176),
         AOM_ICDF(32768) },
        {AOM_ICDF(8129), AOM_ICDF(8290), AOM_ICDF(19920), AOM_ICDF(20008),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(5751), AOM_ICDF(14950), AOM_ICDF(15002),
         AOM_ICDF(32768) },
        {AOM_ICDF(5569), AOM_ICDF(5601), AOM_ICDF(11041), AOM_ICDF(11105),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(28225), AOM_ICDF(29079), AOM_ICDF(32387), AOM_ICDF(32426),
         AOM_ICDF(32768) },
        {AOM_ICDF(21185), AOM_ICDF(22046), AOM_ICDF(30982), AOM_ICDF(31061),
         AOM_ICDF(32768) },
        {AOM_ICDF(13377), AOM_ICDF(13595), AOM_ICDF(25762), AOM_ICDF(25824),
         AOM_ICDF(32768) },
        {AOM_ICDF(8001), AOM_ICDF(8123), AOM_ICDF(20530), AOM_ICDF(20590),
         AOM_ICDF(32768) },
        {AOM_ICDF(4289), AOM_ICDF(4322), AOM_ICDF(9907), AOM_ICDF(9974),
         AOM_ICDF(32768) },
        {AOM_ICDF(3393), AOM_ICDF(3412), AOM_ICDF(6663), AOM_ICDF(6739),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(30529), AOM_ICDF(31014), AOM_ICDF(32651), AOM_ICDF(32664),
         AOM_ICDF(32768) },
        {AOM_ICDF(23489), AOM_ICDF(24268), AOM_ICDF(31627), AOM_ICDF(31682),
         AOM_ICDF(32768) },
        {AOM_ICDF(14017), AOM_ICDF(14239), AOM_ICDF(26653), AOM_ICDF(26707),
         AOM_ICDF(32768) },
        {AOM_ICDF(11201), AOM_ICDF(11317), AOM_ICDF(23122), AOM_ICDF(23169),
         AOM_ICDF(32768) },
        {AOM_ICDF(6721), AOM_ICDF(6768), AOM_ICDF(14810), AOM_ICDF(14863),
         AOM_ICDF(32768) },
        {AOM_ICDF(6593), AOM_ICDF(6632), AOM_ICDF(13188), AOM_ICDF(13245),
         AOM_ICDF(32768) } } },
    {   // Inter
      { // Band 0
        { AOM_ICDF(29888), AOM_ICDF(30492), AOM_ICDF(32500), AOM_ICDF(32766),
          AOM_ICDF(32767), AOM_ICDF(32768) },
        { AOM_ICDF(18752), AOM_ICDF(23235), AOM_ICDF(29846), AOM_ICDF(32214),
          AOM_ICDF(32442), AOM_ICDF(32768) },
        { AOM_ICDF(5568), AOM_ICDF(17762), AOM_ICDF(25039), AOM_ICDF(31213),
          AOM_ICDF(31651), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(26433), AOM_ICDF(29681), AOM_ICDF(32757), AOM_ICDF(32760),
         AOM_ICDF(32768) },
        {AOM_ICDF(24769), AOM_ICDF(28761), AOM_ICDF(32722), AOM_ICDF(32734),
         AOM_ICDF(32768) },
        {AOM_ICDF(22209), AOM_ICDF(26975), AOM_ICDF(32418), AOM_ICDF(32500),
         AOM_ICDF(32768) },
        {AOM_ICDF(16321), AOM_ICDF(21333), AOM_ICDF(28368), AOM_ICDF(29283),
         AOM_ICDF(32768) },
        {AOM_ICDF(12865), AOM_ICDF(14775), AOM_ICDF(22545), AOM_ICDF(23553),
         AOM_ICDF(32768) },
        {AOM_ICDF(12353), AOM_ICDF(12354), AOM_ICDF(12473), AOM_ICDF(12532),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(27457), AOM_ICDF(30005), AOM_ICDF(32738), AOM_ICDF(32745),
         AOM_ICDF(32768) },
        {AOM_ICDF(24897), AOM_ICDF(27541), AOM_ICDF(32723), AOM_ICDF(32731),
         AOM_ICDF(32768) },
        {AOM_ICDF(15297), AOM_ICDF(19106), AOM_ICDF(30414), AOM_ICDF(30711),
         AOM_ICDF(32768) },
        {AOM_ICDF(6593), AOM_ICDF(8826), AOM_ICDF(19732), AOM_ICDF(20840),
         AOM_ICDF(32768) },
        {AOM_ICDF(4161), AOM_ICDF(4233), AOM_ICDF(16509), AOM_ICDF(16557),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2652), AOM_ICDF(7276), AOM_ICDF(7351),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(28609), AOM_ICDF(30482), AOM_ICDF(32761), AOM_ICDF(32763),
         AOM_ICDF(32768) },
        {AOM_ICDF(25665), AOM_ICDF(27830), AOM_ICDF(32727), AOM_ICDF(32733),
         AOM_ICDF(32768) },
        {AOM_ICDF(21057), AOM_ICDF(23803), AOM_ICDF(30367), AOM_ICDF(30721),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(21878), AOM_ICDF(32726), AOM_ICDF(32737),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(5750), AOM_ICDF(14739), AOM_ICDF(14792),
         AOM_ICDF(32768) },
        {AOM_ICDF(2881), AOM_ICDF(2913), AOM_ICDF(8427), AOM_ICDF(8498),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(28993), AOM_ICDF(30944), AOM_ICDF(32762), AOM_ICDF(32764),
         AOM_ICDF(32768) },
        {AOM_ICDF(26561), AOM_ICDF(28695), AOM_ICDF(32733), AOM_ICDF(32739),
         AOM_ICDF(32768) },
        {AOM_ICDF(17985), AOM_ICDF(19028), AOM_ICDF(31008), AOM_ICDF(31079),
         AOM_ICDF(32768) },
        {AOM_ICDF(7873), AOM_ICDF(8039), AOM_ICDF(19981), AOM_ICDF(20068),
         AOM_ICDF(32768) },
        {AOM_ICDF(5313), AOM_ICDF(5366), AOM_ICDF(14376), AOM_ICDF(14430),
         AOM_ICDF(32768) },
        {AOM_ICDF(2753), AOM_ICDF(2789), AOM_ICDF(8909), AOM_ICDF(8979),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(30273), AOM_ICDF(32029), AOM_ICDF(32764), AOM_ICDF(32766),
         AOM_ICDF(32768) },
        {AOM_ICDF(28609), AOM_ICDF(30847), AOM_ICDF(32745), AOM_ICDF(32751),
         AOM_ICDF(32768) },
        {AOM_ICDF(21313), AOM_ICDF(24377), AOM_ICDF(31986), AOM_ICDF(32098),
         AOM_ICDF(32768) },
        {AOM_ICDF(32705), AOM_ICDF(32709), AOM_ICDF(32739), AOM_ICDF(32741),
         AOM_ICDF(32768) },
        {AOM_ICDF(4929), AOM_ICDF(5579), AOM_ICDF(16402), AOM_ICDF(16866),
         AOM_ICDF(32768) },
        {AOM_ICDF(3009), AOM_ICDF(3246), AOM_ICDF(10158), AOM_ICDF(10533),
         AOM_ICDF(32768) } } } }
};
static const coeff_cdf_model default_coef_head_cdf_32x32[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { AOM_ICDF(2240), AOM_ICDF(5407), AOM_ICDF(18304), AOM_ICDF(25601),
          AOM_ICDF(27911), AOM_ICDF(32768) },
        { AOM_ICDF(960), AOM_ICDF(4633), AOM_ICDF(8197), AOM_ICDF(16254),
          AOM_ICDF(18796), AOM_ICDF(32768) },
        { AOM_ICDF(192), AOM_ICDF(3061), AOM_ICDF(3557), AOM_ICDF(8701),
          AOM_ICDF(9762), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(11969), AOM_ICDF(15846), AOM_ICDF(25660), AOM_ICDF(26667),
         AOM_ICDF(32768) },
        {AOM_ICDF(11713), AOM_ICDF(15794), AOM_ICDF(25737), AOM_ICDF(26760),
         AOM_ICDF(32768) },
        {AOM_ICDF(9281), AOM_ICDF(12675), AOM_ICDF(23181), AOM_ICDF(24351),
         AOM_ICDF(32768) },
        {AOM_ICDF(7105), AOM_ICDF(8757), AOM_ICDF(18383), AOM_ICDF(19437),
         AOM_ICDF(32768) },
        {AOM_ICDF(4289), AOM_ICDF(4579), AOM_ICDF(11353), AOM_ICDF(11792),
         AOM_ICDF(32768) },
        {AOM_ICDF(1857), AOM_ICDF(1874), AOM_ICDF(4695), AOM_ICDF(4777),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(20929), AOM_ICDF(22297), AOM_ICDF(29370), AOM_ICDF(29646),
         AOM_ICDF(32768) },
        {AOM_ICDF(17473), AOM_ICDF(18985), AOM_ICDF(28079), AOM_ICDF(28413),
         AOM_ICDF(32768) },
        {AOM_ICDF(13121), AOM_ICDF(14064), AOM_ICDF(24902), AOM_ICDF(25217),
         AOM_ICDF(32768) },
        {AOM_ICDF(9793), AOM_ICDF(10214), AOM_ICDF(20069), AOM_ICDF(20329),
         AOM_ICDF(32768) },
        {AOM_ICDF(5825), AOM_ICDF(5987), AOM_ICDF(13350), AOM_ICDF(13559),
         AOM_ICDF(32768) },
        {AOM_ICDF(2241), AOM_ICDF(2260), AOM_ICDF(5520), AOM_ICDF(5600),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(25921), AOM_ICDF(26891), AOM_ICDF(31632), AOM_ICDF(31729),
         AOM_ICDF(32768) },
        {AOM_ICDF(18241), AOM_ICDF(19463), AOM_ICDF(29222), AOM_ICDF(29419),
         AOM_ICDF(32768) },
        {AOM_ICDF(11585), AOM_ICDF(12065), AOM_ICDF(23294), AOM_ICDF(23488),
         AOM_ICDF(32768) },
        {AOM_ICDF(6593), AOM_ICDF(6686), AOM_ICDF(16153), AOM_ICDF(16234),
         AOM_ICDF(32768) },
        {AOM_ICDF(3137), AOM_ICDF(3170), AOM_ICDF(8751), AOM_ICDF(8821),
         AOM_ICDF(32768) },
        {AOM_ICDF(1345), AOM_ICDF(1359), AOM_ICDF(3739), AOM_ICDF(3824),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(27713), AOM_ICDF(28504), AOM_ICDF(32068), AOM_ICDF(32132),
         AOM_ICDF(32768) },
        {AOM_ICDF(19265), AOM_ICDF(20354), AOM_ICDF(29789), AOM_ICDF(29943),
         AOM_ICDF(32768) },
        {AOM_ICDF(11201), AOM_ICDF(11538), AOM_ICDF(22701), AOM_ICDF(22848),
         AOM_ICDF(32768) },
        {AOM_ICDF(6337), AOM_ICDF(6424), AOM_ICDF(15268), AOM_ICDF(15353),
         AOM_ICDF(32768) },
        {AOM_ICDF(3649), AOM_ICDF(3681), AOM_ICDF(9052), AOM_ICDF(9121),
         AOM_ICDF(32768) },
        {AOM_ICDF(1601), AOM_ICDF(1618), AOM_ICDF(4584), AOM_ICDF(4667),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(30913), AOM_ICDF(31044), AOM_ICDF(32635), AOM_ICDF(32640),
         AOM_ICDF(32768) },
        {AOM_ICDF(22081), AOM_ICDF(22261), AOM_ICDF(30452), AOM_ICDF(30477),
         AOM_ICDF(32768) },
        {AOM_ICDF(10561), AOM_ICDF(10625), AOM_ICDF(21535), AOM_ICDF(21568),
         AOM_ICDF(32768) },
        {AOM_ICDF(6081), AOM_ICDF(6130), AOM_ICDF(14369), AOM_ICDF(14423),
         AOM_ICDF(32768) },
        {AOM_ICDF(3777), AOM_ICDF(3809), AOM_ICDF(9156), AOM_ICDF(9225),
         AOM_ICDF(32768) },
        {AOM_ICDF(1857), AOM_ICDF(1875), AOM_ICDF(4936), AOM_ICDF(5018),
         AOM_ICDF(32768) } } },
    {   // Intra
      { // Band 0
        { AOM_ICDF(4672), AOM_ICDF(6927), AOM_ICDF(23534), AOM_ICDF(29846),
          AOM_ICDF(30928), AOM_ICDF(32768) },
        { AOM_ICDF(3776), AOM_ICDF(6784), AOM_ICDF(18075), AOM_ICDF(25863),
          AOM_ICDF(27926), AOM_ICDF(32768) },
        { AOM_ICDF(1344), AOM_ICDF(5588), AOM_ICDF(12166), AOM_ICDF(20966),
          AOM_ICDF(23504), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(19393), AOM_ICDF(22016), AOM_ICDF(31280), AOM_ICDF(31444),
         AOM_ICDF(32768) },
        {AOM_ICDF(21185), AOM_ICDF(24329), AOM_ICDF(31706), AOM_ICDF(31865),
         AOM_ICDF(32768) },
        {AOM_ICDF(20673), AOM_ICDF(23240), AOM_ICDF(31186), AOM_ICDF(31379),
         AOM_ICDF(32768) },
        {AOM_ICDF(17857), AOM_ICDF(20035), AOM_ICDF(29594), AOM_ICDF(29889),
         AOM_ICDF(32768) },
        {AOM_ICDF(13633), AOM_ICDF(14929), AOM_ICDF(24883), AOM_ICDF(25337),
         AOM_ICDF(32768) },
        {AOM_ICDF(7873), AOM_ICDF(8416), AOM_ICDF(17452), AOM_ICDF(17886),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(25665), AOM_ICDF(27145), AOM_ICDF(32256), AOM_ICDF(32314),
         AOM_ICDF(32768) },
        {AOM_ICDF(21057), AOM_ICDF(22826), AOM_ICDF(31465), AOM_ICDF(31576),
         AOM_ICDF(32768) },
        {AOM_ICDF(13633), AOM_ICDF(14885), AOM_ICDF(27873), AOM_ICDF(28088),
         AOM_ICDF(32768) },
        {AOM_ICDF(8769), AOM_ICDF(9515), AOM_ICDF(21941), AOM_ICDF(22248),
         AOM_ICDF(32768) },
        {AOM_ICDF(6209), AOM_ICDF(6594), AOM_ICDF(15598), AOM_ICDF(15950),
         AOM_ICDF(32768) },
        {AOM_ICDF(1985), AOM_ICDF(2014), AOM_ICDF(6855), AOM_ICDF(6931),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(26817), AOM_ICDF(27824), AOM_ICDF(32362), AOM_ICDF(32399),
         AOM_ICDF(32768) },
        {AOM_ICDF(21185), AOM_ICDF(22321), AOM_ICDF(31389), AOM_ICDF(31466),
         AOM_ICDF(32768) },
        {AOM_ICDF(13761), AOM_ICDF(14154), AOM_ICDF(27163), AOM_ICDF(27245),
         AOM_ICDF(32768) },
        {AOM_ICDF(8897), AOM_ICDF(9011), AOM_ICDF(20600), AOM_ICDF(20659),
         AOM_ICDF(32768) },
        {AOM_ICDF(4673), AOM_ICDF(4774), AOM_ICDF(15044), AOM_ICDF(15131),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(194), AOM_ICDF(384), AOM_ICDF(479),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(28865), AOM_ICDF(29687), AOM_ICDF(32655), AOM_ICDF(32667),
         AOM_ICDF(32768) },
        {AOM_ICDF(23233), AOM_ICDF(24218), AOM_ICDF(32080), AOM_ICDF(32118),
         AOM_ICDF(32768) },
        {AOM_ICDF(15041), AOM_ICDF(15444), AOM_ICDF(28787), AOM_ICDF(28845),
         AOM_ICDF(32768) },
        {AOM_ICDF(9921), AOM_ICDF(10248), AOM_ICDF(22818), AOM_ICDF(22944),
         AOM_ICDF(32768) },
        {AOM_ICDF(7745), AOM_ICDF(7866), AOM_ICDF(16591), AOM_ICDF(16702),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(194), AOM_ICDF(384), AOM_ICDF(479),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(31169), AOM_ICDF(31559), AOM_ICDF(32741), AOM_ICDF(32744),
         AOM_ICDF(32768) },
        {AOM_ICDF(24769), AOM_ICDF(25583), AOM_ICDF(32347), AOM_ICDF(32370),
         AOM_ICDF(32768) },
        {AOM_ICDF(15937), AOM_ICDF(16169), AOM_ICDF(29120), AOM_ICDF(29152),
         AOM_ICDF(32768) },
        {AOM_ICDF(7489), AOM_ICDF(7578), AOM_ICDF(22647), AOM_ICDF(22677),
         AOM_ICDF(32768) },
        {AOM_ICDF(7617), AOM_ICDF(7689), AOM_ICDF(19849), AOM_ICDF(19887),
         AOM_ICDF(32768) },
        {AOM_ICDF(2113), AOM_ICDF(2183), AOM_ICDF(7202), AOM_ICDF(7377),
         AOM_ICDF(32768) } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { AOM_ICDF(23232), AOM_ICDF(24301), AOM_ICDF(30231), AOM_ICDF(31582),
          AOM_ICDF(32091), AOM_ICDF(32768) },
        { AOM_ICDF(7872), AOM_ICDF(11041), AOM_ICDF(22542), AOM_ICDF(27086),
          AOM_ICDF(29145), AOM_ICDF(32768) },
        { AOM_ICDF(1344), AOM_ICDF(3989), AOM_ICDF(18125), AOM_ICDF(25340),
          AOM_ICDF(27820), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(15937), AOM_ICDF(29000), AOM_ICDF(32210), AOM_ICDF(32434),
         AOM_ICDF(32768) },
        {AOM_ICDF(12353), AOM_ICDF(26626), AOM_ICDF(31533), AOM_ICDF(31993),
         AOM_ICDF(32768) },
        {AOM_ICDF(11457), AOM_ICDF(29187), AOM_ICDF(30896), AOM_ICDF(31750),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(21278), AOM_ICDF(28169), AOM_ICDF(29764),
         AOM_ICDF(32768) },
        {AOM_ICDF(7489), AOM_ICDF(8855), AOM_ICDF(13365), AOM_ICDF(15620),
         AOM_ICDF(32768) },
        {AOM_ICDF(4289), AOM_ICDF(4833), AOM_ICDF(8572), AOM_ICDF(10108),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(25025), AOM_ICDF(30783), AOM_ICDF(32603), AOM_ICDF(32666),
         AOM_ICDF(32768) },
        {AOM_ICDF(24385), AOM_ICDF(29586), AOM_ICDF(31803), AOM_ICDF(32142),
         AOM_ICDF(32768) },
        {AOM_ICDF(22337), AOM_ICDF(23002), AOM_ICDF(27573), AOM_ICDF(27903),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(12336), AOM_ICDF(21900), AOM_ICDF(22590),
         AOM_ICDF(32768) },
        {AOM_ICDF(8257), AOM_ICDF(8830), AOM_ICDF(19986), AOM_ICDF(20298),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(10990), AOM_ICDF(18660), AOM_ICDF(18701),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(29761), AOM_ICDF(31473), AOM_ICDF(32693), AOM_ICDF(32715),
         AOM_ICDF(32768) },
        {AOM_ICDF(20417), AOM_ICDF(24512), AOM_ICDF(31394), AOM_ICDF(31650),
         AOM_ICDF(32768) },
        {AOM_ICDF(11713), AOM_ICDF(13283), AOM_ICDF(25819), AOM_ICDF(26206),
         AOM_ICDF(32768) },
        {AOM_ICDF(13121), AOM_ICDF(14099), AOM_ICDF(21909), AOM_ICDF(22514),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(248), AOM_ICDF(9546), AOM_ICDF(9614),
         AOM_ICDF(32768) },
        {AOM_ICDF(2497), AOM_ICDF(2524), AOM_ICDF(7050), AOM_ICDF(7125),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(30657), AOM_ICDF(31885), AOM_ICDF(32691), AOM_ICDF(32715),
         AOM_ICDF(32768) },
        {AOM_ICDF(19393), AOM_ICDF(26050), AOM_ICDF(31698), AOM_ICDF(31988),
         AOM_ICDF(32768) },
        {AOM_ICDF(15809), AOM_ICDF(15863), AOM_ICDF(24985), AOM_ICDF(25008),
         AOM_ICDF(32768) },
        {AOM_ICDF(23489), AOM_ICDF(28138), AOM_ICDF(32751), AOM_ICDF(32756),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(16450), AOM_ICDF(16545), AOM_ICDF(16593),
         AOM_ICDF(32768) },
        {AOM_ICDF(2369), AOM_ICDF(2395), AOM_ICDF(6822), AOM_ICDF(6898),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(32705), AOM_ICDF(32744), AOM_ICDF(32766), AOM_ICDF(32767),
         AOM_ICDF(32768) },
        {AOM_ICDF(21953), AOM_ICDF(24962), AOM_ICDF(32156), AOM_ICDF(32246),
         AOM_ICDF(32768) },
        {AOM_ICDF(13121), AOM_ICDF(15358), AOM_ICDF(26284), AOM_ICDF(26835),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(7417), AOM_ICDF(20132), AOM_ICDF(20885),
         AOM_ICDF(32768) },
        {AOM_ICDF(4417), AOM_ICDF(4939), AOM_ICDF(15104), AOM_ICDF(15535),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2680), AOM_ICDF(8218), AOM_ICDF(8338),
         AOM_ICDF(32768) } } },
    {   // Inter
      { // Band 0
        { AOM_ICDF(25280), AOM_ICDF(25678), AOM_ICDF(32446), AOM_ICDF(32622),
          AOM_ICDF(32724), AOM_ICDF(32768) },
        { AOM_ICDF(10560), AOM_ICDF(11822), AOM_ICDF(28682), AOM_ICDF(29919),
          AOM_ICDF(31276), AOM_ICDF(32768) },
        { AOM_ICDF(3264), AOM_ICDF(5170), AOM_ICDF(21779), AOM_ICDF(24026),
          AOM_ICDF(27905), AOM_ICDF(32768) } },
      { // Band 1
        {AOM_ICDF(24257), AOM_ICDF(30554), AOM_ICDF(32719), AOM_ICDF(32738),
         AOM_ICDF(32768) },
        {AOM_ICDF(17217), AOM_ICDF(27413), AOM_ICDF(32617), AOM_ICDF(32667),
         AOM_ICDF(32768) },
        {AOM_ICDF(22977), AOM_ICDF(27600), AOM_ICDF(32482), AOM_ICDF(32552),
         AOM_ICDF(32768) },
        {AOM_ICDF(16833), AOM_ICDF(24360), AOM_ICDF(30746), AOM_ICDF(31293),
         AOM_ICDF(32768) },
        {AOM_ICDF(17089), AOM_ICDF(20060), AOM_ICDF(28880), AOM_ICDF(29370),
         AOM_ICDF(32768) },
        {AOM_ICDF(10945), AOM_ICDF(11009), AOM_ICDF(21900), AOM_ICDF(21932),
         AOM_ICDF(32768) } },
      { // Band 2
        {AOM_ICDF(27201), AOM_ICDF(30217), AOM_ICDF(32736), AOM_ICDF(32745),
         AOM_ICDF(32768) },
        {AOM_ICDF(22721), AOM_ICDF(27676), AOM_ICDF(32749), AOM_ICDF(32754),
         AOM_ICDF(32768) },
        {AOM_ICDF(5057), AOM_ICDF(12431), AOM_ICDF(25246), AOM_ICDF(26620),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(321), AOM_ICDF(22016), AOM_ICDF(22048),
         AOM_ICDF(32768) },
        {AOM_ICDF(5313), AOM_ICDF(5363), AOM_ICDF(13839), AOM_ICDF(13894),
         AOM_ICDF(32768) },
        {AOM_ICDF(2625), AOM_ICDF(2652), AOM_ICDF(7276), AOM_ICDF(7351),
         AOM_ICDF(32768) } },
      { // Band 3
        {AOM_ICDF(27713), AOM_ICDF(30739), AOM_ICDF(32759), AOM_ICDF(32762),
         AOM_ICDF(32768) },
        {AOM_ICDF(26177), AOM_ICDF(30430), AOM_ICDF(32756), AOM_ICDF(32760),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(384), AOM_ICDF(32706), AOM_ICDF(32707),
         AOM_ICDF(32768) },
        {AOM_ICDF(9409), AOM_ICDF(9528), AOM_ICDF(21591), AOM_ICDF(21646),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(194), AOM_ICDF(384), AOM_ICDF(479),
         AOM_ICDF(32768) },
        {AOM_ICDF(2881), AOM_ICDF(2913), AOM_ICDF(8427), AOM_ICDF(8498),
         AOM_ICDF(32768) } },
      { // Band 4
        {AOM_ICDF(28993), AOM_ICDF(31156), AOM_ICDF(32747), AOM_ICDF(32753),
         AOM_ICDF(32768) },
        {AOM_ICDF(25153), AOM_ICDF(28701), AOM_ICDF(32754), AOM_ICDF(32758),
         AOM_ICDF(32768) },
        {AOM_ICDF(16449), AOM_ICDF(16544), AOM_ICDF(32737), AOM_ICDF(32738),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(321), AOM_ICDF(22016), AOM_ICDF(22048),
         AOM_ICDF(32768) },
        {AOM_ICDF(193), AOM_ICDF(194), AOM_ICDF(384), AOM_ICDF(479),
         AOM_ICDF(32768) },
        {AOM_ICDF(2753), AOM_ICDF(2789), AOM_ICDF(8909), AOM_ICDF(8979),
         AOM_ICDF(32768) } },
      { // Band 5
        {AOM_ICDF(30785), AOM_ICDF(32088), AOM_ICDF(32765), AOM_ICDF(32766),
         AOM_ICDF(32768) },
        {AOM_ICDF(22977), AOM_ICDF(26623), AOM_ICDF(32750), AOM_ICDF(32754),
         AOM_ICDF(32768) },
        {AOM_ICDF(21953), AOM_ICDF(21954), AOM_ICDF(22017), AOM_ICDF(22049),
         AOM_ICDF(32768) },
        {AOM_ICDF(5697), AOM_ICDF(7486), AOM_ICDF(20238), AOM_ICDF(21009),
         AOM_ICDF(32768) },
        {AOM_ICDF(4929), AOM_ICDF(5579), AOM_ICDF(16402), AOM_ICDF(16866),
         AOM_ICDF(32768) },
        {AOM_ICDF(3009), AOM_ICDF(3246), AOM_ICDF(10158), AOM_ICDF(10533),
         AOM_ICDF(32768) } } } }
};

/* clang-format on */

static void extend_to_full_distribution(aom_prob *probs, aom_prob p) {
  assert(p != 0);
  memcpy(probs, av1_pareto8_full[p - 1], MODEL_NODES * sizeof(aom_prob));
}

void av1_model_to_full_probs(const aom_prob *model, aom_prob *full) {
  if (full != model)
    memcpy(full, model, sizeof(aom_prob) * UNCONSTRAINED_NODES);
  extend_to_full_distribution(&full[UNCONSTRAINED_NODES], model[PIVOT_NODE]);
}

static void build_tail_cdfs(aom_cdf_prob cdf_tail[CDF_SIZE(ENTROPY_TOKENS)],
                            aom_cdf_prob cdf_head[CDF_SIZE(ENTROPY_TOKENS)],
                            int band_zero) {
  int probNZ, prob1, prob_idx, i;
  int phead[HEAD_TOKENS + 1], sum;
  const int is_dc = !!band_zero;
  aom_cdf_prob prev_cdf;
  prev_cdf = 0;
  for (i = 0; i < HEAD_TOKENS + is_dc; ++i) {
    phead[i] = AOM_ICDF(cdf_head[i]) - prev_cdf;
    prev_cdf = AOM_ICDF(cdf_head[i]);
  }
  // Do the tail
  probNZ = CDF_PROB_TOP - phead[ZERO_TOKEN + is_dc] - (is_dc ? phead[0] : 0);
  prob1 = phead[is_dc + ONE_TOKEN_EOB] + phead[is_dc + ONE_TOKEN_NEOB];
  prob_idx =
      AOMMIN(COEFF_PROB_MODELS - 1, AOMMAX(0, ((256 * prob1) / probNZ) - 1));

  sum = 0;
  for (i = 0; i < TAIL_TOKENS; ++i) {
    sum += av1_pareto8_tail_probs[prob_idx][i];
    cdf_tail[i] = AOM_ICDF(sum);
  }
}

static void build_head_cdfs(const aom_prob *pdf_model,
                            const aom_prob *blockz_model,
                            aom_cdf_prob cdf_head[ENTROPY_TOKENS + 1]) {
  int i, p, p1, p2, phead[6], prob_NZ, prob_EOB_1, prob_EOB_2p, prob_NEOB_1,
      prob_NEOB_2p;
  int prob8_blocknz;
  // We have the first coefficient position and so an extended CDF
  const int is_dc = blockz_model != NULL;
  const int last_head_val = HEAD_TOKENS - 1 + is_dc;

  assert(pdf_model != NULL);
  assert(pdf_model[2] != 0);

  /* FIXME: maintain true CDF counts. */

  /* Values are 0=BLOCK_ZERO 1=ZERO_TOKEN, 2=ONE_TOKEN_EOB
     3=ONE_TOKEN_NEOB, 4=TWO_TOKEN_PLUS_EOB, 5=TWO_TOKEN_PLUS_NEOB
     */
  // Block zero probability
  if (is_dc) {
    phead[0] =
        ((*blockz_model) << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));
    phead[0] = AOMMIN(CDF_PROB_TOP - (HEAD_TOKENS + 1), AOMMAX(1, phead[0]));
  }

  // Will scale the remaining probabilities by the probability of the block
  // being non-zero
  prob8_blocknz = is_dc ? (256 - *blockz_model) : 256;

  // Probability of zero
  phead[is_dc + ZERO_TOKEN] =
      (pdf_model[1] << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));

  // Will scale the non-zero values
  prob_NZ = CDF_PROB_TOP - phead[is_dc + ZERO_TOKEN];

  // Will scale the EOBs by the probability of and EOB_TOKEN ..
  prob_EOB_1 =
      (pdf_model[0] << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));
  // .. use a lower probability of EOB for larger values
  prob_EOB_2p = prob_EOB_1 / 2;

  prob_NEOB_1 = CDF_PROB_TOP - prob_EOB_1;
  prob_NEOB_2p = CDF_PROB_TOP - prob_EOB_2p;
  if (prob_NZ == 0 || prob_NZ == CDF_PROB_TOP) abort();
  if (prob_EOB_1 == 0 || prob_EOB_1 == CDF_PROB_TOP) abort();
  if (prob_EOB_2p == 0 || prob_EOB_2p == CDF_PROB_TOP) abort();

  // ONE_CONTEXT_NODE prob
  p = (pdf_model[2] << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));
  // Scale by the non-zero factor to get the probability of token = 1
  p1 = ROUND_POWER_OF_TWO(prob_NZ * p, 15);

  // Scale by the EOB factors
  phead[is_dc + ONE_TOKEN_EOB] = ROUND_POWER_OF_TWO(p1 * prob_EOB_1, 15);
  phead[is_dc + ONE_TOKEN_NEOB] = ROUND_POWER_OF_TWO(p1 * prob_NEOB_1, 15);

  // Probability token is 2 or more
  p2 = CDF_PROB_TOP - p1 - phead[is_dc + ZERO_TOKEN];

  phead[is_dc + TWO_TOKEN_PLUS_EOB] = ROUND_POWER_OF_TWO(p2 * prob_EOB_2p, 15);
  phead[is_dc + TWO_TOKEN_PLUS_NEOB] =
      ROUND_POWER_OF_TWO(p2 * prob_NEOB_2p, 15);

  // Now use block non-zerp prob to scale the values
  for (i = is_dc; i < last_head_val; ++i) {
    phead[i] = (prob8_blocknz * phead[i] + 128) >> 8;
  }

  for (i = 0; i < last_head_val; ++i) {
    int c0;
    c0 = i > 0 ? AOM_ICDF(cdf_head[i - 1]) : 0;
    p = AOMMAX(1, AOMMIN(CDF_PROB_TOP - (last_head_val - i) - c0, phead[i]));
    cdf_head[i] = AOM_ICDF(c0 + p);
  }
  cdf_head[last_head_val] = AOM_ICDF(CDF_PROB_TOP);
}

static void av1_default_coef_cdfs(FRAME_CONTEXT *fc) {
  int i, j, k, l;
  for (i = 0; i < PLANE_TYPES; ++i)
    for (j = 0; j < REF_TYPES; ++j)
      for (k = 0; k < COEF_BANDS; ++k)
        for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
#if CONFIG_CHROMA_2X2
          av1_copy(fc->coef_head_cdfs[TX_2X2][i][j][k][l],
                   default_coef_head_cdf_4x4[i][j][k][l]);
#endif
          av1_copy(fc->coef_head_cdfs[TX_4X4][i][j][k][l],
                   default_coef_head_cdf_4x4[i][j][k][l]);
          av1_copy(fc->coef_head_cdfs[TX_8X8][i][j][k][l],
                   default_coef_head_cdf_8x8[i][j][k][l]);
          av1_copy(fc->coef_head_cdfs[TX_16X16][i][j][k][l],
                   default_coef_head_cdf_16x16[i][j][k][l]);
          av1_copy(fc->coef_head_cdfs[TX_32X32][i][j][k][l],
                   default_coef_head_cdf_32x32[i][j][k][l]);
#if CONFIG_TX64X64
          av1_copy(fc->coef_head_cdfs[TX_64X64][i][j][k][l],
                   default_coef_head_cdf_32x32[i][j][k][l]);
#endif
        }
}

void av1_coef_head_cdfs(FRAME_CONTEXT *fc) {
  TX_SIZE t;
  int i, j, k, l;
  for (t = 0; t < TX_SIZES; ++t)
    for (i = 0; i < PLANE_TYPES; ++i)
      for (j = 0; j < REF_TYPES; ++j)
        for (k = 0; k < COEF_BANDS; ++k)
          for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
            build_head_cdfs(fc->coef_probs[t][i][j][k][l],
                            k == 0 ? &fc->blockzero_probs[t][i][j][l] : NULL,
                            fc->coef_head_cdfs[t][i][j][k][l]);
          }
}

void av1_coef_pareto_cdfs(FRAME_CONTEXT *fc) {
  /* Build the tail based on a Pareto distribution */
  TX_SIZE t;
  int i, j, k, l;
  for (t = 0; t < TX_SIZES; ++t)
    for (i = 0; i < PLANE_TYPES; ++i)
      for (j = 0; j < REF_TYPES; ++j)
        for (k = 0; k < COEF_BANDS; ++k)
          for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l)
            build_tail_cdfs(fc->coef_tail_cdfs[t][i][j][k][l],
                            fc->coef_head_cdfs[t][i][j][k][l], k == 0);
}

void av1_default_coef_probs(AV1_COMMON *cm) {
#if CONFIG_Q_ADAPT_PROBS
  const int index = AOMMIN(
      ROUND_POWER_OF_TWO(cm->base_qindex, 8 - QCTX_BIN_BITS), QCTX_BINS - 1);
  av1_copy(cm->fc->coef_probs, default_qctx_coef_probs[index]);
#else
#if CONFIG_CHROMA_2X2
  av1_copy(cm->fc->coef_probs[TX_2X2], default_coef_probs_4x4);
#endif
  av1_copy(cm->fc->coef_probs[TX_4X4], default_coef_probs_4x4);
  av1_copy(cm->fc->coef_probs[TX_8X8], default_coef_probs_8x8);
  av1_copy(cm->fc->coef_probs[TX_16X16], default_coef_probs_16x16);
  av1_copy(cm->fc->coef_probs[TX_32X32], default_coef_probs_32x32);
#if CONFIG_TX64X64
  av1_copy(cm->fc->coef_probs[TX_64X64], default_coef_probs_64x64);
#endif  // CONFIG_TX64X64
#endif  // CONFIG_Q_ADAPT_PROBS
  av1_copy(cm->fc->blockzero_probs, av1_default_blockzero_probs);
  /* Load the head tokens */
  av1_default_coef_cdfs(cm->fc);
  av1_coef_pareto_cdfs(cm->fc);
}

#if !CONFIG_LV_MAP
static void adapt_coef_probs(AV1_COMMON *cm, TX_SIZE tx_size,
                             unsigned int count_sat,
                             unsigned int update_factor) {
  const FRAME_CONTEXT *pre_fc = cm->pre_fc;
  av1_coeff_probs_model *const probs = cm->fc->coef_probs[tx_size];
  const av1_coeff_probs_model *const pre_probs = pre_fc->coef_probs[tx_size];
  const av1_coeff_count_model *const counts =
      (const av1_coeff_count_model *)cm->counts.coef[tx_size];
  const unsigned int(*eob_counts)[REF_TYPES][COEF_BANDS][COEFF_CONTEXTS] =
      (const unsigned int(*)[REF_TYPES][COEF_BANDS][COEFF_CONTEXTS])
          cm->counts.eob_branch[tx_size];
  const av1_blockz_probs_model *const pre_blockz_probs =
      pre_fc->blockzero_probs[tx_size];
  av1_blockz_probs_model *const blockz_probs = cm->fc->blockzero_probs[tx_size];
  const av1_blockz_count_model *const blockz_counts =
      (const av1_blockz_count_model *)&cm->counts.blockz_count[tx_size][0];
  int i, j, k, l, m;
#if CONFIG_RECT_TX
  assert(!is_rect_tx(tx_size));
#endif  // CONFIG_RECT_TX

  for (i = 0; i < PLANE_TYPES; ++i)
    for (j = 0; j < REF_TYPES; ++j)
      for (k = 0; k < COEF_BANDS; ++k)
        for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
          const int n0 = counts[i][j][k][l][ZERO_TOKEN];
          const int n1 = counts[i][j][k][l][ONE_TOKEN];
          const int n2 = counts[i][j][k][l][TWO_TOKEN];
          const int neob = counts[i][j][k][l][EOB_MODEL_TOKEN];
          const unsigned int branch_ct[UNCONSTRAINED_NODES][2] = {
            { neob, eob_counts[i][j][k][l] - neob }, { n0, n1 + n2 }, { n1, n2 }
          };
          for (m = 0; m < UNCONSTRAINED_NODES; ++m)
            probs[i][j][k][l][m] =
                av1_merge_probs(pre_probs[i][j][k][l][m], branch_ct[m],
                                count_sat, update_factor);
        }

  for (i = 0; i < PLANE_TYPES; ++i) {
    for (j = 0; j < REF_TYPES; ++j) {
      for (k = 0; k < BLOCKZ_CONTEXTS; ++k) {
        const int n0 = blockz_counts[i][j][k][0];
        const int n1 = blockz_counts[i][j][k][1];
        const unsigned int branch_ct[2] = { n0, n1 };
        blockz_probs[i][j][k] = av1_merge_probs(
            pre_blockz_probs[i][j][k], branch_ct, count_sat, update_factor);
      }
    }
  }
}
#endif  // !CONFIG_LV_MAP

void av1_adapt_coef_probs(AV1_COMMON *cm) {
  unsigned int count_sat, update_factor;

  if (!frame_is_intra_only(cm) && cm->last_frame_type == KEY_FRAME) {
    update_factor = COEF_MAX_UPDATE_FACTOR_AFTER_KEY; /* adapt quickly */
    count_sat = COEF_COUNT_SAT_AFTER_KEY;
  } else {
    update_factor = COEF_MAX_UPDATE_FACTOR;
    count_sat = COEF_COUNT_SAT;
  }

#if CONFIG_LV_MAP
  av1_adapt_txb_probs(cm, count_sat, update_factor);
#else
  TX_SIZE tx_size;
  for (tx_size = 0; tx_size < TX_SIZES; tx_size++)
    adapt_coef_probs(cm, tx_size, count_sat, update_factor);
#endif
}

#if CONFIG_EC_ADAPT
static void av1_average_cdf(aom_cdf_prob *cdf_ptr[], aom_cdf_prob *fc_cdf_ptr,
                            int cdf_size, const int num_tiles) {
  int i;
  for (i = 0; i < cdf_size;) {
    do {
      int sum = 0;
      int j;
      assert(i < cdf_size);
      for (j = 0; j < num_tiles; ++j) sum += AOM_ICDF(cdf_ptr[j][i]);
      fc_cdf_ptr[i] = AOM_ICDF(sum / num_tiles);
    } while (fc_cdf_ptr[i++] != AOM_ICDF(CDF_PROB_TOP));
    // Zero symbol counts for the next frame
    assert(i < cdf_size);
    fc_cdf_ptr[i++] = 0;
    // Skip trailing zeros until the start of the next CDF.
    for (; i < cdf_size && fc_cdf_ptr[i] == 0; ++i) {
    }
  }
}

#define AVERAGE_TILE_CDFS(cname)                            \
  for (i = 0; i < num_tiles; ++i)                           \
    cdf_ptr[i] = (aom_cdf_prob *)&ec_ctxs[i]->cname;        \
  fc_cdf_ptr = (aom_cdf_prob *)&fc->cname;                  \
  cdf_size = (int)sizeof(fc->cname) / sizeof(aom_cdf_prob); \
  av1_average_cdf(cdf_ptr, fc_cdf_ptr, cdf_size, num_tiles);

void av1_average_tile_coef_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                                aom_cdf_prob *cdf_ptr[], int num_tiles) {
  int i, cdf_size;

  aom_cdf_prob *fc_cdf_ptr;

  AVERAGE_TILE_CDFS(coef_head_cdfs)
  AVERAGE_TILE_CDFS(coef_tail_cdfs)
}

void av1_average_tile_mv_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                              aom_cdf_prob *cdf_ptr[], int num_tiles) {
  int i, k, cdf_size;

  aom_cdf_prob *fc_cdf_ptr;

  int j;
  for (j = 0; j < NMV_CONTEXTS; ++j) {
    AVERAGE_TILE_CDFS(nmvc[j].joint_cdf)

    for (k = 0; k < 2; ++k) {
      AVERAGE_TILE_CDFS(nmvc[j].comps[k].class_cdf);
      AVERAGE_TILE_CDFS(nmvc[j].comps[k].class0_fp_cdf);
      AVERAGE_TILE_CDFS(nmvc[j].comps[k].fp_cdf);
    }
  }
}

void av1_average_tile_intra_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                                 aom_cdf_prob *cdf_ptr[], int num_tiles) {
  int i, cdf_size;

  aom_cdf_prob *fc_cdf_ptr;

  AVERAGE_TILE_CDFS(tx_size_cdf);

#if CONFIG_VAR_TX
// FIXME: txfm_partition probs
#endif

  // FIXME: skip probs

  AVERAGE_TILE_CDFS(intra_ext_tx_cdf)
  AVERAGE_TILE_CDFS(inter_ext_tx_cdf);

  AVERAGE_TILE_CDFS(seg.tree_cdf)
  AVERAGE_TILE_CDFS(uv_mode_cdf)

  AVERAGE_TILE_CDFS(partition_cdf)

#if CONFIG_DELTA_Q
  AVERAGE_TILE_CDFS(delta_q_cdf)
#if CONFIG_EXT_DELTA_Q
  AVERAGE_TILE_CDFS(delta_lf_cdf)
#endif
#endif
#if CONFIG_EXT_INTRA && CONFIG_INTRA_INTERP
  AVERAGE_TILE_CDFS(intra_filter_cdf)
#endif  // CONFIG_EXT_INTRA && CONFIG_INTRA_INTERP
}

void av1_average_tile_inter_cdfs(AV1_COMMON *cm, FRAME_CONTEXT *fc,
                                 FRAME_CONTEXT *ec_ctxs[],
                                 aom_cdf_prob *cdf_ptr[], int num_tiles) {
  int i, cdf_size;

  aom_cdf_prob *fc_cdf_ptr;

  // FIXME: comp_inter_cdf not defined

  // FIXME: comp_ref_cdf and comp_bwd_ref not defined

  // FIXME: single_ref_cdf not defined

  // FIXME: cdfs not defined for newmv_mode, zeromv_mode, drl_mode, new2mv_mode

  // FIXME: cdfs not defined for motion_mode_prob, obmc_prob

  // FIXME: cdfs not defined for super_tx

  // FIXME: CONFIG_EXT_INTER cdfs not defined for inter_compound_mode,
  // interintra_mode etc

  AVERAGE_TILE_CDFS(y_mode_cdf)

  if (cm->interp_filter == SWITCHABLE) {
    AVERAGE_TILE_CDFS(switchable_interp_cdf)
  }
}

#if CONFIG_PVQ
// Averaging PVQ's expected values for symbol coding
static void av1_average_pvq_ex(int *cxt_ptr[], int *fc_cxt_ptr, int cxt_size,
                               const int num_tiles) {
  int i, j;
  for (i = 0; i < cxt_size; ++i) {
    int sum = 0;
    for (j = 0; j < num_tiles; ++j) sum += cxt_ptr[j][i];
    fc_cxt_ptr[i] = sum / num_tiles;
  }
}

#define AVERAGE_TILE_PVQ_EX(cname)                                        \
  for (i = 0; i < num_tiles; ++i) cxt_ptr[i] = (int *)&ec_ctxs[i]->cname; \
  fc_cxt_ptr = (int *)&fc->cname;                                         \
  cxt_size = (int)sizeof(fc->cname) / sizeof(int);                        \
  av1_average_pvq_ex(cxt_ptr, fc_cxt_ptr, cxt_size, num_tiles);

void av1_default_pvq_probs(AV1_COMMON *cm) {
  od_adapt_ctx *adapt = &cm->fc->pvq_context;

  // Init with flat probabilities.
  od_adapt_ctx_reset(adapt, 0);

  // TODO(yushin): Prepare offline cdf and context table for PVQ,
  // i.e. od_adapt_ctx, then load them from table,
  // for example od_adapt_ctx default_pvq_context.
  // Then do sth like this:
  // av1_copy(cm->fc->pvq_context, default_pvq_context);
}

void av1_average_tile_pvq_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                               const int num_tiles) {
  int i, j, cdf_size, cxt_size;

  aom_cdf_prob *cdf_ptr[MAX_TILE_ROWS * MAX_TILE_COLS];
  aom_cdf_prob *fc_cdf_ptr;
  int *cxt_ptr[MAX_TILE_ROWS * MAX_TILE_COLS];
  int *fc_cxt_ptr;

  AVERAGE_TILE_PVQ_EX(pvq_context.ex_dc)
  AVERAGE_TILE_PVQ_EX(pvq_context.ex_g)

  for (j = 0; j < OD_NPLANES_MAX; j++) {
    AVERAGE_TILE_CDFS(pvq_context.model_dc[j].cdf)
  }

  AVERAGE_TILE_CDFS(pvq_context.skip_cdf)

  AVERAGE_TILE_PVQ_EX(pvq_context.pvq.pvq_codeword_ctx.pvq_adapt)
  AVERAGE_TILE_CDFS(pvq_context.pvq.pvq_codeword_ctx.pvq_k1_cdf)
  AVERAGE_TILE_CDFS(pvq_context.pvq.pvq_codeword_ctx.pvq_split_cdf)

  for (j = 0; j < 3; j++) {
    AVERAGE_TILE_CDFS(pvq_context.pvq.pvq_param_model[j].cdf)
  }

  AVERAGE_TILE_PVQ_EX(pvq_context.pvq.pvq_ext)
  AVERAGE_TILE_PVQ_EX(pvq_context.pvq.pvq_exg)
  AVERAGE_TILE_CDFS(pvq_context.pvq.pvq_gaintheta_cdf)
  AVERAGE_TILE_CDFS(pvq_context.pvq.pvq_skip_dir_cdf)
}
#endif  // CONFIG_PVQ
#endif  // CONFIG_EC_ADAPT
